// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.SLListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;

public class InteractiveProver implements java.io.Serializable {

    /** the proof the interactive prover works on */
    private Proof proof;
    /** the user focused goal */
    private Goal focusedGoal;
    /** list with the activated heuristics */
    private ListOfHeuristic activeHeuristics = SLListOfHeuristic.EMPTY_LIST;
    /** true iff in interactive mode */
    private boolean interactive = true;
    /** limits the maximal amount heuristics application before an
     * interaction is needed. Default is: 10
     */
    private int maxHeuristicsSteps = 10;

    /** the central heuristics processor including GUI signalling */
    private ApplyHeuristics applyHeuristics;
    // list of proof listeners and interactive proof listeners
    private List listenerList = Collections.synchronizedList(new ArrayList(5));
    /** a cached ProofEvent that is fired if heuristics have been
     * started or stopped
     */
    private final ProofEvent stockProofEvent = new ProofEvent(this);

    /** listens to the current selected proof and node */
    private KeYSelectionListener selListener;

    /** the mediator */
    private KeYMediator mediator;
    
    private boolean autoHeuristicsOn = false;

    /** creates a new interactive prover object 
     */
    public InteractiveProver(KeYMediator mediator) {
	selListener = new InteractiveProverKeYSelectionListener();
	this.mediator = mediator;
	mediator.addKeYSelectionListener(selListener);
	
    }

    /** returns the KeYMediator */
    KeYMediator mediator() {
	return mediator;
    }

    /** sets up a new proof 
     * @param p a Proof that contains the root of the proof tree
     */
    public void setProof(Proof p) {
	proof = p;
    }

    /** set maximal amount of heuristics steps before an interactive
     * one is needed
     */
    public void setMaxHeuristicSteps(int steps) {
	maxHeuristicsSteps = steps;
    }

    /** returns the maximal amount of heuristics steps before an interactive
     * one is needed
     * @return  the maximal amount of heuristics steps before an interactive
     * one is needed
     */
    public int getMaxHeuristicSteps() {
	return maxHeuristicsSteps;
    }

    /** adds an interactive ProofListener
     * @param p the InteractiveProofListener to be added
     */
    public synchronized void addInteractiveProofListener
	(InteractiveProofListener p) { 
	synchronized(listenerList) {
	    listenerList.add(p);
	}
    }

    /** removes an interactive ProofListener
     * @param p the InteractiveProofListener to be removed
     */
    public synchronized void removeInteractiveProofListener
	(InteractiveProofListener p) { 
	synchronized(listenerList) {	
	    listenerList.remove(p);
	}
    }

    /** fires the event that automatic execution has started */
    protected synchronized void fireAutoModeStarted(ProofEvent e) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    autoModeStarted(e);
	    }
	}
    }

    /** fires the event that automatic execution has stopped */
    protected synchronized void fireAutoModeStopped(ProofEvent e) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    autoModeStopped(e);
	    }
	}
    }
    
    

    public void setAutoHeuristics(boolean b) {
       autoHeuristicsOn = b;
    }
 

    public boolean autoHeuristics() {
	return autoHeuristicsOn;
    }

    public boolean interactiveMode() {
	return interactive;
    }

    public void setInteractive(boolean interact) {
	interactive=interact;
    }

    /** fires the event that a rule has been applied */
    protected synchronized void fireRuleApplied() {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    ruleApplied(stockProofEvent);
	    }
	}
    }

    /** The rule application is performed on the given goal
     * @param app the RuleApp to apply 
     * @param goal the goal where to apply the rule application
     */
    public void apply(RuleApp app, Goal goal) {
	Debug.assert(app instanceof TacletApp);
	Debug.out(" >>>>> INTERACTIVE APPLICATION <<<< ", app);
	Debug.outIfEqual("auto_insert_eq applied", 
			 ((Taclet)app.rule()).name(), "auto_insert_eq");
	TacletApp tacletApp = (TacletApp)app;
	JavaInfo javaInfo = mediator().getJavaInfo();
	if (tacletApp.taclet().matchIf(goal.node().sequent(), 
				   tacletApp.instantiations(), javaInfo)
	    != SLListOfSVInstantiations.EMPTY_LIST) {   
	    proof.replace(goal, tacletApp.execute(goal, javaInfo));
	} else {
	    proof.replace(goal, tacletApp.executeForceIf(goal, javaInfo));
	}
	fireRuleApplied();
	if (autoHeuristicsOn) startHeuristics();
    }

    /** returns the proof the interactive prover is working on
     * @return the proof the interactive prover is working on
     */
    public Proof getProof() {
	return proof;
    }    

    /** sets the active heuristics 
     * @param activeHeuristics the ListOfHeuristic containing all
     * activated Heuristics 
     */
    public void setActiveHeuristics(ListOfHeuristic activeHeuristics) {
	this.activeHeuristics = activeHeuristics;
    }

    /** returns active heuristics 
     */
    public ListOfHeuristic activeHeuristics() {
	return activeHeuristics;
    }

    /** starts the execution of rules with active heuristics 
     * @param maxApplicationOfHeuristics maximal allowed heuristcal
     * applications of Taclets
     */
    public void startHeuristics() {
	startHeuristics(proof.openGoals());
    }


    /** starts the execution of rules with active heuristics. The
     * heuristics will only be applied on the goals of the list that
     * is handed over and on the new goals an applied rule adds
     */
    public void startHeuristics(ListOfGoal goals) {
	if (activeHeuristics() == SLListOfHeuristic.EMPTY_LIST) {
	    mediator().popupWarning("No heuristics selected", "Oops...");
           return;
        }
	if (goals == SLListOfGoal.EMPTY_LIST) { 
	    mediator().popupWarning("No (more) goals", "Oops...");
	    return;
	}
	applyHeuristics = new ApplyHeuristics(this, proof, 
					      maxHeuristicsSteps);
	applyHeuristics.start(goals);
    }

    
    /** true iff pos describes a term not a sequent or semisequent */
    private boolean isTerm(PosInSequent pos) {
	return !pos.isSequent() && 
	    !pos.isAntecedent() && 
	    !pos.isSuccedent();
    }

    /** collects all applicable NoFindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable NoFindTaclets
     */
    ListOfTaclet getNoFindTaclet() {	
	return filterTaclet(focusedGoal.ruleAppIndex().
		       getNoFindTaclet(interactive,
				     activeHeuristics(), 
				     mediator.getJavaInfo()));
    }    

    /** collects all applicable FindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable FindTaclets
     */

    ListOfTaclet getFindTaclet(PosInSequent pos) {
	if (isTerm(pos) && (focusedGoal!=null) && (pos!=null)) {
	    return filterTaclet(focusedGoal.ruleAppIndex().
			      getFindTaclet(interactive,
					  pos.getPosInOccurrence(),
					  activeHeuristics(),
					  mediator.getJavaInfo()));
	}
	return SLListOfTaclet.EMPTY_LIST;
    }
    
    /** collects all applicable RewriteTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable RewriteTaclets
     */
    ListOfTaclet getRewriteTaclet(PosInSequent pos) {
	if (isTerm(pos))  {  
	    return filterTaclet(focusedGoal.ruleAppIndex().
		   getRewriteTaclet(interactive,
				  pos.getPosInOccurrence(),
				  activeHeuristics(),
				  mediator.getJavaInfo())); 
	}

	return SLListOfTaclet.EMPTY_LIST;
    }

    /**
     * takes NoPosTacletApps as arguments and returns a duplicate free list of
     * the contained Taclets
     */
    private ListOfTaclet filterTaclet(ListOfNoPosTacletApp tacletInstances) {
	java.util.HashSet applicableRules = new java.util.HashSet();
	ListOfTaclet result = SLListOfTaclet.EMPTY_LIST;
	IteratorOfNoPosTacletApp it = tacletInstances.iterator();		
	while (it.hasNext()) {
	    Taclet taclet = it.next().taclet();
	    if (!applicableRules.contains(taclet)) {
		applicableRules.add(taclet);
		result = result.prepend(taclet);
	    } 
	}
	
       	return result;
    }


    /** collects all Taclet applications at the given position of the specified
     * taclet
     * @param pos the PosInOccurrence describing the position
     * @param taclet the Taclet
     * @return a list of all found rule applications of the given rule at
     * position pos  
     */
    protected SetOfTacletApp getTacletApplications(Goal goal, Taclet taclet, 
                                                   PosInOccurrence pos) {
	SetOfTacletApp result = SetAsListOfTacletApp.EMPTY_SET;  
        ListOfTacletApp fittingApps = 
           goal.ruleAppIndex().getTacletAppAt(activeHeuristics(), 
                                              interactive, pos, 
			                      mediator.getJavaInfo());
	IteratorOfTacletApp it = fittingApps.iterator();
	// filter fitting applications
	while (it.hasNext()) {
	    TacletApp app = it.next();
	    if ((app.rule()).equals(taclet) ){//&& pos.equals(app.posInOccurrence())) {
		if (taclet.name().toString().equals("insert_eq")) {
		    TacletApp x=result.iterator().next();
		    if (x!=null) {
			System.out.println("XXX"+app.equals(x)+"  "+app.instantiations().equals(x.instantiations())+"    "+app.posInOccurrence().equals(x.posInOccurrence()));
					   
					   }
		}
		result = result.add(app);
	    }
	}
	return result;
    }
    
    /** listener to KeYSelection Events in order to be informed if the
     * current proof changed
     */
    private class InteractiveProverKeYSelectionListener 
	implements KeYSelectionListener {

	/** focused node has changed */
	public void selectedNodeChanged(KeYSelectionEvent e) {
	    focusedGoal = e.getSource().getSelectedGoal();
	}

	/** the selected proof has changed (e.g. a new proof has been
	 * loaded) */ 
	public void selectedProofChanged(KeYSelectionEvent e) {
	    Debug.out("InteractiveProver: initialize with new proof");
	    setProof(e.getSource().getSelectedProof());
	    focusedGoal = e.getSource().getSelectedGoal();
	}

    }
}
