// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.gui;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.SLListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.IteratorOfNode;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ProofTreeListener;
import de.uka.ilkd.key.proof.ProofTreeAdapter;
import de.uka.ilkd.key.proof.ProofTreeEvent;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.util.Debug;

class KeYSelectionModel {

    /** the proof to listen to */
    private Proof proof;
    /** if true the selectedGoal field below can be trusted */
    private boolean goalIsValid;
    /** is the selected node a goal */
    private Goal selectedGoal;
    /** the current displayed node */
    private Node selectedNode;
    /** the listeners to this model */
    private List listenerList;
    /** cached selected node event */
    private KeYSelectionEvent selectionEvent = 
	new KeYSelectionEvent(this);
    
    
    KeYSelectionModel() {
	listenerList = Collections.synchronizedList(new ArrayList(5));
	goalIsValid = false;
    }

    /** sets the selected proof 
     * @param p the Proof that is selected
     */
    public void setSelectedProof(Proof p) {
	goalIsValid = false;
	proof = p;
	Goal g = proof.openGoals().iterator().next();	    
	if ( g == null) {
	    selectedNode = proof.root().leavesIterator().next();
	} else {
	    goalIsValid = true;
	    selectedNode = g.node();
	    selectedGoal = g;
	}	

	fireSelectedProofChanged();
    }

    /** returns the proof that is selected by the user
     * @return  the proof that is selected by the user
     */ 
    public Proof getSelectedProof() {
	return proof;
    }

    /** sets the node that is focused by the user
     * @param n the selected node
     */
    public void setSelectedNode(Node n) {
	goalIsValid = false;
	selectedNode = n;
	fireSelectedNodeChanged();
    }

    /** sets the node that is focused by the user
     * @param g the Goal that contains the selected node
     */
    public void setSelectedGoal(Goal g) {	
	selectedGoal = g;
	selectedNode = g.node();
	goalIsValid = true;
	fireSelectedNodeChanged();
    }

    /** returns the node that is selected by the user
     * @return  the node that is selected by the user
     */ 
    public Node getSelectedNode() {
	return selectedNode;
    }

    /** returns the goal the selected node belongs to, null if it is
     * an inner node
     * @return  the goal the selected node belongs to, null if it is
     * an inner node
     */ 
    public Goal getSelectedGoal() {
	if (proof==null) {
	    throw new IllegalStateException("No proof loaded.");
	}
	if (!goalIsValid) {
	    selectedGoal = proof.getGoal(selectedNode);
	    goalIsValid = true;
	} 
	return selectedGoal;
    }

    /** returns true iff the selected node is a goal
     * @return true iff the selected node is a goal
     */
    public boolean isGoal() {
	if (!goalIsValid) {
	    return (getSelectedGoal() == null);
	}	
	return selectedGoal == null;
    }

   
    /** selectes the first goal in the goal list of proof if available
     * if not it selectes a leaf of the proof tree
     */
    public void defaultSelection() {		
	Goal g = null;
	if (selectedNode != null) {
	    IteratorOfNode nodeIt = selectedNode.leavesIterator();	
	    while (nodeIt.hasNext()) {
		g = proof.getGoal(nodeIt.next());
		if (g != null) {
		    break;
		}
	    }
	}
	if (g == null && proof.openGoals() != SLListOfGoal.EMPTY_LIST ) {	       	    
	    g = proof.openGoals().iterator().next();
	}
	if (g != null) {
	    setSelectedGoal(g);
	} else {
	    setSelectedNode(proof.root().leavesIterator().next());	
	}
    }

    public void addKeYSelectionListener(KeYSelectionListener listener) {
	synchronized(listenerList) {
	    listenerList.add(listener);	    
	}
    }

    public void removeKeYSelectionListener(KeYSelectionListener listener) {
	synchronized(listenerList) {
	    listenerList.remove(listener);	    
	}
    }		

    public synchronized void fireSelectedNodeChanged() {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((KeYSelectionListener)it.next()).
		    selectedNodeChanged(selectionEvent);
	    }
	}
    }

    public synchronized void fireSelectedProofChanged() {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((KeYSelectionListener)it.next()).
		    selectedProofChanged(selectionEvent);
	    }
	}
    }    
    
}
