// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import java.util.HashMap;
import java.util.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;

/** Stores the mapping from operators to {@link Notation}s.  Each 
 * {@link Notation} represents the concrete syntax for some 
 * {@link de.uka.ilkd.key.logic.op.Operator}.  The {@link SequentPrinter}
 * asks the NotationInfo to find out which Notation to use for a given term.
 * 
 * The Notation associated with an operator might change.  New Notations can
 * be added and abbreviations for epsilon terms may be introduced.
 */
public class NotationInfo {
    
    /** This maps operators and classes of operators to {@link
     * Notation}s.  The idea is that we first look wether the operator has
     * a Notation registered.  Otherwise, we see if there is one for the
     * <em>class</em> of the operator.
     */
    private HashMap tbl;
    
    /** This maps epsilon terms to {@link EpsilonPrintInfo}s.  The
     * latter are also {@link Notation}s, which might print an
     * abbreviation instead of the full syntax.
     */
    private HashMap epsAbbrevs;

    /** Create a new NotationInfo. */
    public NotationInfo() {
	createDefaultNotationTable();
    }

    /** Set all notations back to default. */
    public void setBackToDefault() {
	createDefaultNotationTable();
    }
    
    /** Register the standard set of notations.  This means no epsilon 
     * abbreviations, and a set of Notations for the built-in operators
     * which corresponds to the parser syntax. 
     */
    private void createDefaultNotationTable() {
	epsAbbrevs = new HashMap();
	tbl=new HashMap();
	createDefaultOpNotation();
	createDefaultTermSymbolNotation();
    }

    /**
     * Registers notations for the built-in operators.  The priorities
     * and associativities correspond to the parser syntax.  
     */
    private void createDefaultOpNotation() {
	tbl.put(Op.TRUE ,new Notation.Constant("true",80));
	tbl.put(Op.FALSE,new Notation.Constant("false",80));
	tbl.put(Op.NOT,new Notation.Prefix("!" ,60,60));
	tbl.put(Op.AND,new Notation.Infix("&"  ,50,60,50));
	tbl.put(Op.OR, new Notation.Infix("|"  ,40,50,40));
	tbl.put(Op.IMP,new Notation.Infix("->" ,30,40,30));
	tbl.put(Op.EQV,new Notation.Infix("<->",20,30,20));

	tbl.put(Op.EPSILON,new Notation.Epsilon(this,"eps",60,60));
	tbl.put(Op.ALL,new Notation.Quantifier("all",60,60));
	tbl.put(Op.EX, new Notation.Quantifier("ex", 60,60));
	tbl.put(Op.DIA,new Notation.Modality("<",">",60,60));
	tbl.put(Op.BOX,new Notation.Modality("[","]",60,60));

	tbl.put(Op.SUBST,new Notation.Subst());
    }    
    

    /** 
     * Register notations for standard classes of operators.  This
     * includes Function operators, all kinds of variables, etc.
     */
    private void createDefaultTermSymbolNotation() {
	tbl.put(Function.class, new Notation.Function());
	tbl.put(CondTermOp.class, new Notation.ConditionalTerm());
	tbl.put(LogicVariable.class, new Notation.Variable());
	tbl.put(SchemaVariable.class, new Notation.Variable());
	tbl.put(Metavariable.class, new Notation.Variable());
	tbl.put(ProgramVariable.class, new Notation.Variable());
	tbl.put(ProgramMethod.class, new Notation.ProgramMethod());
	tbl.put(Equality.class, new Notation.Infix("=",70,80,80)); 
	tbl.put(UpdateOp.class, new Notation.Update());
	tbl.put(AttributeOp.class, new Notation.Attribute(80,80));
    }

    /**
     * Get the notation for a particular epsilon term.  Epsilon terms
     * are identified via object identity.  This is usually no problem,
     * due to sharing.  It might be confusing if two syntactically identical
     * epsilon terms appear during a proof.  If no Notation exists,
     * one is created.
     */
    public Notation getEpsilonNotation(Term eps) {
	EpsilonPrintInfo info = (EpsilonPrintInfo)(epsAbbrevs.get(eps));
	if (info!=null) {
	    return info;
	}
	info = new EpsilonPrintInfo(this, eps);
	epsAbbrevs.put(eps, info);
	return info;
    }
   
    /**
     * Get the {@link EpsilonPrintInfo} for an epsilon term.  If none
     * is registered, return <code>null</code>.
     */
    EpsilonPrintInfo getEpsilonPrintInfo(Term eps) {
	return (EpsilonPrintInfo)(epsAbbrevs.get(eps));
    }

    /** Returns whether a certain abbreviation is already used.  This
     * currently uses the {@link EpsilonPrintInfo#equals(Object)}
     * method which currently compares abbreviations, which is extremely
     * bad design.  */
    boolean containsAbbrev(EpsilonPrintInfo info) {
	return epsAbbrevs.containsValue(info);	
    }

    /** Get the Notation for a given Operator.  
     * If no notation is registered, a Function notation is returned.
     */
    public Notation getNotation(Operator op) {
	if (tbl.get(op)!=null) {	    
	    return (Notation)tbl.get(op);
	} else if ((Notation)tbl.get(op.getClass()) != null) {
	    return (Notation)tbl.get(op.getClass());
	}
	return new Notation.Function(); 
    }

    /** Returns the priority of an operator.
     * @return the priority of operator op
     * @param op the operator we want to know the priority of
     */
    public int getPriority(Operator op){
	return getNotation(op).getPriority(op);
    }

    /** Registers an infix notation for a given operator 
     * @param op    the operator
     * @param token the string representing the operator
     */
    public void createInfixNotation(Operator op, String token) {
	tbl.put(op, new Notation.Infix(token,90,100,100));
    }

    /** Registers a prefix notation for a given operator 
     * @param op    the operator
     * @param token the string representing the operator
     */
    public void createPrefixNotation(Operator op, String token) {
	tbl.put(op, new Notation.Prefix(token, 110, 0));
    }

    /** Registers a number literal notation for a given operator.
     * This is done for the `Z' operator which marks number literals.
     * A term <code>Z(3(2(#)))</code> gets printed simply as
     * <code>23</code>.
     * @param op the operator */
    public void createNumLitNotation(Operator op) {
	tbl.put(op, new Notation.NumLiteral());
    }


}
