// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.table.TableColumn;
import javax.swing.event.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.util.Debug;
import java.awt.dnd.*;
import java.awt.datatransfer.*;

public class TacletMatchCompletionDialog extends ApplyTacletDialog {

    // the table showing the instantiations
    private DataTable[] dataTable;

    // the current chosen model
    private int current;

    // the gui component used to display the different instantiation
    // alternatives
    private JTabbedPane alternatives;
    
    /** the goal the application of the rule has to be performed */
    private Goal goal;
 
    public TacletMatchCompletionDialog(ApplyTacletDialogModel[] model,
				     Goal goal, KeYMediator mediator) { 
	super(model, mediator);	
	this.goal = goal;

	current = 0;
	dataTable = new DataTable[model.length];

	setStatus(model[current].getStatusString());
	for (int i = 0; i < model.length; i++) {
	    model[i].prepareUnmatchedInstantiation();
	}

	// layout dialog
	layoutDialog();
	pack();
	// set at the middle of the main frame
	//setLocation(dialogPosition());
	setSize(800,(int)getSize().getHeight());
	setLocation(50,50);
	setVisible(true);
    }

    public TacletMatchCompletionDialog(ApplyTacletDialogModel model,
				     Goal goal,
				     KeYMediator mediator) { 
	this(new ApplyTacletDialogModel[]{model}, goal, mediator);
    }

    /** calculates the position where the dialog has to be displayed.
     * It is tried to the dialog in the mid of the main frame
     * @return the point describing the position of the top-left
     * corner of the dialog 
     */
    private Point dialogPosition() {
	Point mainLocation = mediator().mainFrame().getLocationOnScreen();

	// calculate relative position to the main frame
	Dimension mainFrame_size = mediator().mainFrame().getSize();
	Dimension dialog_size = getSize();

	int mainFrame_middle_x = (int) mainFrame_size.getWidth() / 2;
	int mainFrame_middle_y = (int) mainFrame_size.getHeight() / 2;
	int x = 0;
	int y = 0;

	if (dialog_size.getWidth() < mainFrame_size.getWidth()) {
	    x = (int) (mainFrame_middle_x - (dialog_size.getWidth() / 2)); 
	} else {
	    x = mainFrame_middle_x / 2;
	}

	if (dialog_size.getHeight() < mainFrame_size.getHeight()) {
	    y = (int) (mainFrame_middle_y - (dialog_size.getHeight() / 2)); 
	} else {
	    y = mainFrame_middle_y / 2;
	}	

	// calculate absolut position
	x += mainLocation.getX();
	y += mainLocation.getY();
	
	return new Point(x, y);
    }
    
    private void layoutDialog() {
	
	JPanel instPanel=new JPanel();
	instPanel.setLayout(new BoxLayout(instPanel, BoxLayout.Y_AXIS));
	
	instPanel.add(createTacletPanel());	     
	instPanel.add(createInfoPanel());
	instPanel.add(createStatusPanel());
	instPanel.add(createButtonPanel(new ButtonListener(this)));

	JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
                                   createTacletDisplay(), instPanel);

	getContentPane().add(splitPane);
	// add button listener
	updateDataModel();
    }
        
    private JPanel createTacletPanel() {
	// the tabbedPane contains different possible instantiations of 
	// the applied Taclet
	JPanel panel = new JPanel(new GridLayout(1,1));	
	panel.setBorder(new TitledBorder("Variable Instantiations"));
	
	alternatives = new JTabbedPane();	
	// some layout stuff 
	EmptyBorder indents = new EmptyBorder(5,5,5,5);	
	
	for (int i = 0; i < model.length; i++) {
	    JPanel tabContent = new JPanel();
	    tabContent.setLayout(new BoxLayout(tabContent, BoxLayout.Y_AXIS));

	    JPanel instPanel = createInstantiationDisplay(i);
	    instPanel.setBorder(indents);
	    tabContent.add(instPanel);

	    if (model[i].application().taclet().
		ifseq() != Sequent.EMPTY_SEQUENT) {

		TacletIfSelectionDialog ifSelection = 
		    new TacletIfSelectionDialog(model[i]);		
		dataTable[i].setIfSelectionPanel(ifSelection);
		tabContent.add(ifSelection);
	    } 
	    alternatives.addTab("Alt "+i, null, tabContent, 
			  "Instantiations Alternatives");
	}

	panel.add(alternatives);
	return panel;
    }

    /** returns the current selected model 
     * @return the current selected model 
     */
    protected int current() {
	return alternatives.getSelectedIndex();
    }

    protected void pushAllInputToModel() {
	pushAllInputToModel(current());
    }

    protected void pushAllInputToModel(int i) {
	if (dataTable[i].hasIfSelectionPanel()) {
	    dataTable[i].getIfSelectionPanel().pushAllInputToModel();
	}
	if (dataTable[i].isEditing()) {
	    dataTable[i].getCellEditor().stopCellEditing();
	}
    }

    private JPanel createInstantiationDisplay(int i) {
	JPanel panel = new JPanel(new BorderLayout());
	// show instantiation
	dataTable[i] = new DataTable(this, i);
	JScrollPane tablePane = new JScrollPane(dataTable[i]);

	int tableSize_x = dataTable[i].getTotalColumnWidth();
	// maximum of 5 row may be visible 
	int visible_rows = dataTable[i].getRowCount() > 8 ? 8 : 
	    dataTable[i].getRowCount();
	int tableSize_y = (int)(
	    (visible_rows + 1) * // number of visible rows + header
	    (dataTable[i].getRowHeight()*1.3 + // height of one row
	     (dataTable[i].getRowMargin()*1.3))); // space between rows 
	                                       // + buffer space
	Dimension tableDim = new Dimension(tableSize_x, tableSize_y);

	tablePane.setMinimumSize(tableDim);
	tablePane.setPreferredSize(tableDim);
	tablePane.setMaximumSize(tableDim);
	panel.add(tablePane, 
		  BorderLayout.CENTER);
	return panel;
    }

    private void setColumnName(int model, int col, String name) {
	dataTable[model].
	    getColumn(dataTable[model].getColumnName(col)).
	    setHeaderValue(name);
    }

    /** shows next instantiation suggestion */
    private void updateDataModel() {	
	for (int i = 0; i < model.length; i++) {
	    if (model[i] != null) {
		dataTable[i].setModel(model[i].tableModel());		
		//sets column names
		setColumnName(i, 0, "Variable");
		setColumnName(i, 1, "Instantiation");	    
	    }
	}
    }

    class ButtonListener implements ActionListener {
    
       private JDialog parent = null;
       
       public ButtonListener(JDialog d) {
          parent = d;
       }

	public void actionPerformed(ActionEvent e) {
	    if (e.getSource() == cancelButton) {
		closeDialog();
	    } else if (e.getSource() == applyButton) {		      
		pushAllInputToModel();
		try {
		    TacletApp app = model[current()].createTacletApp(); 		    
		    if (app == null) {
			JOptionPane.showMessageDialog
			    (TacletMatchCompletionDialog.this,
			     "Could not apply rule",
			     "Rule Application Failure",
			     JOptionPane.ERROR_MESSAGE);   
			return ;
		    }
		    mediator().apply(app, goal);
		} catch (SVInstantiationException ie) {
		    JOptionPane.showMessageDialog
			(TacletMatchCompletionDialog.this,
			 ie.getMessage(),
			 "Instantiation Error",
			 JOptionPane.ERROR_MESSAGE);   
		    // select table cell where the error occured
		    dataTable[current()].setEditingRow(ie.getRow());
		    dataTable[current()].setEditingColumn(1);
		    return ;		    
		} catch (Exception ex) {		    
		    ErrorMessages.showBugMessage
			(TacletMatchCompletionDialog.this,"Could not apply rule",ex); 
		    return ;
		} 
		
		closeDialog();
	    }
	}

	private void closeDialog() {
	    setVisible(false);
	    dispose();
            synchronized(parent) { parent.notifyAll(); }
	}

    }

    static private class DataTable extends JTable 
	implements ModelChangeListener {
	
	/** the number of the model the data table belongs to */
	private int modelNr;
	/** the enclosing dialog */	
	private TacletMatchCompletionDialog owner;
	/** the TacletIfSelectionPanel that shows the different possible
	 * instantiations of the if-sequent or a manual entered
	 * instantiation. The value is null if and only if
	 * the taclet has no if-sequent
	 */
	private TacletIfSelectionDialog ifSelectionPanel;

	private DataTable(TacletMatchCompletionDialog owner,
			  int modelNr) {

	    super(owner.model[modelNr].tableModel());
	    this.modelNr = modelNr;
	    this.owner = owner;
	    owner.model[modelNr].addModelChangeListener(this);


	    // And now the Drag'n'drop stuff ...
	    DropTarget aDropTarget = 
		new DropTarget
		    (this, 
		     new DropTargetListener(){
			     
			     /**
			      * is invoked when you are dragging over the DropSite
			      * 
			      */
			     
			     public void dragEnter (DropTargetDragEvent event) {
				 // event.acceptDrag (DnDConstants.ACTION_MOVE);
			     }
			     
			     /**
			      * is invoked when you are exit the DropSite without dropping
			      *
			      */
			     
			     public void dragExit (DropTargetEvent event) {
				 
			     }
			     
			     /**
			      * is invoked when a drag operation is going on
			      * 
			      */
			     
			     public void dragOver (DropTargetDragEvent event) {
			     }
				       
			     /**
			      * a drop has occurred
			      * 
			      */			     
			     public void drop (DropTargetDropEvent event) {
				 String droppedString;
				 
				 Point dropLocation = event.getLocation();
				 int row = DataTable.this.rowAtPoint( dropLocation );
				 int column = DataTable.this.columnAtPoint( dropLocation );

				 if ((row != -1) && (column == 1)){
				     // The point lies within the table and within the instantiation
				     // column ...

				     try {
					 Transferable transferable = event.getTransferable();
						   
					 // we accept only Strings      
					 if (transferable.isDataFlavorSupported (DataFlavor.stringFlavor)){
						       
					     event.acceptDrop(DnDConstants.ACTION_MOVE);
					     droppedString = (String)transferable.getTransferData ( DataFlavor.stringFlavor);
					     // now set the new entry in the table ...

					     if(droppedString != null){
						 String s = cleanUp(droppedString);
							   
						 DataTable.this.setValueAt(s, row, column);
						 DataTable.this.repaint();
					     }
					     event.getDropTargetContext().dropComplete(true);
					 } 
					 else{
					     event.rejectDrop();
					 }
				     }
				     catch (IOException exception) {
					 exception.printStackTrace();
					 event.rejectDrop();
				     } 
				     catch (UnsupportedFlavorException ufException ) {
					 ufException.printStackTrace();						   
					 event.rejectDrop();
				     }
				 } else {
				     event.rejectDrop();
				 }
			     }
					    
			     /**
			      * is invoked if the use modifies the current drop gesture
			      * 
			      */
			     
			     
			     public void dropActionChanged ( DropTargetDragEvent event ) {
			     }
				       
			     /**
			      * Replaces all newline symbols by spaces ...
			      */
			     public String cleanUp(String s){
				 String result;
				 result = s.replace('\n',' ');
				 return result;
			     }

				       
			 });   
	
 
	    this.setDropTarget(aDropTarget);
	    
	} // end constructor
        
        
        /** Provide sane single-click editing in table */
        public javax.swing.table.TableCellEditor getCellEditor(int row, int col) {
           javax.swing.DefaultCellEditor ed =
              (javax.swing.DefaultCellEditor)super.getCellEditor(row, col);
           ed.setClickCountToStart(1);
           return ed;
        }
        

	/** sets the if selection panel */
	private void setIfSelectionPanel
	    (TacletIfSelectionDialog ifSelectionPanel) {
	    this.ifSelectionPanel = ifSelectionPanel;	    
	}

	/** returns the if selection panel
	 * @return the if selection panel, null if not available
	 */
	private TacletIfSelectionDialog getIfSelectionPanel() {
	    return ifSelectionPanel;	    
	}

	/** returns true the model has a non empty if sequent and
	 * the ifSelectionPanel has been created and set. So that the
	 * method getIfSelectionPanel will not return null
	 * @returns true iff getIfSelectionPanel does not return null
	 */
	private boolean hasIfSelectionPanel() {
	    return getIfSelectionPanel() != null;	    
	}

	public void modelChanged(ModelEvent me) {
	    if (me.getSource() instanceof ApplyTacletDialogModel) {
		setModel(((ApplyTacletDialogModel)me.getSource()).tableModel());
		repaint();
	    }
	}

	public int getTotalColumnWidth() {
	    return getColumnModel().getTotalColumnWidth();
	}
		
	public void editingStopped(ChangeEvent e) {	    
	    if (modelNr == owner.current()) {
		super.editingStopped(e);
		owner.pushAllInputToModel(modelNr);
		if (owner.checkAfterEachInput()) {
		    owner.setStatus(owner.model[modelNr].
				    getStatusString());
		}
	    }
	}	
    }


}
