// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.util.Debug;


/** Abstract constraint interface for constraints offering unification
 * of terms and joins. There are no public constructors to build up a
 * new Constraint use the BOTTOM constraint (final static class
 * variable) and add the needed constraints if a constraint would not
 * be satisfiable (cycles, unification failed) the Constraint TOP is
 * returned. TOP is as well as BOTTOM a final static class variable
 * (means usage of singleton pattern for Constraints BOTTOM and TOP).
 * Unsatisfiable constrains should only be representated by the TOP
 * element.
 */

public interface Constraint extends java.io.Serializable {

    /** unsatisfiable Constraint */
    public static final Constraint TOP     = new Top ();
    /** standard constraint class implementing the offered functionality */ 
    public static final Constraint BOTTOM  = new EqualityConstraint ();


    /** returns true if Bottom
     * @return true if Bottom 
     */
    public boolean isBottom();


    /** a constraint being instance of this class is satisfiable. If a
     * method realizes that an unsatisfiable Constraint would be built
     * because of failed unification, cycle or s.th. similar it
     * returns the singleton TOP being instance of the subclass Top
     * @return true always 
     */
    public boolean isSatisfiable();


    /** executes unification for terms t1 and t2
     * PRECONDITION: sorts of the given terms are equal 
     * @param t1 Term to be unified 
     * @param t2 term to be unified
     * @return TOP if not possible, else a new constraint with after
     * unification of t1 and t2
     */
    public Constraint unify(Term t1, Term t2);

        
    /** executes unification for terms t1 and t2. 
     * PRECONDITION: sorts of the given terms are equal. 
     * @param t1 Term to be unfied 
     * @param t1 Term to be unfied
     * @param subsumed  true iff the new constraint is subsumed by
     * this one
     * @return TOP if not possible, else a new constraint with after
     * unification of t1 and t2
     */
    public Constraint unify(Term t1, Term t2, BooleanContainer subsumed);


    /** checks equality of constraints
     */
    public boolean equals(Object obj);


    /** joins the given constraint with this constraint and returns
     * the joint new constraint.  Every implementing class should
     * handle the cases co == TOP and ( co instanceof EqualityConstraint ).
     * @param co Constraint to be joined with this one
     * @return the joined constraint 
     */	
    public Constraint join(Constraint co);
    

    /** joins constraint co with this constraint and returns the joint
     * new constraint. The BooleanContainer is used to wrap a second
     * return value and indicates a subsumption of co by this
     * constraint. Every implementing class should handle the cases co
     * == TOP and ( co instanceof EqualityConstraint ).
     * @param co Constraint to be joined with this one
     * @param subsuemd the BooleanContainers value set true, iff this
     * constraint subsumes co 
     * @return the joined constraint     
     */
    public Constraint join(Constraint co, BooleanContainer subsumed);
    

    /** @return String representation of the constraint */
    public String toString();


    /** Constraint class for representating the TOP (unsatisfiable) constraint. */
    static class Top implements Constraint {
	
	/** creation of TOP */
	public Top() {}
	
	/** used for serializing. As we have a unique Top we will
	 * replace the deserialized double by the unique original
	 */
	private Object readResolve() 
	    throws java.io.ObjectStreamException {
	    System.err.println("Replaced TOP");	
	    return Constraint.TOP;
	}


	/** is an unsatisfiable Constraint satisfiable? NO.
	 * @return always false
	 */
	public boolean isSatisfiable() {
	    return false;
	}
	
	/** adding new constraints to an unsatisfiable constraint
	 * results in an unsatisfiable constraint so this one is returned
	 * 
	 * @return always this
	 */
	public Constraint unify(Term t1, Term t2) {
	    return this;
	}

	public Constraint unify(Term t1, Term t2, BooleanContainer subsumed) {
	    subsumed.setVal ( true );
	    return this;
	}

	public boolean equals(Object obj) {
	    return ( obj instanceof Top );
	}

	/** joint of Top and co is Top
	 * @return this 
	 */
	public Constraint join(Constraint co) { 
	    return this; 
	}

	/** joint of Top and co is Top and Top subsumes every constraint
	 * @return this 
	 */    
	public Constraint join(Constraint co, BooleanContainer c) {
	    c.setVal ( true );
	    return this;
	}
	
	/** returns true if Bottom
	 * @return true if Bottom 
	 */
	public boolean isBottom() {
	    return false;
	}
	
	/** @return String representing the TOP constraint 
	 */
	public String toString() {
	    return "TOP";
	}
	
    }

}
