// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import java.util.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.rule.TacletVariableCollector;
import de.uka.ilkd.key.rule.SyntacticalReplaceVisitor;
import de.uka.ilkd.key.rule.inst.SVInstantiations;

/** This class represents the succedent or antecendent part of a
 * sequent. It is more or less a list without duplicates. This is
 * ensured using method removeRedundancy. In future versions it can be
 * enhanced to do other simplifications. A sequent and so a
 * semisequent has to be immutable.
 * 
 */
public class Semisequent implements java.io.Serializable {

    /** the empty semisequent (using singleton pattern) */
    public final static Semisequent EMPTY_SEMISEQUENT = new Empty();
    /** list with the Constraintformulae of the Semisequent */
    private ListOfConstrainedFormula
	seqList=SLListOfConstrainedFormula.EMPTY_LIST; 
 
    /** used by inner class Empty*/
    private Semisequent() {
    }

    /** creates a new Semisequent with one element 
     * @param conForm ConstrainedFormula being the unique element
     */ 
    private Semisequent(ConstrainedFormula conForm) {
	this.seqList = seqList.prepend(conForm);
    }
    

    /** creates a new Semisequent with the Semisequent elements in
     * seqList */ 
    private Semisequent(ListOfConstrainedFormula seqList) {
	this.seqList = seqList;
    }

    private Object readResolve() 
	throws java.io.ObjectStreamException {
	if (seqList == SLListOfConstrainedFormula.EMPTY_LIST) {
	    return Semisequent.EMPTY_SEMISEQUENT;
	}
	return this;
    }


    /** inserts an element at the specified index  
     * @param idx int encoding the place the element has to be put	
     * @param conForm ConstrainedFormula to be inserted
     * @return new semisequent with conForm at place idx
     */    
    private Semisequent insertNoCheck(int idx, 
				    ConstrainedFormula conForm) {
	int index=0;
	ListOfConstrainedFormula newList = seqList;  
	ListOfConstrainedFormula queue =
	    SLListOfConstrainedFormula.EMPTY_LIST;  

	// go to index idx
	while(index<idx) {
	    index++;
	    queue=queue.append(newList.head());
	    newList=newList.tail();	    	    
	}
	// insert new element
	newList=newList.prepend(conForm);
	newList=newList.prepend(queue);
	return new Semisequent(newList).checked();
    }


    /** inserts an element at a specified index performing redundancy
     * checks, this may result in returning same semisequent if
     * inserting would create redundancies 
     * @param idx int encoding the place the element has to be put	
     * @param conForm ConstrainedFormula to be inserted
     * @return new semisequent with conForm at place idx
     */    
    public Semisequent insert(int idx, ConstrainedFormula conForm) {	
	return removeRedundance(idx, conForm);
    }

    /** inserts element at index 0 performing redundancy
     * checks, this may result in returning same semisequent if
     * inserting would create redundancies 
     * @param conForm ConstrainedFormula to be inserted
     * @return new semisequent with conForm at place idx
     */
    public Semisequent insertFirst(ConstrainedFormula conForm) {
	return insert(0,conForm);
    }
    

    /** inserts element at the end of the semisequent performing
     * redundancy checks, this may result in returning same
     * semisequent if inserting would create redundancies 
     * @param conForm ConstrainedFormula to be inserted
     * @return new semisequent with conForm at place idx
     */	
    public Semisequent insertLast(ConstrainedFormula conForm) {
	return insert(size(),conForm);
    }


    /** .inserts new ConstrainedFormula at index idx and removes
     * duplicates, perform simplifications etc.
     * @return new Semisequent with conForm at index idx and removed
     * redundancies 
     */
    private Semisequent removeRedundance(int idx, ConstrainedFormula conForm) {
	if (containsEqual(conForm)) { // no duplicates
	    return this;
	} 
	return insertNoCheck(idx,conForm);
    }


    /** replaces the element at place idx with conForm
     * @param idx an int specifying the index of the element that
     * has to be replaced 
     * @param conForm the ConstrainedFormula replacing the old
     * element at index idx
     */
    public Semisequent replace(int idx, ConstrainedFormula conForm) {	
	return remove(idx).insert(idx,conForm);
    }

    /** @return int counting the elements of this semisequent */
    public int size() {
	return seqList.size();
    }

    /** removes an element 
     * @param idx int being the index of the element that has to
     * be removed
     * @return Semisequent without the element at idx 
     */
    public Semisequent remove(int idx) {	       
	ListOfConstrainedFormula newList=seqList;  
	ListOfConstrainedFormula queue=
	    SLListOfConstrainedFormula.EMPTY_LIST;  
	int index=0;
	
	if (idx<0 || idx>=size()) {
	    return this;
	}

	while (index<idx) {// go to idx 
	    queue=queue.append(newList.head());// stores all former 
	                                       //list elements 	    
	    newList=newList.tail();
	    index++;
	    
	}
	// remove the element that is at head of newList	
	newList=newList.tail();
	newList=newList.prepend(queue);

	return new Semisequent(newList).checked();
    }

    /**
     * returns the EMPTY_SEMISEQUENT if the list of ConstrainedFormulae
     * constituting the Semisequent is an empty list or this object
     * otherwise. Return the result of this method in all methods that can
     * remove elements from the semisequent, to avoid empty semisequents that
     * are not the EMPTY_SEMISEQUENT.
     * @return a non empty semisequent or the constant EMPTY_SEMISEQUENT
     */
    private Semisequent checked() {
	return (seqList!=SLListOfConstrainedFormula.EMPTY_LIST) ? 
	    this : EMPTY_SEMISEQUENT;
    }

    /**
     * returns index of a ConstrainedFormula. An identity check ('==')
     * is used when searching for the ConstrainedFormula, NOT equals on
     * ConstrainedFormula, because a ConstrainedFormulae identifies the origin
     * of a formula.
     * @param conForm the ConstrainedFormula the index want to be
     * determined 
     * @return index of conForm (-1 if not found)
     */
    public int indexOf(ConstrainedFormula conForm) {
	ListOfConstrainedFormula searchList=seqList;  
	int index=0;
	while (searchList!=SLListOfConstrainedFormula.EMPTY_LIST)
	    { 
		if (searchList.head()==conForm) {
		    return index;
		} else {
		    index++;
		}
		searchList=searchList.tail();
	    } 
	return -1;
    }

    /** gets the element at a specific index
     * @param idx int representing the index of the element we
     * want to have
     * @return ConstrainedFormula found at index idx
     */
    public ConstrainedFormula get(int idx) {
	ListOfConstrainedFormula searchList=seqList;  
	for (int i=0;i<idx;i++) {
	    searchList=searchList.tail();
	}
	return searchList.head();
    }

    /** @return the first ConstrainedFormula of this Semisequent */
    public ConstrainedFormula getFirst() {
	return seqList.head();
    }

    /** checks if a ConstrainedFormula is in this Semisequent
     * (identity check)
     * @param conForm the ConstraintForumla to look for
     * @return true iff. conForm has been found in this
     * Semisequent 
     */
    public boolean contains(ConstrainedFormula conForm) {
	return (indexOf(conForm)!=-1);
    }
    
    /** checks if a ConstrainedFormula is in this Semisequent
     * (equality check)
     * @param conForm the ConstraintForumla to look for
     * @return true iff. conForm has been found in this
     * Semisequent 
     */
    public boolean containsEqual(ConstrainedFormula conForm) {
	ListOfConstrainedFormula searchList=seqList;  
	while (searchList!=SLListOfConstrainedFormula.EMPTY_LIST)
	    { 
		if (searchList.head().equals(conForm)) {
		    return true;
		} 
		searchList=searchList.tail();
	    }
	return false;
    }

    /** looks for variable var in Term term
     * @param var the SchemaVariable to be looked for
     * @param term the Term to be looked through
     * @return true iff variable has been found in the term 
     */    
    private boolean lookupInTerm(SchemaVariable var, Term term) {
	TacletVariableCollector vcv=new TacletVariableCollector();
	term.execPostOrder(vcv);
	return vcv.contains(var);
    }
    
    /** looks for variable var in the Semisequent
     * @param var the SchemaVariable to be looked for
     * @return true iff variable has been found 
     */
    public boolean lookup(SchemaVariable var) {
	IteratorOfConstrainedFormula it=iterator();
	while (it.hasNext()) {
	    if (lookupInTerm(var, it.next().formula())) {
		return true;
	    }	    
	}
	return false;
    }



    /**
     * executes the syntactical Replacement defined by the given map on all
     * formulas of the Semisequent and returns a new Semisequent where the
     * replacement is done. The Constraints of the new ConstrainedFormulas are
     * set to Constraint.BOTTOM.
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the SVInstantiations containing the replacements
     * @return a new Semisequent with the replaced terms 
     */
    protected Semisequent syntacticalReplace(JavaInfo javaInfo, SVInstantiations svInst) {
	IteratorOfConstrainedFormula it = iterator();	
	ListOfConstrainedFormula replaced = SLListOfConstrainedFormula.EMPTY_LIST;
	while (it.hasNext()) {
	    SyntacticalReplaceVisitor srv = new SyntacticalReplaceVisitor(javaInfo, svInst);
	    it.next().formula().execPostOrder(srv);
	    replaced=replaced.append(new ConstrainedFormula(srv.getTerm(), 
							   Constraint.BOTTOM));
	}	
	return new Semisequent(replaced).checked();
    }



    /** returns iterator about the elements of the sequent 
     * @return IteratorOfConstrainedFormula
     */

    public IteratorOfConstrainedFormula iterator() {
	return seqList.iterator();
    }

    public boolean equals(Object o) {
 	if ( ! ( o instanceof Semisequent ) ) 
	    return false;
	Semisequent o1 = (Semisequent) o;
	return seqList.equals(o1.seqList);
    }


    /** @return String representation of this Semisequent */
    public String toString() {
	return seqList.toString();
    }


    
    // inner class used to represent an empty semisequent 
    private static class Empty extends Semisequent{

	Empty() {
	    super();
	}

	private Object readResolve() 
	    throws java.io.ObjectStreamException {
	    return Semisequent.EMPTY_SEMISEQUENT;
	}

	
	/** inserts the element always at index 0 ignores the first
	 * argument 
	 * @param idx int encoding the place the element has to be put	
	 * @param conForm ConstrainedFormula to be inserted
	 * @return new semisequent with conForm at place idx
	 */
	public Semisequent insert(int idx, ConstrainedFormula conForm) {	
	    return insertFirst(conForm);
	}

	/** inserts the element at index 0
	 * @param conForm ConstrainedFormula to be inserted
	 * @return new semisequent with conForm at place idx
	 */
	public Semisequent insertFirst(ConstrainedFormula conForm) {
	    return new Semisequent(conForm).checked();
	}

	/** inserts the element at the end of the semisequent
	 * @param conForm ConstrainedFormula to be inserted
	 * @return new semisequent with conForm at place idx
	 */	
	public Semisequent insertLast(ConstrainedFormula conForm) {
	    return insertFirst(conForm);
	}

	/** replaces the element at place idx with conForm
	 * @param idx an int specifying the index of the element that
	 * has to be replaced 
	 * @param conForm the ConstrainedFormula replacing the old
	 * element at index idx
	 */
	public Semisequent replace(int idx, ConstrainedFormula conForm) {
	    return insertFirst(conForm);
	}

	/** @return int counting the elements of this semisequent */ 
	public int size() { 
	    return 0; 
	}

	/** removes an element 
	 * @param idx int being the index of the element that has to
	 * be removed
	 * @return Semisequent without the element at idx 
	 */
	public Semisequent remove(int idx) {
	    return this;
	}

	/** returns index of a ConstrainedFormula 
	 * @param conForm the ConstrainedFormula the index want to be
	 * determined 
	 * @return index of conForm
	 */
	public int indexOf(ConstrainedFormula conForm) {
	    return -1;
	}

	/** gets the element at a specific index
	 * @param idx int representing the index of the element we
	 * want to have
	 * @return ConstrainedFormula found at index idx
	 */
	public ConstrainedFormula get(int idx) {
	    return null;
	}

	/** @return the first ConstrainedFormula of this Semisequent 
	 */
	public ConstrainedFormula getFirst() {
	    return null;
	}

	/** checks if a ConstrainedFormula is in this Semisequent
	 * @param conForm the ConstraintForumla to look for
	 * @return true iff. conForm has been found in this
	 * Semisequent 
	 */
	public boolean contains(ConstrainedFormula conForm) {
	    return false;
	}

	public boolean equals(Object o) {
	    return o == EMPTY_SEMISEQUENT;
	}

	/** @return String representation of this Semisequent */
	public String toString() {
	    return "[]";
	}
	

    }
    
}










