// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.logic;

import java.util.List;
import java.util.ArrayList;

import de.uka.ilkd.key.java.expression.literal.NullLiteral;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.Rulevariable;
import de.uka.ilkd.key.rule.ListOfUpdatePair;
import de.uka.ilkd.key.rule.UpdatePair;
import de.uka.ilkd.key.rule.SLListOfUpdatePair;
import de.uka.ilkd.key.rule.ListOfUpdatePair;

/*
 * creates instances of subclasses of {@link
 * de.uka.ilkd.key.logic.Term}.  As class {@link
 * de.uka.ilkd.key.logic.Term} is abstract, but access to
 * subclasses is only allowed via {@link
 * de.uka.ilkd.key.logic.Term} this class is used to create
 * special kind of Terms.
 */	
public class TermFactory {

    public static final Term[] NO_SUBTERMS = new Term[0];

    /** An instance of TermFactory */
    public static final TermFactory DEFAULT=new TermFactory();
    /*#private Term lnkTerm;*/

    //
    // CHANGE these two methods!  vars should be something like an 
    // array of arrays! 
    //


    /**
     * creates a term using the other methods of this class depending on the
     * given operator. If the kind of term is known before (without using
     * if-else cascades on the kind of operator) the other methods in this
     * factory should be preferred. Depending on the needed parameters for
     * the terms that should be created some of the parameters of this method
     * might be ignored.
     * @param op the top level operator for the new term.
     * @param subTerms the subterms for the new term. The first n elements
     * are taken if op is a Junctor or TermSymbol and n is the arity
     * of op. Only the first entry is taken if op is a Quantifier,
     * a Diamond or an Epsilon. The first (representing the replacing term
     * for a variable) and the second (representing the term behind the
     * substitution operator) entries are taken if op is a SubstOp.
     * @param vars the variables that are bound to a subterm. Not considered
     * if op is a Junctor, TermSymbol or Diamond. If op is an Epsilon or a
     * SubstOp only the first element is taken and the variable is bound to
     * the second subterm. In all other cases all variables are taken and
     * bound to the first subterm.
     * @param javaBlock representing a java code block. Only taken if op is a
     * Diamond.
     * @return the created Term
     */
    public Term createTerm(Operator op, Term[] subTerms, 
			   QuantifiableVariable[] vars, 
			   JavaBlock javaBlock, Rulevariable ruleVar) {
	return createTerm(op, subTerms, new ArrayOfQuantifiableVariable(vars),
			  javaBlock, ruleVar);
    }


   public Term createTerm(Operator op, Term[] subTerms, 
			  ArrayOfQuantifiableVariable vars, 
			  JavaBlock javaBlock, Rulevariable ruleVar) {
	if (op==null) {
	    throw new IllegalArgumentException("null-Operator at TermFactory");
	} else if (op instanceof Operator2) {
	    /* If the operator is an instance of Operator2, the
	     * operator can create the term by itself. Operator2
	     * supports no Java blocks and no ruleVar. */
	    if (javaBlock != JavaBlock.EMPTY_JAVABLOCK) {
		throw new IllegalArgumentException("Non-empty JavaBlock in Operator2.");
	    } else if (ruleVar != null) {
		throw new IllegalArgumentException("Operator2 can not have Rulevariable.");
	    }
	    return ((Operator2) op).createTerm(vars, subTerms);
	} else if (op instanceof SkolemFunction) { 
	    return createSkolemFunctionTerm((SkolemFunction)op,subTerms);
	} else if (op instanceof Proofobl) {
	    return createProofObligationTerm(ruleVar);
	} else if (op instanceof Equality) {
	    return createEqualityTerm(subTerms);
	} else if (op instanceof TermSymbol) {
	    return createFunctionTerm((TermSymbol)op,subTerms);
	} else if (op instanceof Junctor) {
	    return createJunctorTerm((Junctor)op,subTerms);
	} else if (op instanceof Quantifier) {
	    return createQuantifierTerm((Quantifier)op, vars, subTerms[0]);
	} else if (op instanceof SubstOp) {
	    return createSubstitutionTerm((SubstOp)op, 
					  vars.getQuantifiableVariable(0),
					  subTerms[0], 
					  subTerms[1]);
	} else if (op instanceof UpdateOp) {
	    return createUpdateTerm((UpdateOp)op, subTerms);
	} else if (op instanceof Modality) {
	    return createProgramTerm((Modality)op, javaBlock, subTerms[0]); 
	} else if (op instanceof Epsilon) {
	    return createEpsilonTerm(vars.getQuantifiableVariable(0), 
				     subTerms[0]);
	} else if (op instanceof AttributeOp) {
	    return createAttributeTerm((AttributeOp)op, subTerms[0]);
	} else if (op instanceof CondTermOp) {
	    return createCondTerm((CondTermOp)op, subTerms);
	} else if (op instanceof MetaOperator) {
	    return createMetaTerm((MetaOperator)op, subTerms);
	} else {
	    de.uka.ilkd.key.util.Debug.assert(false, "Should never be"+
					      " reached. Missing case.");
	    return null;
	}       
   }


   public Term createTerm(Operator op, Term[] subTerms, 
			  ArrayOfQuantifiableVariable[] bv,
			  JavaBlock javaBlock, Rulevariable ruleVar) {
	if (op==null) {
	    throw new IllegalArgumentException("null-Operator at TermFactory");
	} else if (op instanceof Operator2) {
	    /* If the operator is an instance of Operator2, the
	     * operator can create the term by itself. Operator2
	     * supports no Java blocks and no ruleVar. 
	     *
	     * Operator2 restrict the use of quantified variables. */
	    if (javaBlock != JavaBlock.EMPTY_JAVABLOCK) {
		throw new IllegalArgumentException("Non-empty JavaBlock in Operator2.");
	    } else if (ruleVar != null) {
		throw new IllegalArgumentException("Operator2 can not have Rulevariable.");
	    }
	    List list = new ArrayList();
	    for (int i = 0; i < bv.length; ++i) {
		for (int j = 0; j < bv[i].size(); ++j) {
		    list.add(bv[i].getQuantifiableVariable(j));
		}
	    }
	    QuantifiableVariable[] vars = (QuantifiableVariable[]) list.toArray(new QuantifiableVariable[list.size()]);
	    return ((Operator2) op).createTerm(new ArrayOfQuantifiableVariable(vars) , subTerms);
	} else if (op instanceof SkolemFunction) { 
	    return createSkolemFunctionTerm((SkolemFunction)op,subTerms);
	} else if (op instanceof Proofobl) {
	    return createProofObligationTerm(ruleVar);
	} else if (op instanceof Equality) {
	    return createEqualityTerm(subTerms);
	} else if (op instanceof TermSymbol) {
	    return createFunctionTerm((TermSymbol)op,subTerms);
	} else if (op instanceof Junctor) {
	    return createJunctorTerm((Junctor)op,subTerms);
	} else if (op instanceof Quantifier) {
	    return createQuantifierTerm((Quantifier)op, bv[0], subTerms[0]);
	} else if (op instanceof SubstOp) {
	    return createSubstitutionTerm((SubstOp)op, 
					  bv[1].getQuantifiableVariable(0),
					  subTerms[0], 
					  subTerms[1]);
	} else if (op instanceof UpdateOp) {
	    return createUpdateTerm((UpdateOp)op, subTerms);
	} else if (op instanceof Modality) {
	    return createProgramTerm((Modality)op, javaBlock, subTerms[0]); 
	} else if (op instanceof Epsilon) {
	    return createEpsilonTerm(bv[0].getQuantifiableVariable(0), 
				     subTerms[0]);
	} else if (op instanceof AttributeOp) {
	    return createAttributeTerm((AttributeOp)op, subTerms[0]);
	} else if (op instanceof CondTermOp) {
	    return createCondTerm((CondTermOp)op, subTerms);
	} else if (op instanceof MetaOperator) {
	    return createMetaTerm((MetaOperator)op, subTerms);
	} else {
	    de.uka.ilkd.key.util.Debug.assert(false, "Should never be"+
					      " reached. Missing case.");
	    return null;
	}       
   }



     /** creates a FunctionTerm with top operator op, some subterms
     * @param op Operator at the top of the termstructure that starts
     * here
     * @param an array containing subTerms (an array with length 0 if
     * there are no more subterms
     */
    public OpTerm createFunctionTerm(TermSymbol op, Term[] subTerms) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }

    public OpTerm createFunctionTerm(TermSymbol op) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");    
	return (OpTerm) (new OpTerm(op, NO_SUBTERMS)).checked();
    }

    public OpTerm createFunctionTerm(TermSymbol op, Term s1) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");    
	return (OpTerm) (new OpTerm(op, new Term[]{s1})).checked();
    }

    public OpTerm createFunctionTerm(TermSymbol op, Term s1,Term s2) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");    
	return (OpTerm) (new OpTerm(op, new Term[]{s1,s2})).checked();
    }

     /** creates a term consisting of the given variable.
     * @param v the variable
     */
    public OpTerm createVariableTerm(LogicVariable v) {
	return (OpTerm) (new OpTerm(v, NO_SUBTERMS)).checked();
    }
    
    public OpTerm createVariableTerm(SchemaVariable v) {
	return (OpTerm) (new OpTerm(v, NO_SUBTERMS)).checked();
    }
    
    public OpTerm createVariableTerm(ProgramVariable v) {
	return (OpTerm) (new OpTerm(v, NO_SUBTERMS)).checked();
    }
    
    /** creates a SkolemFunctionTerm with top operator op, some subterms
     * @param op Operator at the top of the termstructure that starts
     * here
     * @param an array containing subTerms (an array with length 0 if
     * there are no more subterms
     */
    public OpTerm createSkolemFunctionTerm(SkolemFunction op, Term[] subTerms) {
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }

    /**
     * creates a EqualityTerm with a given equality operator. USE THIS WITH
     * CARE! THERE IS NO CHECK THAT THE EQUALITY OPERATOR MATCHES THE TERMS.
     */
    public OpTerm createEqualityTerm(Equality op, Term[] subTerms) {
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }

    /**
     * creates a EqualityTerm with a given equality operator. USE THIS WITH
     * CARE! THERE IS NO CHECK THAT THE EQUALITY OPERATOR MATCHES THE TERMS.
     */
    public OpTerm createEqualityTerm(Equality op, Term t1, Term t2) {
	return (OpTerm) (new OpTerm(op, new Term[]{t1, t2} )).checked();
    }

 
    /** create an EqualityTerm with the correct equality symbol for
     * the sorts involved, according to {@link Sort#getEqualitySymbol}
     */
    public OpTerm createEqualityTerm(Term t1, Term t2) {
	Equality eq=t1.sort().getEqualitySymbol();
	if (t1.op() instanceof SchemaVariable) {
	    t2.sort().getEqualitySymbol();
	} 
	if (eq==null) eq=Op.EQUALS;
	return (OpTerm) new OpTerm(eq, new Term[]{t1, t2} ).checked();
    }


    /** create an EqualityTerm with the correct equality symbol for
     * the sorts involved, according to {@link Sort#getEqualitySymbol}
     */
    public OpTerm createEqualityTerm(Term[] terms) {
	return createEqualityTerm(terms[0], terms[1]);
    }
 
    

    /** creates a JunctorTerm with top operator op, some subterms
     * @param op Operator at the top of the termstructure that starts
     * here
     * @param an array containing subTerms (an array with length 0 if
     * there are no more subterms */
    public OpTerm createJunctorTerm(Junctor op, Term[] subTerms) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }

    public OpTerm createJunctorTerm(Junctor op) {
	return createJunctorTerm(op, NO_SUBTERMS);
    }

    public OpTerm createJunctorTerm(Junctor op, Term t1) {
	return createJunctorTerm(op, new Term[]{t1});
    } 

    public OpTerm createJunctorTerm(Junctor op, Term t1, Term t2) {
	return createJunctorTerm(op, new Term[]{t1, t2});
    }

    /** some methods for the creation of junctor terms with automatically performed simplification
     * like  ( b /\ true ) == (b) ...
     * Currently only the AND and OR Operator will be simplified (if possible) when on of the given
     * subterms is neutral with respect to the given junctor.
     */

    public Term createJunctorTermAndSimplify(Junctor op, Term t1, Term t2) {
	if (op == Op.AND) {
	    // if one of the terms is true skip the subterm.
	    if (t1.op() == Op.TRUE) {
		return  t2;
	    } else if(t2.op() == Op.TRUE) {
		return t1;
	    } else { // nothing to simplifiy ...
		return createJunctorTerm(op, t1, t2);
	    }
	} else if (op == Op.OR) {
	    // if one of the terms is false skip the subterm.
	    if (t1.op() == Op.FALSE) {
		return t2;
	    } else if(t2.op() == Op.FALSE) {
		return t1;
	    } else { // nothing to simplifiy ...
		return createJunctorTerm(op, t1, t2);
	    } 
	} else if (op == Op.IMP) {
	    // if t1 is true or t2 is false skip that subterm.
	    if (t1.op() == Op.TRUE) {
		return t2;
	    } else if(t2.op() == Op.FALSE) {
		return t1;
	    } else { // nothing to simplifiy ...
		return createJunctorTerm(op, t1, t2);
	    }
	
	} else { // all other Junctors ...
	    return createJunctorTerm(op, t1, t2);
	}
    }
    
    

    /** some methods to handle the equality for formulas (equiv - operator) ... */

   

    public OpTerm createJunctorTerm(Equality op, Term t1, Term t2) {
	return createEqualityTerm(op, new Term[]{t1, t2});
    }
    
    public OpTerm createJunctorTerm(Equality op, Term[] subTerms) {
	return createEqualityTerm(op, subTerms);
    }


     /** creates a substitution term
     * @param substVar the QuantifiableVariable to be substituted
     * @param substTerm the Term that replaces substVar
     * @param origTerm the Term that is substituted
     */
    public SubstitutionTerm createSubstitutionTerm(SubstOp op,
            QuantifiableVariable substVar, 
            Term substTerm, 
            Term origTerm) {
	return (SubstitutionTerm) new SubstitutionTerm(op, substVar, 
						       substTerm, 
						       origTerm).checked();
    }
    
    public UpdateOp createUpdateOp(ProgramVariable var) {
	return new UpdateOp(createVariableTerm(var));
    }

    public UpdateOp createUpdateOp(SchemaVariable var) {
	return new UpdateOp(createVariableTerm(var));
    }

    public UpdateOp createUpdateOp(Term t) {
	if (t.op() instanceof ProgramVariable) {
	    return createUpdateOp((ProgramVariable)t.op());
	} else {
	    return new UpdateOp(t);
	}
    }


    private Term createUpdateTermHelp(ListOfUpdatePair opList,
				      Term t) {
	if (opList == SLListOfUpdatePair.EMPTY_LIST) {
	    return t;
	} else {
	    UpdatePair pair = opList.head();
	    return createUpdateTerm
		(pair.updateOp(), pair.update(),
		 createUpdateTermHelp(opList.tail(), t));
	}
    }

    public UpdateTerm createUpdateTerm(ListOfUpdatePair opList,
				       Term t) {
	return (UpdateTerm)createUpdateTermHelp(opList, t);
    }


     /** creates an update-term
     * @param origTerm the Term that is updated
     */
    public UpdateTerm createUpdateTerm(UpdateOp op,
				       Term[] origTerm) {
	return (UpdateTerm) new UpdateTerm(op, origTerm[0], 
					   origTerm[1]).checked();
    }

     /** creates an update-term
     * @param value the Term that the variable is updated with
     * @param origTerm the Term that is updated
     */
    public UpdateTerm createUpdateTerm(UpdateOp op,
				       Term value,
				       Term origTerm) {
	return (UpdateTerm) new UpdateTerm(op, value, 
					   origTerm).checked();
    }

    /** creates an update-term
     * @param updatedVar the ProgramVariable to be updated
     * @param value the Term that the variable is updated with
     * @param origTerm the Term that is updated
     */
    public UpdateTerm createUpdateTerm(ProgramVariable updatedVar,
				       Term value,
				       Term origTerm) {

	return (UpdateTerm) new UpdateTerm(createUpdateOp(updatedVar), 
					   value, origTerm).checked();
    }

    /** creates an update-term
     * @param updatedVar the SchemaVariable to be updated
     * @param value the Term that the variable is updated with
     * @param origTerm the Term that is updated
     */
    public UpdateTerm createUpdateTerm(SchemaVariable updatedVar,
				       Term value,
				       Term origTerm) {

	return (UpdateTerm) new UpdateTerm(createUpdateOp(updatedVar), 
					   value, origTerm).checked();
    }

    /** creates an update-term
     * @param updatedVar the ProgramVariable to be updated
     */
    public UpdateTerm createUpdateTerm(ProgramVariable updatedVar,
				       Term[] subs) {

	return (UpdateTerm) new UpdateTerm(createUpdateOp(updatedVar), 
					   subs[0], subs[1]).checked();
    }
    

    /** creates an update-term
     * @param updatedAtt the Term that is the attribute to be updated
     * @param value the Term that the variable is updated with
     * @param origTerm the Term that is updated
     */
    public UpdateTerm createUpdateTerm(Term updatedAtt,
				       Term value,
				       Term origTerm) {

	return (UpdateTerm) new UpdateTerm
	    (createUpdateOp(updatedAtt), value, origTerm).checked();
    }

    /** creates an update-term
     * @param updatedAtt the Term that is the attribute to be updated
     */
    public UpdateTerm createUpdateTerm(Term updatedAtt,
				       Term[] subs) {
	return (UpdateTerm) new UpdateTerm
	    (createUpdateOp(updatedAtt), subs[0], subs[1]).checked();
    }


    /** creates an `conditional term' (t1 ?= t2) /-> t3
     * @param attr the AttributeOp that is the attribute to be updated
     * @param t1 the first part of the conditional
     * @param t2 the second part of the conditional
     * @param t3 the result of the conditional term
     */
    public OpTerm createCondTerm
	(AttributeOp attr, Term t1, Term t2, Term t3) {
	CondTermOp op = new CondTermOp(attr);
	return (OpTerm) new OpTerm
	    (op, new Term[]{t1,t2,t3}).checked();
    }

    /** creates an `conditional term' (t1 ?= t2).var /-> t3
     * @param var the variable the attribute term references to
     * @param prefixSort the sort of the type where the variable is declared
     * @param t1 the first part of the conditional
     * @param t2 the second part of the conditional
     * @param t3 the result of the conditional term
     */
    public OpTerm createCondTerm
	(ProgramVariable var, Sort prefixSort, 
	 Term t1, Term t2, Term t3) {
	AttributeOp attr = new AttributeOp(var, prefixSort);
	CondTermOp op = new CondTermOp(attr);
	return (OpTerm) new OpTerm
	    (op, new Term[]{t1,t2,t3}).checked();
    }

    /** creates an `conditional term' (t1 ?= t2) /-> t3
     * @param attr the AttributeOp that is the attribute to be updated
     * @param t1 the first part of the conditional
     * @param t2 the second part of the conditional
     * @param t3 the result of the conditional term
     */
    public OpTerm createCondTerm
	(CondTermOp op, Term t1, Term t2, 
	 Term t3) {
	return (OpTerm) new OpTerm
	    (op, new Term[]{t1,t2,t3}).checked();
    }

    /** creates an `conditional term' (t1 ?= t2) /-> t3
     * @param attr the AttributeOp that is the attribute to be updated
     * @param subs an array of Term with t1, t2, t3
     */
    public OpTerm createCondTerm
	(CondTermOp op, Term[] subs) {
	return (OpTerm) new OpTerm
	    (op, subs).checked();
    }

    


    /** creates an attribute term that references to a field of a class 
     * @param var the variable the attribute term references to
     * @param prefixSort the sort of the type where the variable is declared
     * @param term the Term describing the class/object of which the
     * attribute value has to be determined      
     * @return the attribute term "term.var"
     */
    public OpTerm createAttributeTerm(SchemaVariable var, 
				      Term term) {
	return (OpTerm) new OpTerm(new AttributeOp(var), 
			  new Term[]{term}).checked();
    }


    /** creates an attribute term that references to a field of a class 
     * @param var the variable the attribute term references to
     * @param prefixSort the sort of the type where the variable is declared
     * @param term the Term describing the class/object of which the
     * attribute value has to be determined      
     * @return the attribute term "term.var"
     */
    public OpTerm createAttributeTerm(ProgramVariable var, 
				      Sort prefixSort, 
				      Term term) {
	return (OpTerm) new OpTerm(new AttributeOp(var, prefixSort), 
			  new Term[]{term}).checked();
    }

    /** creates an attribute term that references to a field of a class 
     * @param attribOp the AttributeOperator 
     * @param term the Term the operator is applied at
     * @return the attribute term 
     */
    public OpTerm createAttributeTerm(AttributeOp attribOp,
				      Term term) {
	return (OpTerm) new OpTerm(attribOp, new Term[]{term}).checked();
    }

    public ProgramTerm createDiamondTerm(JavaBlock javaBlock, Term subTerm) {
	return (ProgramTerm) createProgramTerm(Op.DIA, javaBlock, subTerm);
    }

    public ProgramTerm createBoxTerm(JavaBlock javaBlock, Term subTerm) {
	return (ProgramTerm) createProgramTerm(Op.BOX, javaBlock, subTerm);
    }

    public ProgramTerm createProgramTerm(Modality op, 
					 JavaBlock javaBlock, 
					 Term subTerm) {
	return (ProgramTerm) new ProgramTerm(op, javaBlock, subTerm).checked();
    }

    /** creates a proofobligation term for the given rulevariable */   
    public ProofObligationTerm createProofObligationTerm(Rulevariable r) {
	return (ProofObligationTerm) new ProofObligationTerm(r).checked();
    }
    




    /**
     * creates a quantifier term 
     * @param op Operator representing the
     * Quantifier (all, exist) of this term 
     * @param varsBoundHere an
     * array of QuantifiableVariable containing all variables bound by the
     * quantifier
     */
    public QuantifierTerm createQuantifierTerm(Quantifier quant,
					       ArrayOfQuantifiableVariable 
					       varsBoundHere, 
					       Term subTerm) {
       return (QuantifierTerm) new QuantifierTerm(quant, varsBoundHere, 
						  subTerm).checked();
    }


    /**
     * creates a quantifier term 
     * @param op Operator representing the
     * Quantifier (all, exist) of this term 
     * @param varsBoundHere an
     * array of QuantifiableVariable containing all variables bound by the
     * quantifier
     */
    public QuantifierTerm createQuantifierTerm(Quantifier quant,
					       QuantifiableVariable[] 
					       varsBoundHere, 
					       Term subTerm) {
	return createQuantifierTerm(quant, new ArrayOfQuantifiableVariable
	    (varsBoundHere), subTerm);
    }



    /**
     * creates a quantifier term 
     * @param op Operator representing the
     * Quantifier (all, exist) of this term 
     * @param varsBoundHere a QuantifiableVariable representing the only bound
     * variable of this quantifier.
     */
    public QuantifierTerm createQuantifierTerm(Quantifier quant,
					       QuantifiableVariable var, 
					       Term subTerm) {
       return createQuantifierTerm
	   (quant, new QuantifiableVariable[]{var}, subTerm);
    }

    /**
     * creates an epsilon term 
     * @param varsBoundHere an
     * array of QuantifiableVariable containing all variables bound by the
     * quantifier
     */
    public QuantifierTerm createEpsilonTerm(QuantifiableVariable varBoundHere, 
					    Term subTerm) {
       return (QuantifierTerm) 
	   new QuantifierTerm(Op.EPSILON, 
			      new QuantifiableVariable[]{varBoundHere}, 
			      subTerm).checked();
    }

     /** creates a OpTerm with a meta operator as top operator. These
     * terms are only used in the replacewith areas of taclets. And are
     * replaced by the SyntacticalReplaceVisitor
     * @param op Operator at the top of the termstructure that starts
     * here
     * @param an array containing subTerms (an array with length 0 if
     * there are no more subterms
     */
    public OpTerm createMetaTerm(MetaOperator op, Term[] subTerms) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");    
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }


}
