// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import java.io.*;
import de.uka.ilkd.key.gui.*; // prettyprinters
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.gui.PresentationFeatures;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.util.Debug;
import javax.swing.JOptionPane;

public class ProofSaver {


   private KeYMediator mediator;
   private String filename;
   private Proof proof;
   

   public ProofSaver(KeYMediator mediator, String filename) {
      super();
      this.mediator = mediator;
      this.filename = filename;
      this.proof = mediator.getSelectedProof();
   }
   
   
   public String save() {
      String errorMsg = null;
      FileOutputStream fos = null;
      PrintStream ps = null;
      try {
          fos = new FileOutputStream(filename);
          ps = new PrintStream(fos);


          Sequent problemSeq = proof.root().sequent();
          SequentPrinter printer = 
             new PureSequentPrinter(new ProgramPrinter(null), 
	                            new NotationInfo(), problemSeq);

          ps.println("problem {");
          printer.printSemisequent(problemSeq.succedent());
          ps.println(printer.result());
          ps.println("}\n");
   //                ps.println(mediator.sort_ns());
          ps.println("proof {");
          ps.println(node2Proof(proof.root()));
          ps.println("}");

      } catch (IOException ioe) {
          errorMsg = "Could not save \n"+filename+".\n";
          errorMsg += ioe.toString();	    
      } catch (NullPointerException e) {
          errorMsg = "Could not save \n"+filename+"\n";
          errorMsg += "No proof present?";
          e.printStackTrace();
      } finally {
          try {
	      if (fos != null) fos.close();
          } catch (Exception e) {
	      System.err.println("Exception closing stream: "+e);
          }
          return errorMsg; // null if success
      }	    	    
   }
   


   private StringBuffer collectProof(Node node, String prefix, 
                                     StringBuffer tree) {       

      tree.append(prefix); 
      RuleApp appliedRuleApp = node.getAppliedRuleApp();
      if (appliedRuleApp == null) { // open goal
         tree.append("(opengoal \"");
         SequentPrinter printer = 
            new PureSequentPrinter(new ProgramPrinter(null), 
	                           new NotationInfo(), node.sequent());
         printer.printSequent(node.sequent());
         tree.append(printer.result().toString().replace('\n',' '));
         tree.append("\")\n");
         return tree;
      }

      IteratorOfNode childrenIt = node.childrenIterator(); 

      tree.append("(rule \"");
      tree.append(appliedRuleApp.rule().name());	
      tree.append("\" ");
      tree.append(posInOccurrence2Proof(appliedRuleApp.posInOccurrence()));
      tree.append(getInteresting(((TacletApp)appliedRuleApp).instantiations()));
      tree.append(")\n");
       
      if (node.childrenCount() == 0) return tree;
      if (node.childrenCount() == 1) 
         collectProof(childrenIt.next(), prefix, tree);
         else
	 while (childrenIt.hasNext()) {
            tree.append(prefix+"   ");
            tree.append("(branch \"dummy ID\"\n");
	    collectProof(childrenIt.next(), prefix+"   ", tree);
            tree.append(prefix+"   )\n");
	 }

      return tree;
   }


   public String node2Proof(Node node) {
       StringBuffer tree=new StringBuffer();
       return "(branch \"dummy ID\"\n"+collectProof(node, "",tree)+")\n"; 
   }

   
   public String posInOccurrence2Proof(PosInOccurrence pos) {
      if (pos == null) return "";
      return "(formula \""+pos.formulaNumberInSequent()+"\")"+
	    posInTerm2Proof(pos.posInTerm());
   }


   public String posInTerm2Proof(PosInTerm pos) {
      if (pos == PosInTerm.TOP_LEVEL) return "";
      String s = " (term \"";
      String list = pos.reversePositionList().toString(); // cheaper to read in
      s = s + list.substring(1,list.length()-1); // chop off "[" and "]"
      s = s + "\")";
      return s;
   }
   
   


   public String getInteresting(SVInstantiations inst) {
      String s = "";
      IteratorOfEntryOfSchemaVariableAndInstantiationEntry pairIt =
         inst.pairIterator();
      java.io.StringWriter sw = new java.io.StringWriter();
      ProgramPrinter prgPrinter = new ProgramPrinter(sw);
      //SequentPrinter printer = 
      //new PureSequentPrinter(prgPrinter, new NotationInfo(), null);
      while (pairIt.hasNext()) {
         EntryOfSchemaVariableAndInstantiationEntry pair = pairIt.next();
         SchemaVariable var = pair.key();
         if (var.isInteresting()) {
	     //s += " (inst \"";
	     //printer.printSchemaVariable(var);
	     //s += printer.result() + "=";
	     s += " (inst \""+var.name()+"=";
	     Object value = pair.value();
	     if (value instanceof TermInstantiation) {
		 Term term = ((TermInstantiation) value).getTerm();
		 s += "XXX"+term+"XXX";
	     }
	     if (value instanceof ProgramInstantiation) {
		 ProgramElement pe = 
		     ((ProgramInstantiation) value).getProgramElement();
		 try{
		     pe.prettyPrint(prgPrinter);
		 } catch(IOException ioe) {System.err.println(ioe);}
		 s += sw;
	     }
	     s += "\")";
         }
      }
      return s;
   }





}
