// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import java.io.*;
import java.util.HashMap;
import java.util.ArrayList;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.*;

/** This class has various public methods that take a logic data structure,
 * like an Taclet, etc. and replace entities of one sort by newly created ones
 * of another sort. 
 * <P> This is supposed to be used when creating standard Taclets for a sort.
 * you might need to be careful when using this class for anything else!
 * */

public class SortInstantiator {

    HashMap cache;

    Sort from,to;

    TermFactory fact;

    Namespace func_ns;

    public SortInstantiator(Sort from, Sort to, Namespace func_ns) {
	this.from = from;
	this.to = to;
	this.func_ns = func_ns;
	fact = TermFactory.DEFAULT;
	cache = new HashMap();
    }

    public ListOfTaclet inst(ListOfTaclet l) {
	if ( l.size() == 0 ) {
	    return l;
	} else {
	    return inst(l.tail()).prepend(inst(l.head()));
	}
    }

    public Taclet inst(Taclet taclet) {
	if (taclet instanceof FindTaclet) {
	    if (taclet instanceof RewriteTaclet) {
		return inst((RewriteTaclet)taclet);
	    } else if (taclet instanceof AntecTaclet) {
		return inst((AntecTaclet)taclet);
	    } else if (taclet instanceof SuccTaclet) {
		return inst((SuccTaclet)taclet);
	    } else {
		throw new RuntimeException
		    ("FindTaclet not instance of Rewrite-,Antec- or SuccTaclet");
	    }
	} else if (taclet instanceof NoFindTaclet) {
	    return inst((NoFindTaclet)taclet);
	} else {
	    throw new RuntimeException
		("Taclet not instance of FindTaclet or NoFindTaclet");
	}
    }

    public RewriteTaclet inst(RewriteTaclet taclet) {
	RewriteTacletBuilder b = new RewriteTacletBuilder();
	addFindTacletInfo(b,taclet);
	b.setFind(inst(taclet.find()));
	addGoalTemplates(b,taclet);
	return b.getRewriteTaclet();
    }

    public AntecTaclet inst(AntecTaclet taclet) {
	AntecTacletBuilder b = new AntecTacletBuilder();
	addFindTacletInfo(b,taclet);
	b.setFind(inst(taclet.find()));
	addGoalTemplates(b,taclet);
	return b.getAntecTaclet();
    }

    public SuccTaclet inst(SuccTaclet taclet) {
	SuccTacletBuilder b = new SuccTacletBuilder();
	addFindTacletInfo(b,taclet);
	b.setFind(inst(taclet.find()));
	addGoalTemplates(b,taclet);
	return b.getSuccTaclet();
    }

    public NoFindTaclet inst(NoFindTaclet taclet) {
	NoFindTacletBuilder b = new NoFindTacletBuilder();
	addTacletInfo(b,taclet);
	addGoalTemplates(b,taclet);
	return b.getNoFindTaclet();
    }

    public Sequent inst(Sequent seq) {
	return Sequent.createSequent(inst(seq.antecedent()),
				     inst(seq.succedent())  );
    }

    public Semisequent inst(Semisequent sseq) {
	Semisequent new_sseq = Semisequent.EMPTY_SEMISEQUENT;
	IteratorOfConstrainedFormula it = sseq.iterator();
	while( it.hasNext() ) {
	    new_sseq = new_sseq.insertLast(inst(it.next()));
	}
	return new_sseq;
    }

    public ConstrainedFormula inst(ConstrainedFormula cf) {
	return new ConstrainedFormula(inst(cf.formula()),inst(cf.constraint()));
    }

    private Sort getSort(Sort s) {
	if (s==from) {
	    return to;
	} else if (s instanceof CollectionSort
		   && to instanceof NonCollectionSort) {
	    if (s instanceof SetSort) {
		return ((NonCollectionSort)to).getSetSort();
	    } else if (s instanceof SequenceSort) {
		return ((NonCollectionSort)to).getSequenceSort();
	    } else if (s instanceof BagSort) {
		return ((NonCollectionSort)to).getBagSort();
	    }
	}
	return s;	
    }

    private Name createName(Sort s, Function f) {
	// %% hacky
	String prefix="";
	int offset=0;
	if (f.name().toString().startsWith("SetOfany::")) {
	    prefix="SetOf";
	    offset=10;
	} else if (f.name().toString().startsWith("SequenceOfany::")) {
	    prefix="SequenceOf";
	    offset=15;
	} else if (f.name().toString().startsWith("BagOfany::")) {
	    prefix="BagOf";
	    offset=10;
	}
	return new Name(prefix+s.name()+"::"
			+f.name().toString().substring(offset));
    }

    public Function inst(Function f) {
	Function result=(Function)cache.get(f);
	boolean changed=false;
	if (result==null) {
	    Sort[] sorts=new Sort[f.arity()];
	    for (int i=0; i<f.arity(); i++) {
		sorts[i]=getSort(f.argSort(i));
		if (sorts[i]!=f.argSort(i)) {
		    changed=true;
		}
	    }
	    Sort retSort=getSort(f.sort());
	    if (retSort!=f.sort()) {
		changed=true;
	    }
	    if (changed) {
		Name newName=createName(to, f);
		result = new Function(newName, retSort, sorts);
	    } else {
		result=f;
	    }
	    cache.put(f, result);
	}
	return result;
    }

    /** We don't really need to do anything with constraints yet, as these
     * don't appear in parsed standard Taclets. */
    private Constraint inst(Constraint c) {
	return Constraint.BOTTOM;
    }
    
    /** This just builds new terms.  It seems improbable that we can
     * share much for the typical `for all sorts' Taclet. 
     * Also, it can cope with boolean junctors and equality and functions.
     */
    private Term inst(Term t) {
	Operator op = t.op();
	if ( op instanceof SchemaVariable ) {
	    return fact.createVariableTerm(map((SchemaVariable)op));
	} else if ( op instanceof Equality ) {
	    return fact.createEqualityTerm(inst(t.sub(0)),inst(t.sub(1)));
	} else {
	    if ((op.name().toString().startsWith("SetOf"+from.name()+"::")) ||
		(op.name().toString().startsWith("SequenceOf"+from.name()
						 +"::")) ||
		(op.name().toString().startsWith("BagOf"+from.name()+"::"))) {
		String oldName=op.name().toString();
		int firstocc=oldName.indexOf(from.name().toString());
		int endocc=firstocc+from.name().toString().length();
		String newName;
// 		if (to instanceof ObjectSort) {
// 		    newName=oldName.substring(0,firstocc)+"Object"
// 			+oldName.substring(endocc);
// 		} else {
		newName=oldName.substring(0,firstocc)+to.name()
		    +oldName.substring(endocc);
		    //		}
		op=(Function)func_ns.lookup(new Name(newName));
		if (op==null) {
		    throw new InstantiationImpossibleException();
		}
	    } 
	    int arity = t.arity();
	    Term subTerms[] = new Term[arity];
	    // Stupid hack because of another stupid hack in TermFactory
	    ArrayList vars = new ArrayList();
	    for ( int i = 0; i<arity; i++ ) {
		ArrayOfQuantifiableVariable vbh = t.varsBoundHere(i);
		for( int j=0; j<vbh.size(); j++ ) {
		    vars.add(map((SchemaVariable)
				 (vbh.getQuantifiableVariable(j))));
		}
		subTerms[i] = inst(t.sub(i));
	    }
	    QuantifiableVariable v1[] = new QuantifiableVariable[0];
	    v1 = (QuantifiableVariable[])(vars.toArray(v1));
	    return fact.createTerm(op,subTerms,v1,t.javaBlock(),null);
	    //!!!! javaBlock not instantiated !!!%%%
	}
    }

    private void addFindTacletInfo(FindTacletBuilder b,FindTaclet taclet) {
	addTacletInfo(b,taclet);
	// can't really set find here, because it is defined further up.
    }
    
    private void addTacletInfo(TacletBuilder b,Taclet taclet) {
	b.setName(new Name(taclet.name()+"_"+to));
	b.setIfseq(inst(taclet.ifseq()));
	addVarsNew(b,taclet);
	addVarsNotFreeIn(b,taclet);
	addHeuristics(b,taclet);
	// We don't handle the constraint, since that can't appear in the 
	// template files, and I don't want to traverse them.
	addAttrs(b,taclet);
    }

    private void addVarsNew(TacletBuilder b,Taclet taclet) {
	IteratorOfNewVarcond it = taclet.varsNew().iterator();
	while( it.hasNext() ) {
	    NewVarcond v = it.next();
	    if (v.isDefinedByType()) {
		b.addVarsNew(map(v.getSchemaVariable()), v.getSort());
	    } else {
		b.addVarsNew(map(v.getSchemaVariable()), map(v.getPeerSchemaVariable()));
	    }
	}
    }
    
    private void addVarsNotFreeIn(TacletBuilder b,Taclet taclet) {
	IteratorOfNotFreeIn it = taclet.varsNotFreeIn();
	while( it.hasNext() ) {
	    NotFreeIn xy = it.next();
	    b.addVarsNotFreeIn((SchemaVariable)map(xy.first()), map(xy.second()));
	}
    }

    private void addHeuristics(TacletBuilder b,Taclet taclet) {
	IteratorOfHeuristic it = taclet.heuristics();
	while( it.hasNext() ) {
	    b.addHeuristic(it.next());
	}
    }

    private void addGoalTemplates(TacletBuilder b,Taclet taclet) {
	IteratorOfTacletGoalTemplate it = taclet.goalTemplates().iterator();
	while( it.hasNext() ) {
	    TacletGoalTemplate gt = it.next();
	    TacletGoalTemplate new_gt;
	    if ( gt instanceof AntecSuccTacletGoalTemplate ) {
		Sequent new_rw = inst(((AntecSuccTacletGoalTemplate)gt)
				      .replaceWith());
		new_gt = new AntecSuccTacletGoalTemplate(inst(gt.seq()),
						       inst(gt.rules()),
						       gt.ruleVars(),
						       new_rw);
	    } else if ( gt instanceof RewriteTacletGoalTemplate ) {
		Term new_rw = inst(((RewriteTacletGoalTemplate)gt)
				   .replaceWith());
		new_gt = new RewriteTacletGoalTemplate(inst(gt.seq()),
						     inst(gt.rules()),
						     gt.ruleVars(),
						     new_rw);
	    } else {
		new_gt = new TacletGoalTemplate(inst(gt.seq()),
					      inst(gt.rules()),
					      gt.ruleVars());

	    }
	    b.addTacletGoalTemplate(new_gt);
	}
    }

    

    private void addAttrs(TacletBuilder b,Taclet taclet) {
	b.setRecursive(taclet.recursive());
	b.setNoninteractive(taclet.noninteractive());
        b.setDisplayName(taclet.displayName());
        b.setHelpText(taclet.helpText());
    }
    
    private SchemaVariable map(SchemaVariable s) {
	if ( s.sort() == from ) {
	    SchemaVariable v = (SchemaVariable)(cache.get(s));
	    if ( v == null ) {
		if (s.isVariableSV()) {
		    v = SchemaVariable.createVariableSV(s.name(), to);
		} else {
		    v = SchemaVariable.createTermSV(s.name(), to);
		}
		cache.put(s,v);
	    }
	    return v;
	} else if (s.sort() instanceof CollectionSort 
		   && (to instanceof NonCollectionSort)
		   && (((CollectionSort)s.sort()).elementSort() == from)) {
	    SchemaVariable v = (SchemaVariable)(cache.get(s));
	    if ( v == null ) {
		Sort instTo=null;
		if (s.sort() instanceof SequenceSort) {
		    instTo=((NonCollectionSort)to).getSequenceSort();
		} else if (s.sort() instanceof BagSort) {
		    instTo=((NonCollectionSort)to).getBagSort();
		} else if (s.sort() instanceof SetSort) {
		    instTo=((NonCollectionSort)to).getSetSort();
		}
		if (s.isVariableSV()) {
		    v = SchemaVariable.createVariableSV(s.name(), instTo);
		} else {
		    v = SchemaVariable.createTermSV(s.name(), instTo);
		}
		cache.put(s,v);
	    }
	    return v;
	    
	} else{
	    return s;
	}
    }


    public static SetOfTaclet addSortRules(SetOfTaclet taclets, Sort s, 
					   Namespace func_ns, 
					   SetOfTaclet instTaclets,
					   Sort anySort) {
	IteratorOfTaclet it = taclets.iterator();	
	while(it.hasNext()) {
	    Taclet taclet = it.next();
	    Taclet instTaclet=null;
	    try {
		instTaclet = new SortInstantiator(anySort,s,func_ns).inst(taclet);
		instTaclets=instTaclets.add(instTaclet);
	    } catch (SortInstantiator.InstantiationImpossibleException e) {
		//don't do anything: skip collection sorts
	    }
	}
	return instTaclets;
    }

    public static SetOfTaclet addSortRules(SetOfTaclet taclets, Namespace sort_ns, 
					   Namespace func_ns, Sort anySort) {
	SetOfTaclet newTaclets=SetAsListOfTaclet.EMPTY_SET;
	IteratorOfNamed it = sort_ns.allElements().iterator();
	while(it.hasNext()) {
	    Sort s = (Sort)(it.next());
	    newTaclets=addSortRules(taclets, s, func_ns, newTaclets, anySort);
	}
	return newTaclets;
    }


    public static class InstantiationImpossibleException extends RuntimeException{
	
    }
}

