// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import javax.swing.table.AbstractTableModel;
import javax.swing.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Label;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Expression;
import java.io.*;
import java.util.Vector;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.IteratorOfSchemaVariable;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.rule.TacletVariableCollector;
import de.uka.ilkd.key.java.TypeConverter;

public class TacletInstantiationsTableModel extends AbstractTableModel {
	
    /** the instantiations entries */
    private Vector entries; 
    /** the related rule application */
    private TacletApp app;
    /** the integer defines the row until which no editing is possible */
    private int noEditRow;
    /** universal namespace of variables, minimum for input in a row */
    private Namespace var_ns;
    /** namespace of functions */
    private Namespace func_ns;
    /** namespace of sorts */
    private Namespace sort_ns;
    /** the java information object */
    private JavaInfo javaInfo;
    

    // This stuff is for generating unique name proposal for schema variables
    // for labels or program variables ...
    private static final String LABEL_NAME_PREFIX = "_label";
    private static final String VARIABLE_NAME_PREFIX = "_var";
    
    public static int globalLabelNameCnt = 0;
    public static int globalVariableNameCnt = 0;

	
    /** create new data model for tree 
     * @param app the TacletApp where to get the necessary entries
     */
    public TacletInstantiationsTableModel(TacletApp app, 
					JavaInfo javaInfo,
					Namespace var_ns,
					Namespace func_ns, 
					Namespace sort_ns) {
	this.app = app;
	this.javaInfo = javaInfo;
	this.var_ns = var_ns;
	this.func_ns = func_ns;
	this.sort_ns = sort_ns;

	entries = createEntryArray(app);
    }
	
    /** creates a Vector with the row entries of the table
    */
    private Vector createEntryArray(TacletApp app) {
	Vector rowVec = new Vector();       
	IteratorOfSchemaVariable it = app.instantiations().svIterator();	
	int count = 0;

	while (it.hasNext()) {
	    SchemaVariable sv = it.next();
	    Object[] column = new Object[2];
	    column[0] = sv;
	    column[1] = app.instantiations().getInstantiation(sv);	    
	    rowVec.add(column);
	    count++;
	}
	noEditRow = count - 1;

	IteratorOfSchemaVariable varIt = app.uninstantiatedVars().iterator();
	TacletVariableCollector svc = new TacletVariableCollector();
	svc.visit((app.taclet()).ifseq());

	while (varIt.hasNext()) {
	    Object[] column = new Object[2];	    
	    SchemaVariable var = varIt.next();
	    if (!svc.contains(var)) {
		column[0] = var;
                var.markInteresting();
		
		// NEW ... : create an appropriate and unique proposal for the name ...
	
		String nameProposal = getNameProposalForSchemaVariable(var);	
	       
		if(nameProposal != null) {
		    // A proposal is available ...
		    column[1] = nameProposal;
		}
		
		
		rowVec.add(column);	
	    }
	}

	return rowVec;
    }

    /** adds an instantiation of a schemavariable */
    public void addInstantiationEntry(int row, Term instantiation) {
	((Object[])entries.get(row))[1] = instantiation;	
    }

    /** return the rule application which is the table models base
     *  @return the Ruleapp 
     */
    public TacletApp application() {
	return app;
    }

    /** number of colums
     * @return number of colums
     */
    public int getColumnCount() { 
	return 2; 
    }

    /** number of rows
     * @return number of rows
     */
    public int getRowCount() { 
	return entries.size();
    }

    /** returns true iff an instantiation is missing 
     * @return true iff an instantiation is missing
     */
    public boolean isCellEditable(int rowIndex, int columnIndex) {
	return (rowIndex > noEditRow) && (columnIndex > 0);
    }


    /** parses the given string that represents the term (or formula)
     * using the given variable namespace and the default namespaces
     * for functions and sorts
     * @param s the String to parse 
     * @param sort the expected sort
     * @param varNS the variable namespace
     */
    public Term parseTerm(String s, Sort sort, Namespace varNS) 
	throws ParserException
    {
	return TermParserFactory.createInstance()
	    .parse(new StringReader(s), sort, javaInfo, 
		   varNS, func_ns, sort_ns);
    }
    
    /**
     * throws an exception iff no input in indicated row
     *
     */
    private void checkInputAvailable(int irow) throws SVInstantiationException {
	if ((getValueAt(irow, 1) == null) 
	    ||(((String)getValueAt(irow,1)).length() == 0)) {
	    throw new SVInstantiationException
		("No instantiation entered.", "", irow, -1, null);		    
	}
    }


    /**
     * parses the indicated row and returns a Term corresponding to the
     * entry in the row
     *
     * @param irow the row to be parsed
     * @return the parsed term
     */
    private Term parseRow(int irow, Namespace varNS) 
	throws SVInstantiationException {

	String instantiation = (String) getValueAt(irow, 1); 
	try {
	    SchemaVariable sv=(SchemaVariable)getValueAt(irow, 0);
	    return parseTerm(instantiation, sv.sort(), varNS);
	} catch (ParserException pe) {
	    Location loc = pe.getLocation();
	    if (loc != null) {
		throw new SVInstantiationException("Error parsing ", instantiation, 
						   irow + (loc.getLine() <= 0 ? 0 : loc.getLine()), 
						   loc.getColumn(), pe);
	    } else {
		throw new SVInstantiationException("Error parsing ", instantiation, 
						   irow, -1, pe);
	    }
	} catch (Exception exc) {
	    /* This catch clause should be removed, because it is to
	     * general, 07.12.2002, -hs. */	    
	    throw new SVInstantiationException("Error parsing ", instantiation, 
					       irow, -1, exc);
	}
    }

    public static ProgramElement getProgramElement(TacletApp app, 
						   String instantiation,
						   SchemaVariable sv, 
						   JavaInfo javaInfo) {
	Sort svSort = sv.sort();
	if (svSort == ProgramSVSort.LABEL) {
	    return new ProgramElementName(instantiation);
	} else if (svSort == ProgramSVSort.VARIABLE 
		   || svSort == ProgramSVSort.BOOLEANVARIABLE) {
	    Object o = app.taclet().varDeclaredNew(sv);
	    if (o != null) {		
		KeYJavaType kjt;
		if (o instanceof SchemaVariable) {
		    TypeConverter tc=javaInfo.getTypeConverter();
		    SchemaVariable peerSV=(SchemaVariable)o;
		    Object peerInst = app.instantiations().
			getInstantiation(peerSV);
		    Expression peerInstExpr;
		    if (peerInst instanceof Term) {
			peerInstExpr=tc.convertToProgramElement((Term)peerInst);
		    } else {
			peerInstExpr=(Expression)peerInst;
		    }
		    kjt = tc.getPrimitiveKeYJavaType(peerInstExpr);
		} else {
		    kjt = javaInfo.getKeYJavaType(((Sort)o).name().toString());
		}
		return new ProgramVariable
		    (new ProgramElementName(instantiation), kjt); 
	    }	
	}
	return null;
    }


    /**
     * parses the indicated row and returns the ProgramElement 
     * corresponding to the entry in the row
     * @param irow the row to be parsed
     * @return the parsed term
     */
    private ProgramElement parseRow(int irow) 
	throws SVInstantiationException {

	String instantiation=(String) getValueAt(irow, 1); 
	SchemaVariable sv= (SchemaVariable)getValueAt(irow, 0);
	ProgramElement pe = getProgramElement(app, instantiation, sv, javaInfo);
	if (pe==null) {
	    throw new SVInstantiationException
		("Unexpected SchemaVariable: "+sv.sort()
		 +". Label SV or a program"
		 +" variable SV expected declared as new.","",
		 irow+1, -1, null);
	}
	return pe;
    }
    

    /**
     * creates new rule app with all inserted instantiations in the variable
     * instantiations table
     * @throws SVInstantiationException if the instantiation is incorrect
     */
    public TacletApp createTacletAppFromVarInsts() throws SVInstantiationException {
	TacletApp result = app;
	for (int irow = noEditRow+1; irow < entries.size(); irow++) {
	    checkInputAvailable(irow);
	    SchemaVariable sv=(SchemaVariable) getValueAt(irow, 0);
	    if (sv.isVariableSV()) {
		String newVarName=(String)getValueAt(irow, 1);
		LogicVariable lv=new LogicVariable(new Name(newVarName),
						   sv.sort());
		Term instance = TermFactory.DEFAULT.createVariableTerm(lv);
		result = result.addCheckedInstantiation(sv, instance);
	    } else if (sv.isProgramSV()) {
		ProgramElement pe = parseRow(irow);
		result = result.addInstantiation(sv, pe);
	    } 
	}
	SchemaVariable problemVarSV = result.varSVNameConflict();
	if (problemVarSV!=null) {
	    throw new SVInstantiationException
		("Ambiguous instantiation of schema "
		 +"variable "+problemVarSV+".", "", getSVRow(problemVarSV), 
		 -1, null);
	}
	for (int irow = noEditRow+1; irow < entries.size(); irow++) {
	    if (!(getValueAt(irow, 0) instanceof SchemaVariable &&
		  ((SchemaVariable)getValueAt(irow, 0)).isVariableSV())) {
		checkInputAvailable(irow);
		SchemaVariable sv=(SchemaVariable)getValueAt(irow, 0);
		if (sv.isProgramSV()) {
		    ProgramElement pe = parseRow(irow);
		    result = result.addInstantiation(sv, pe);
		} else {		    
		    Namespace extVarNS=result.extendVarNamespaceForSV(var_ns, sv);
		    Term instance = parseRow(irow, extVarNS);
		    result = result.addCheckedInstantiation(sv,instance); 
		}
	    }
	}
	return result;

    }


    /** sets the Value of the cell */
    public void setValueAt(Object instantiation, int rowIndex, 
			   int columnIndex) { 
	((Object[])entries.get(rowIndex))[columnIndex] = instantiation;
	
    }
   

    /** get value at the specified row and col
     * @return the value
     */
    public Object getValueAt(int row, int col) {
	return 	((Object[])entries.get(row))[col]; 
    }

    /** returns the index of the row the given Schemavariable stands 
     *@return the index of the row the given Schemavariable stands (-1
     * if not found)  
     */
    public int getSVRow(SchemaVariable sv) {
	for (int i = 0; i < entries.size(); i++) {
	    if (getValueAt(i, 0).equals(sv)) {
		return i;
	    }
	}
	return -1;
    }

    

    /**
     * Generates a proposal for the instantiation of the given
     * schema variable.
     * Currently this proposal only serves for the case of names (for labels or 
     * variables) and generates a unique new name.
     * @param var the schema var for which a unique name should be proposed as 
     * its instantiation.
     * @return a new and unquie name that can be used as an instantiation for
     *         the given schema variable.
     */
    public static String getNameProposalForSchemaVariable(SchemaVariable var){
	String result = null;

	Sort svSort = var.sort();
	if (svSort == ProgramSVSort.LABEL) {
	    result = LABEL_NAME_PREFIX + globalLabelNameCnt++;
	} else if (svSort == ProgramSVSort.VARIABLE 
		   || svSort == ProgramSVSort.BOOLEANVARIABLE) {
	    result = VARIABLE_NAME_PREFIX + globalVariableNameCnt++;
	}
	
	return result;
    }


}
