// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.TogetherFacade;
import de.uka.ilkd.key.casetool.together.*;
import de.uka.ilkd.key.ocl.tudresden.TUDresdenParserAdapter;
import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.*;
import de.uka.ilkd.key.gui.ProofSettings;
import tudresden.ocl.*;

import java.io.*;

/** represents an OCL proof obligation input to the prover for proving the 
 * behavioral subtyping property.
 */
public class OCLBehavSubPO extends OCLProofOblInput {

    private ReprModelMethod methRepr;
    private String preText;
    private String postText;
    private String preTextParent;
    private String postTextParent;
    private String contextString;
    private String xmifile;

    /** creates a new OCL proof obligation input representing a behavioral
     * subtyping property proof obligation using 
     * an XMI file name for reading the UML model from, a representation 
     * of the model class the PO is invoked from and five strings, (1)
     * describing the contextString, (2) describing the pre condition text
     * of the parent, (3) describing the pre condition text of the parent class,
     * (4) describing the post condtion string of the subclass, and (5)
     * describing the post condition of the parent class.
     */
    public OCLBehavSubPO(ReprModelMethod aMethRepr, String xmifile, 
			 String contextString, String preTextParent, 
			 String preText, String postTextParent, 
			 String postText) {
	super("BehavioralSubtyping");
	this.methRepr=aMethRepr;
	this.xmifile=xmifile;
	this.contextString=contextString;
	this.preTextParent=preTextParent;
	this.preText=preText;
	this.postTextParent=postTextParent;
	this.postText=postText;
    }

    /** reads the Java model for this PO -- do we need it?
     */
    public void readJava() {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	Recoder2KeY r2k = new Recoder2KeY("", initConfig.sortNS(), 
					  initConfig.javaInfo().rec2key(), 
					  initConfig.javaInfo()
					           .getTypeConverter());
	String[] cus = methRepr.getContainingClass().getClassesInPackage();
	r2k.readCompilationUnitsAsFiles(cus);
    }


    /** reads this OCL proof obligation, thus invoking the OCL-DL-Translation
     */
    public void read(ModStrategy mod) {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	TUDresdenParserAdapter aAdapter1 = new TUDresdenParserAdapter();
	TUDresdenParserAdapter aAdapter2 = new TUDresdenParserAdapter();


	OclTree prePOAsTree 
	    = aAdapter1.doParse1
	    (contextString + " pre:",  xmifile, "context " 
	     + contextString + " pre: " + "((" + 
	     FunctionalityOnModel.normalizeConstraint(preTextParent) + ")" + 
	     " implies " + "(" 
	     + FunctionalityOnModel.normalizeConstraint(preText) + "))",
	     methRepr.getContainingPackage());
	
	
	OclTree postPOAsTree 
	    = aAdapter2.doParse1
	    (contextString + " post:",  xmifile, "context " 
	     + contextString + " pre: " + "(" + 
	     FunctionalityOnModel.normalizeConstraint(preTextParent) 
	     + " and " + 
	     FunctionalityOnModel.normalizeConstraint(preText) + ")" 
	     + " post: " + 
	     "((" + FunctionalityOnModel.normalizeConstraint(postText) + ")" 
	     + " implies " +
	     "(" + FunctionalityOnModel.normalizeConstraint(postTextParent) 
	     + "))",
	     methRepr.getContainingPackage());

        OCLTreeToDLAggregator aOCLTreeToDLAggregator;
        try{
	    TogetherFacade po2prover = initConfig.createFacadeForTransl();
	    OCLMod[] mods
		=ProofSettings.THE_SETTINGS
		.getOCLRefinementSettings().getOCLMods();
	    aOCLTreeToDLAggregator =  getAggregator(po2prover, mods);
	    if ( !(aAdapter1.getErrorOccurred() 
		   || aAdapter2.getErrorOccurred()) ){
		Term preProofObl 
		    = aOCLTreeToDLAggregator.treeToTerm(prePOAsTree
							, false, true);
		Term postProofObl 
		    = aOCLTreeToDLAggregator.treeToTerm(postPOAsTree
							, false, false);
		result=new POInputResult();
		result.addTerm("pre conditions", preProofObl);
		result.addTerm("post conditions", postProofObl);
	    } else {
		throw new RuntimeException("Pre-/Post Conditions are "
					   +"syntactically incorrect");
	    }
    	}catch(Exception e){
	    System.out.println("KeY-Error: "+ e);
	    e.printStackTrace();
            throw new RuntimeException( "KeY: "+e.getMessage());
    	}

    }

}
