// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.*;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.TogetherFacade;
import de.uka.ilkd.key.casetool.together.*;
import de.uka.ilkd.key.ocl.tudresden.TUDresdenParserAdapter;
import de.uka.ilkd.key.gui.ProofSettings;
import tudresden.ocl.*;
import java.io.*;

/** Represents the proof obligation input for the prover as an OCL 
 * constraint in textual format.
 */
public abstract class OCLProofOblInput implements ProofOblInput {

    /** The initial configuration for a prover to use for reading and modify*/
    protected InitConfig initConfig = null;

    /** The result of reading the OCL input*/
    protected POInputResult result = null;

    private String name;

    /** Creates a new representation of an OCL input with a name describing
     * the kind of input.
     */
    public OCLProofOblInput(String name) {
	this.name=name;
    }

    /** returns an OCLTreeToDLAggregator for use in the OCL-DL-Translation.
     * @param mods Modifications to OCL constraints and OCL model used
     * to parse model extensions such as excThrown or null. If this parameter
     * has length 0 a usual OCLTODLAggregator is returned, otherwise a 
     * DetailedAggregator considering the modifications is returned.
     */
    protected OCLTreeToDLAggregator getAggregator(TogetherFacade facade, 
                                                         OCLMod[] mods) {
	if (mods.length==0) {
	    return new OCLTreeToDLAggregator(facade);
	}
	return new DetailedAggregator(facade, mods);
    }


    /**
     * generate proof obligation as DL text.
     * @param aTag controls the form of the generated text
     *           values for aTag: purepre, prepost, prepostaspl
     */
    protected Term generateMethodPO(TogetherFacade po2prover,
                                         ReprModelMethod aMethRepr,
                                         String aTag,
                                         String preText, String postText,
                                         String xmifile) {
	Debug.out("generate Method PO");
	String returnvalue = "";
	String contextString = aMethRepr.getContainingClassName() + "::" + 
           aMethRepr.getCallSignature();

	TUDresdenParserAdapter aAdapter = new TUDresdenParserAdapter();
	OclTree methPOAsTree;

	
 	// generating proofobligation: preText -> <m> postText
	methPOAsTree= aAdapter.doParse1
	    (contextString + aTag,  xmifile, 
	     "context " + contextString + 
	     " pre: " + FunctionalityOnModel.normalizeConstraint(preText) +
	     " post: " + FunctionalityOnModel.normalizeConstraint(postText), 
	     aMethRepr.getContainingPackage());
	
        OCLTreeToDLAggregator aOCLTreeToDLAggregator;
        try {
	    OCLMod[] mods=ProofSettings.THE_SETTINGS
		.getOCLRefinementSettings().getOCLMods();
	    aOCLTreeToDLAggregator =  getAggregator(po2prover, mods);
	    if (methPOAsTree!=null) {
		Term methProofObl;
                // only preconditions are considered
		if (aTag.equals("purepre")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    false, true);
		    // is transformed into a DL-form
		} else if ( aTag.equals("prepost")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    true, false);
		    // transformation into PL form where diamond is omitted
		    // and @pre constructs in the postconditions are eliminated
		} else if ( aTag.equals("prepostaspl")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    false, false);
		} else { 
		    throw new RuntimeException("unexpected tag in "
					       +"generatedMethodPO: " + aTag);
		}
            	return methProofObl;
	    } else {
		throw new RuntimeException("generateMethodPO() failed");
	    }
    	} catch(Exception e){
	    System.out.println("KeY-Error: "+ e);
	    e.printStackTrace();
            throw new RuntimeException("generateMethodPO() failed");
    	}
    }

    /** returns the problem term as generated by reading the OCL constraint
     * and translating it to a DL formula.
     */
    public POInputResult getProblemTerm() {
	return result;
    }

    /** returns the initial configuration that is used to read the OCL
     * input and that is used to be modified during reading.
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** reads the OCL constraint and modifies the initial prover configuration
     * according to the given modification strategy.
     */
    public abstract void read(ModStrategy mod);

    /** sets the initial configuration that is used to read the OCL
     * input and that is used to be modified during reading.
     */
    public void setConfig(InitConfig conf) {
	this.initConfig=conf;
    }

    /** returns the name of the OCL proof obligation input.
     */
    public String name() {
	return name;
    }

}
