// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import java.io.*;
import de.uka.ilkd.key.gui.Main;

/** represents a .key file were all rules are instantiated with all available 
 * sorts from an initial configuration. A declared sort with a special name 
 * <code>any</code> is used to help to represent the uninstantiated taclets.
 * As some taclets refer to a number sort that depends on the number LDT
 * read before there is a second template sort <code>NAT</code> that is
 * replaced by the current integer sort as delivered by the integer LDT.
 */
public class SortKeYFile extends KeYFile {

    private Sort anySort=null;

    /** The name of the `parameter' sort in SORT_RULES */
    private static final Name ANY_SORT_NAME = new Name("any");
    /** The name of the number `parameter' sort in SORT_RULES */
    private static final Name NAT_SORT_NAME = new Name("NAT");


    /** creates a new representation of a SortKeYFile with the given name,
     * a rule source representing the physical source of the input, and
     * a graphical representation to call back in order to report the progress
     * while reading.
     */
    public SortKeYFile(String name, RuleSource file, Main main) {
	super(name, file, main);
    }

    /** reads the input, instantiates the sort <code>any</code> to all available
     * sorts in the initial configuration, instantiates the sort <code>NAT</code>
     * to the current integer sort (that in the TypeConverter of the JavaInfo), 
     * and modifies the initial configuration by adding the instantiated rules
     * with respect to the given modification strategy.
     */
    public void read(ModStrategy mod) {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}

	NumberLDT intLDT=initConfig.javaInfo().getTypeConverter().getIntegerLDT();
	Sort intSort=intLDT.targetSort();
	UseInsteadMap uim = new UseInsteadMap(NAT_SORT_NAME.toString(), 
					      intSort.name().toString());	
	try {
	    InputStream inp = file.getNewStream();
	    CountingBufferedInputStream cinp = 
		new CountingBufferedInputStream
		    (inp,main,getNumberOfChars()/100);
	    ProblemParser problemParser = 
		new ProblemParser(new KeYLexer(cinp),
				  initConfig.javaInfo().copy(), 
				  initConfig.namespaces().copy(),
				  uim);
	    problemParser.problem();
	    anySort = (Sort)(problemParser.getSorts().lookup(ANY_SORT_NAME));
	    Debug.assert(anySort!=null);
	    SetOfTaclet newTaclets
		=SortInstantiator.addSortRules(problemParser.getTaclets(), 
					       initConfig.sortNS(),	
					       problemParser.getFunctions(),   
					       anySort);
	    //???	    initConfig.add(newNamespaces);
	    if (mod.modifyTaclets()) initConfig.add(newTaclets);
	} catch (Exception e) {
	    System.err.println(e);
	    e.printStackTrace();
	    throw new RuntimeException("Failed to read "+name()+": "
				       +e.getMessage());
	}
    }



}
