// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.SkolemDeliverer;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.MapFromSchemaVariableToTacletPrefix;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;

/** class implements Schematic Theory Specific Rules  (Taclets) that
 * work on the antecedent.
 */
public class AntecTaclet extends FindTaclet{

     /**
     * creates a Schematic Theory Specific Rule (Taclet) with the given
     * parameters.  
     * @param name the name of the Taclet 
     * @param applPart contains the application part of an Taclet that is
     * the if-sequence, the variable conditions
     * @param goalTemplates a list of goal descriptions.
     * @param heuristics a list of heurisics for the Taclet
     * @param constraint the Constraint under which the Taclet is valid
     * @param attrs attributes for the Taclet; these are boolean values
     * indicating a noninteractive or recursive use of the Taclet. 
     * @param find the find term of the Taclet
     * @param prefixMap a MapFromSchemaVariableToTacletPrefix that contains the
     * prefix for each SchemaVariable in the Taclet
     */
    public AntecTaclet(Name name, TacletApplPart applPart,  
		     ListOfTacletGoalTemplate goalTemplates, 
		     ListOfHeuristic heuristics,
		     Constraint constraint,
		     TacletAttributes attrs,
		     Term find, MapFromSchemaVariableToTacletPrefix prefixMap){
	super(name, applPart, goalTemplates, heuristics, constraint, 
	      attrs, find, prefixMap);
	cacheMatchInfo();
    }


    /**
     * computes the formula capturing the meaning of the if-Part of this rule
     */
    protected Term calculate_b_IF(){
	// returns F_neg(ifseq) \/ F_neg( findseq ) where findseq = (this.find() =>)
	TermFactory tf = TermFactory.DEFAULT;
	Term result = ifseq().getAsNegativeFormula();
	Term secondFormula = tf.createJunctorTerm(Op.NOT,find()); // == F_neg( findseq ) !
	result = tf.createJunctorTermAndSimplify(Op.OR,result,secondFormula);
	return result;
    
    }
    
   /**
     * computes the formula capturing the meaning of the replacewith-Part 
     * of the i-th
     * goal template of this rule where 'rwexp' is the rwexp - term or 
     * sequent of the i-th goal template.
     */
    protected Term calculate_b_RW(Object rwexp){
	// returns 'true' ...
	TermFactory tf = TermFactory.DEFAULT;
	return(tf.createJunctorTerm(Op.TRUE));
    }
    
    /**
     * computes the formula capturing the meaning of the add-Part of the i-th
     * goal template of this rule where 'addseq' is the addseq - Sequent 
     * and 'rwexp' is the
     * rwexp - term or sequent (if one exists) of the i-th goal template.
     */
   
    protected Term calculate_b_A(Sequent addseq, Object rw){
	Debug.assert( rw instanceof Sequent, "RWEXP in current Taclet has to be a sequent");
	Sequent rwexp = (Sequent) rw;
	TermFactory tf = TermFactory.DEFAULT;

	// returns F_pos(addseq) /\ F_pos(rwexp);
	Term result = tf.createJunctorTermAndSimplify(Op.AND,rwexp.getAsPositiveFormula(),
						      addseq.getAsPositiveFormula());

	return result;
    }

    
    /** looks if a schema variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, addexp and rwexp)
     * used by lookupInRule)
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInIfAddRwExp(SchemaVariable var) { 
	if (!ifseq().lookup(var)) {
	    IteratorOfTacletGoalTemplate it = goalTemplates().iterator();
	    TacletGoalTemplate tpl;
	    while (it.hasNext()) {
		tpl=it.next();		
		if (tpl.seq().lookup(var) ||
		    (tpl instanceof AntecSuccTacletGoalTemplate) && 
		     ((AntecSuccTacletGoalTemplate)tpl).replaceWith()
		    .lookup(var)) {
		    return true;
		}
	    }   
	    return false;	    
	}
	return true;	    
    }

    /** this method is used to determine if top level updates are
     * allowed to be ignored. This is the case if we have an Antec or
     * SuccTaclet but not for a RewriteTaclet
     * @return true if top level updates shall be ignored 
     */
    protected boolean ignoreTopLevelUpdates() {
	return true;
    }


    /** CONSTRAINT NOT USED 
     * applies the replacewith part of Taclets
     * @param gt TacletGoalTemplate used to get the replaceexpression in the Taclet
     * @param goal the Goal where the rule is applied
     * @param posOfFind the PosInOccurrence belonging to the find expression
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations, so that the find term matches 
     */
    protected void applyReplacewith(TacletGoalTemplate gt, Goal goal,
				    PosInOccurrence posOfFind,
				    JavaInfo javaInfo, 
				    SVInstantiations svInst) {
	if (gt instanceof AntecSuccTacletGoalTemplate) {
	    Semisequent ante  = goal.seq().antecedent();
	    Semisequent succ  = goal.seq().succedent();
	    Sequent replWith = ((AntecSuccTacletGoalTemplate)gt).replaceWith();
	    applyAdd(replWith, goal, 
		     posOfFind, javaInfo, svInst);	   
	    if (replWith == Sequent.EMPTY_SEQUENT || 
		!(goal.seq().succedent() == succ && 
		  goal.seq().antecedent() == ante)) {
		goal.removeFormula(posOfFind);
	    }
	} else {
	    // Then there was no replacewith...
	}
    }

    /**
     * adds the sequent of the add part of the Taclet to the goal sequent
     * @param add the Sequent to be added
     * @param goal the Goal to be updated
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param posOfFind the PosInOccurrence describes the place where to add
     * the semisequent 
     * @param svInst the SVInstantiations to be used to instantiate the Taclet
     */
    protected void applyAdd(Sequent add, Goal goal,
			    PosInOccurrence posOfFind,
			    JavaInfo javaInfo,
			    SVInstantiations svInst) {
	addToAntec(add.antecedent(), goal, posOfFind, javaInfo, svInst);
	addToSucc(add.succedent(), goal, null, javaInfo, svInst);
    }
        
    /** toString for the find part */
    String toStringFind() {
	return "find " +find()+"==> ";
    }

    protected Taclet setName(String s) {
	AntecTacletBuilder b=new AntecTacletBuilder();
	b.setFind(find());
	return super.setName(s, b);
    }


}
