// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.SkolemDeliverer;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;

/** class implements Schematic Theory Specific Rules  (Taclets) that
 * represents a rewrite rule.
 */
public class RewriteTaclet extends FindTaclet{

     /**
     * creates a Schematic Theory Specific Rule (Taclet) with the given
     * parameters that represents a rewrite rule.  
     * @param name the name of the Taclet 
     * @param applPart contains the application part of an Taclet that is
     * the if-sequence, the variable conditions
     * @param goalTemplates a list of goal descriptions.
     * @param heuristics a list of heurisics for the Taclet
     * @param constraint the Constraint under which the Taclet is valid
     * @param attrs attributes for the Taclet; these are boolean values
     * indicating a noninteractive or recursive use of the Taclet. 
     * @param find the find term of the Taclet
     * @param prefixMap a MapFromSchemaVariableToTacletPrefix that contains the
     * prefix for each SchemaVariable in the Taclet
     */
    public RewriteTaclet(Name name, TacletApplPart applPart,  
		       ListOfTacletGoalTemplate goalTemplates, 
		       ListOfHeuristic heuristics,
		       Constraint constraint,
		       TacletAttributes attrs,
		       Term find, MapFromSchemaVariableToTacletPrefix prefixMap){
	super(name, applPart, goalTemplates, heuristics, constraint,
	      attrs, find, prefixMap);
	cacheMatchInfo();
    }	

      
    /** this method is used to determine if top level updates are
     * allowed to be ignored. This is the case if we have an Antec or
     * SuccTaclet but not for a RewriteTaclet
     * @return true if top level updates shall be ignored 
     */
    protected boolean ignoreTopLevelUpdates() {
	return false;
    }

    /**
     * computes the formula capturing the meaning of the if-Part of this rule.
     */
    protected Term calculate_b_IF(){
	// returns F_neg(ifseq) ...
	return (ifseq().getAsNegativeFormula());
    }
    
    /**
     * computes the formula capturing the meaning of the replacewith-Part of the i-th
     * goal template of this rule where 'rwexp' is the rwexp - term or sequent of the i-th goal template.
     */
    protected Term calculate_b_RW(Object rw){
	Debug.assert( rw instanceof Term, "RWEXP in current Taclet has to be a term (or formula)");
	Term rwexp = (Term) rw;
	// returns (findexp = rwexp) bzw. (findexp <-> rwexp) ...
	Equality eqv = rwexp.sort().getEqualitySymbol();
	TermFactory tf = TermFactory.DEFAULT;
	return(tf.createEqualityTerm(eqv,find(),rwexp));
    }
    
    /**
     * computes the formula capturing the meaning of the add-Part of the i-th
     * goal template of this rule where 'addseq' is the addseq - Sequent and 'rwexp' is the
     * rwexp - term (if one exists) of the i-th goal template.
     */
   
    protected Term calculate_b_A(Sequent addseq, Object rwexp){
	    // returns F_pos(addseq) ...
	    return(addseq.getAsPositiveFormula());
    }
    
    /** looks if a schema variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, addexp and rwexp)
     * used by lookup)
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInIfAddRwExp(SchemaVariable var) { 
	if (!ifseq().lookup(var)) {
	    IteratorOfTacletGoalTemplate it=goalTemplates().iterator();
	    TacletGoalTemplate tpl;
	    while (it.hasNext()) {
		tpl=it.next();		
		if (tpl.seq().lookup(var)
		    || (tpl instanceof RewriteTacletGoalTemplate 
			&& 
			lookupInTerm(var,
				     ((RewriteTacletGoalTemplate)tpl).replaceWith()))		    
		    ) {

		    return true;


		} 
	    }
	    return false;
	}
	return true;	    
    }


    /** looks if a schema variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, and all addseq )
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInAreaOne(SchemaVariable var) { 
	if (!ifseq().lookup(var)) {
	    IteratorOfTacletGoalTemplate it=goalTemplates().iterator();
	    while (it.hasNext()) {				
		if (it.next().seq().lookup(var)) {  
		    return true;
		} 
	    }
	    return false;
	}
	return true;	    
    }

    /** checks if a schema variable occurrs maximal in one instantiation area
     * @param var the SchemaVariable whose instantiation is checked
     * @return true iff instantiation is correct
     */
    protected boolean correctInstantiationAreaOccurrence(SchemaVariable var) {
	Boolean correct = areaCache.get(var);
	if (correct != null) {
	    return correct.booleanValue();
	} else if (var.isVariableSV() || varDeclaredNew(var)!=null) {
	    return super.correctInstantiationAreaOccurrence(var);
	}	
	// var is not new and findeseq is term or formula
	boolean found = lookupInAreaOne(var);
	boolean foundInFind = lookupInTerm(var,find());

	IteratorOfTacletGoalTemplate tplIt = goalTemplates().iterator();
	while (tplIt.hasNext()) {
	    TacletGoalTemplate tpl=tplIt.next();
	    if ((tpl instanceof RewriteTacletGoalTemplate) &&
		(!find().equals(((RewriteTacletGoalTemplate)tpl).replaceWith()))) {  
		if ( foundInFind || 
		     lookupInTerm(var,
			  ((RewriteTacletGoalTemplate)tpl).replaceWith())) {
		    if (found) {
			areaCache.put(var, new Boolean(false));
			return false; 
		    } else {
			found=true;
		    }
		}
	    }
	}

	if (!lookupInAddRulesArea(var,found)) { // does not occur
		areaCache.put(var, new Boolean(true));		
		return true;
	}

	areaCache.put(var, new Boolean(false));
	return false;
    }

    /**
     * does the work for applyReplacewith (wraps recursion) 
     */
    private Term replace(Term term, Term with, IteratorOfInteger it,
			 JavaInfo javaInfo,
			 SVInstantiations svInst) {	
	if (it.hasNext()) {	    
	    int sub=it.next().intValue();
	    Term[] subs=new Term[term.arity()];
	    for (int i=0;i<term.arity();i++) {
		if (i!=sub) {
		    subs[i]=term.sub(i);
		} else {
		    subs[i]=replace(term.sub(i), with, it, javaInfo, svInst);
		}
	    }
	    
	    ArrayOfQuantifiableVariable origvars 
		= (term.varsBoundHere(0).size()==0)
		? term.varsBoundHere(1) 
		: term.varsBoundHere(0);
	    QuantifiableVariable[] vars 
		= new QuantifiableVariable[origvars.size()];
	    for (int i=0; i<origvars.size(); i++) {
		vars[i] = origvars.getQuantifiableVariable(i);
	    }	    	    
	    return TermFactory.DEFAULT.createTerm(term.op(), subs, vars,
						  term.javaBlock(), 
						  term.ruleVar());	    
	} 
	return syntacticalReplace(with, javaInfo, svInst);
    }


    /** CONSTRAINT NOT USED 
     * applies the replacewith part of Taclets
     * @param gt TacletGoalTemplate used to get the replaceexpression in the Taclet
     * @param goal the Goal where the rule is applied
     * @param posOfFind the PosInOccurrence belonging to the find expression
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations
     */
    protected void applyReplacewith(TacletGoalTemplate gt, Goal goal,
				    PosInOccurrence posOfFind,
				    JavaInfo javaInfo,
				    SVInstantiations svInst) {
	if ( gt instanceof RewriteTacletGoalTemplate ) {
	    Term term = posOfFind.constrainedFormula().formula();
	    IteratorOfInteger it = posOfFind.posInTerm().iterator();
	    Term rwTemplate=((RewriteTacletGoalTemplate)gt).replaceWith();
	    Term result = replace(term, rwTemplate, it, javaInfo, svInst);
 	    UpdateSimplifier simplifier = new UpdateSimplifier();
	    int simpDepth=UpdateSimplifier.depthWithoutUpdates(rwTemplate);
	    result = simplifier.simplify(result, simpDepth);
	    goal.changeFormula(new ConstrainedFormula(result, constraint()), 
			       posOfFind);
	} else {
	    // Then there was no replacewith...
	    // This is strange in a RewriteTaclet, but who knows...
	}
    }

    /**
     * adds the sequent of the add part of the Taclet to the goal sequent
     * @param add the Sequent to be added
     * @param goal the Goal to be updated
     * @param posOfFind the PosInOccurrence describes the place where to add
     * the semisequent
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations
     */
    protected void applyAdd(Sequent add, Goal goal,
			    PosInOccurrence posOfFind,
			    JavaInfo javaInfo,
			    SVInstantiations svInst) {
	if (posOfFind.isInAntec()) {
	    addToAntec(add.antecedent(), goal, posOfFind, javaInfo, svInst);
	    addToSucc(add.succedent(), goal, null, javaInfo, svInst);
	} else {
	    addToAntec(add.antecedent(), goal, null, javaInfo, svInst);	
	    addToSucc(add.succedent(), goal, posOfFind, javaInfo, svInst);
	}
    }

    protected Taclet setName(String s) {
	RewriteTacletBuilder b=new RewriteTacletBuilder();
	b.setFind(find());
	return super.setName(s, b);
    }


}
