// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


/**
 * visitor for <t> execPostOrder </t> of 
 * {@link de.uka.ilkd.key.logic.Term}. Called with that method
 * on a term, the visitor builds a new term replacing SchemaVariables with their
 * instantiations that are given as a SVInstantiations object.
 */
package de.uka.ilkd.key.rule;

import java.util.Stack;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.java.visitor.ContextAddVisitor;
import de.uka.ilkd.key.java.visitor.ProgramReplaceVisitor;
import de.uka.ilkd.key.logic.Visitor;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.ClashFreeSubst;


public class SyntacticalReplaceVisitor extends Visitor{	

    private SVInstantiations svInst;
    private JavaInfo javaInfo;
    private Term computedResult = null;
    private TypeConverter typeConverter = null;

    /**
     * the stack contains the subterms that will be added in the next step of
     * execPostOrder in Term in order to build the new term. A boolean value
     * between or under the subterms on the stack indicate that a term using
     * these subterms should build a new term instead of using the old one,
     * because one of its subterms has been built, too.
     */
    private Stack subStack; //of Term (and Boolean)
    private TermFactory tf = TermFactory.DEFAULT;
    private Boolean newMarker = new Boolean(true);

    public SyntacticalReplaceVisitor(JavaInfo javaInfo, 
				     SVInstantiations svInst) { 
	this.javaInfo = javaInfo;	
	this.svInst = svInst;
	this.typeConverter = javaInfo.getTypeConverter();
	subStack = new Stack(); // of Term
    }

    private JavaProgramElement addContext(StatementBlock pe) {
	ContextInstantiationEntry cie=svInst.getContextInstantiation();
	if (cie == null) {
	    throw new IllegalStateException("Context should also "
					    +"be instantiated");
	}	
	if (cie.prefixDepth() > 0) {
	    ContextAddVisitor contextAdder 
		= new ContextAddVisitor(cie.prefix(),
					cie.prefixDepth(), cie.contextWidth(),
					cie.contextProgram(), pe);
	    
	    contextAdder.start();
	    return (JavaProgramElement) contextAdder.result();
	} else {
	    return pe;
	}
    }

    private JavaBlock replacePrg(SVInstantiations svInst, JavaBlock jb) {
	ProgramReplaceVisitor trans;
	JavaProgramElement result = null;
	if (jb.program() instanceof ContextTerm) {
	    trans = new ProgramReplaceVisitor
		(((ContextTerm)jb.program()).getChildAt(0), javaInfo, svInst);
	    trans.start();
	    result = (JavaProgramElement)trans.result();
	    result=addContext((StatementBlock)result);
	} else {
	    trans = new ProgramReplaceVisitor(jb.program(), javaInfo, svInst);
	    trans.start();
	    result = (JavaProgramElement)trans.result();
	}
	return (result==jb.program()) ? jb : JavaBlock.createJavaBlock(result);
    }

    private Term[] neededSubs(int n) {
	boolean newTerm = false;
	Term[] result = new Term[n];
	for (int i = n-1; i >= 0; i--) {
	    Object top=subStack.pop();
	    if (top == newMarker){
		newTerm=true; 
		top=subStack.pop();
	    }
	    result[i] = (Term) top;
	}
	if (newTerm && (subStack.empty() || 
			subStack.peek() != newMarker) ) {
	    subStack.push(newMarker);
	}
	return result;
    }

    private Term peekSub() {
	for (int i=subStack.size()-1; i>=0; i--) {
	    if (subStack.elementAt(i)!=newMarker) {
		return (Term) subStack.elementAt(i);
	    }
	}
	return null;
    }

    private void pushNew(Object t) {
	if (subStack.empty() || subStack.peek() != newMarker) {
	    subStack.push(newMarker);
	}
	subStack.push(t);	
    }

    private Term toTerm(Object o) {
	if (o instanceof Term) {
	    return (Term)o;
	} else if (o instanceof ProgramElement) {
	    return typeConverter.convertToLogicElement
		((ProgramElement)o);
	}
	de.uka.ilkd.key.util.Debug.assert(false, 
					  "Wrong instantiation in SRVisitor");
	return null;
    }

    public void visit(Term visited) {
	// Sort equality has to be ensured before calling this method
	if (visited.op() instanceof SchemaVariable &&
	    svInst.isInstantiated((SchemaVariable)visited.op())) {
	    pushNew(toTerm(svInst.getInstantiation
			   ((SchemaVariable)visited.op())));
	} else if (visited.op() instanceof UpdateOp) {
	    UpdateOp update = (UpdateOp)visited.op();
	    UpdateOp newUpdateOp = update;
	    Operator var = update.variable().op();

	    if (var instanceof AttributeOp) {
		Debug.out("syntacticalreplacementvisitor: WARNING: "+
			  "attribute op handling needs to be implemented.");
	    } else if (var instanceof SchemaVariable &&
		svInst.isInstantiated((SchemaVariable)var)) {
		Object inst = svInst.getInstantiation((SchemaVariable) var);
		if (inst instanceof ProgramElement) {
		    pushNew(TermFactory.DEFAULT.createUpdateTerm
			    (typeConverter.convertToLogicElement
			     ((ProgramElement)inst), neededSubs(2)));
		} else {
		    pushNew(TermFactory.DEFAULT.createUpdateTerm
			    ((Term)inst, neededSubs(2)));
		}
	    } else {
		subStack.push(visited);
	    }
	} else if (visited.op() instanceof ExpressionOperator) {
	    ExpressionOperator exprOp = (ExpressionOperator) visited.op();
	    pushNew(exprOp.resolveExpression(svInst, javaInfo));	    
	} else {
	    Operator op=visited.op();
	    if (op instanceof AttributeOp) {
		AttributeOp ao=(AttributeOp)op;
		if (ao.attribute() instanceof SchemaVariable) {
		    if (svInst.isInstantiated((SchemaVariable)ao.attribute())) {
			Term t = toTerm(svInst.getInstantiation
					((SchemaVariable)ao.attribute()));
			op=new AttributeOp((ProgramVariable)t.op(), 
					   peekSub().sort());
		    }
		}
	    }
       
	    boolean jblockChanged=false;
	    JavaBlock jb=visited.javaBlock();
	    if (jb != JavaBlock.EMPTY_JAVABLOCK) {
		jb = replacePrg(svInst, jb);
		if (jb!=visited.javaBlock()) {
		    jblockChanged=true;
		}
	    }

	    boolean varsChanged=false;
	    
	    ArrayOfQuantifiableVariable boundVars 
		= ((visited.varsBoundHere(0).size() == 0) ?
		   visited.varsBoundHere(1) :
		   visited.varsBoundHere(0)); 
	    	    
	    QuantifiableVariable[] vars = 
		new QuantifiableVariable[boundVars.size()];

	    for (int i = 0; i < vars.length; i++) {
		if (boundVars.getQuantifiableVariable(i) 
		    instanceof SchemaVariable) {
		    vars[i] = (QuantifiableVariable)
			((Term)svInst.getInstantiation
			 ((SchemaVariable)boundVars.
			  getQuantifiableVariable(i))).op();
		    varsChanged = true;
		} else {
		    vars[i] = boundVars.getQuantifiableVariable(i);
		}
	    }

	    Term[] neededsubs = neededSubs(visited.arity());
	    if (varsChanged || jblockChanged || 
		(!subStack.empty() &&  subStack.peek() == newMarker)) {
		pushNew(resolveSubst(tf.createTerm(op, 
						   neededsubs,
						   vars,
						   jb,
						   visited.ruleVar())));
	    } else {
		subStack.push(resolveSubst(visited));
	    }		
	}
    }

    private Term resolveSubst(Term t) {
	if (t.op() instanceof SubstOp) {
	    QuantifiableVariable v=t.varsBoundHere(1).getQuantifiableVariable(0);
	    ClashFreeSubst cfSubst = new ClashFreeSubst(v, t.sub(0));
	    return cfSubst.apply(t.sub(1));	    
	}
	return t;
    }


    /**
     * delivers the new built term
     */
    public Term getTerm() {
	if (computedResult==null) {
	    Object o=null;
	    do {
		o=subStack.pop();
	    } while (o==newMarker);
	    Term t = (Term) o;
// 	    CollisionDeletingSubstitutionTermApplier substVisit
// 		= new CollisionDeletingSubstitutionTermApplier();
// 	    t.execPostOrder(substVisit);	    
// 	    t=substVisit.getTerm();	
	    computedResult=t;
//  	    UpdateSimplifier simplifier=new UpdateSimplifier();
//  	    computedResult=simplifier.simplify(t);	    
	}	
	return computedResult;
    }


    /**
     * this method is called in execPreOrder and execPostOrder in class Term
     * when leaving the subtree rooted in the term subtreeRoot.
     * Default implementation is to do nothing. Subclasses can
     * override this method 
     * when the visitor behaviour depends on informations bound to subtrees.
     * @param subtreeRoot root of the subtree which the visitor leaves.
     */

    public void subtreeLeft(Term subtreeRoot){
	if (subtreeRoot.op() instanceof MetaOperator) {
	    MetaOperator mop = (MetaOperator) subtreeRoot.op();
	    pushNew(mop.calculate((Term)subStack.pop(),svInst, javaInfo));	    
	} 

    }


}
