// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General
// Public License. See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.SkolemDeliverer;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.expression.Literal;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.JavaInfo;

import java.util.*;
import java.text.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.java.visitor.JavaIterator;

/** This class implements the Schematic Theory Specific Rules
 * (Taclets). This class encapsulates the structure of a taclet and offers
 * methods that work mainly and 
 * these structures like matching and instantiating the Taclet. 
 * It is neither used to look where the Taclet can be applied in a sequent nor to
 * create new goals etc. This has to be done by 
 * {@link de.uka.ilkd.key.proof.RuleAppIndex RuleAppIndex} or 
 * {@link de.uka.ilkd.key.proof.TacletAppIndex TacletAppIndex} or 
 * {@link de.uka.ilkd.key.rule.TacletApp TacletApp}
*/

// There used to be methods in this class to calculate proof obligations
// from taclets.  Since v0.82 however, the logic data structures do not
// match those methods.  They can be retreived from PRCS version 0.81 of
// Feb 5th 2001.


public abstract class Taclet implements Rule, Named {

    public static final String PROOFOBL_PREFIX = "_proofobl(";
    public static final String PROOFOBL_SUFFIX = ")_";

    private final String AUTONAME = "taclet";
    private static int tacletNumber = 0;

    private Name name;
    private String displayName = null;
    private String helpText = null;

    /**
     * @label ifSequent
     */
    private Sequent ifseq;
    private ListOfNewVarcond varsNew;
    private ListOfNotFreeIn varsNotFreeIn;
    private boolean recursive;
    /**
     * @link aggregation
     * @supplierCardinality *
     */
    /*#private TacletGoalTemplate lnkTacletGoalTemplate;*/
    private ListOfTacletGoalTemplate goalTemplates;
    /**
     * @link aggregation
     * @supplierCardinality *
     * @label Rule Heuristics
     */
    /*#private Heuristic lnkHeuristic;*/
    protected ListOfHeuristic heuristics;
    private boolean noninteractive;
    /** constraint under which the Taclet is valid */
    private Constraint constraint;
    /** caches variable to variable occurs in less than two
     * instantiation areas*/
    protected HashMapFromSchemaVariableToBoolean areaCache;

    protected MapFromSchemaVariableToTacletPrefix prefixMap;
    
    /** for a variable pair (x,y) in this list: x occurs in the scope
     * of a variable y in this Taclet. At least the pair (y,y) occurs if
     * variable y is bound in the Taclet, i.e. the ys are the bound
     * variables in the Taclet.
     */
    protected SetOfInScopePair varInScope
	=SetAsListOfInScopePair.EMPTY_SET;


    protected SetOfQuantifiableVariable boundVars
	=SetAsListOfQuantifiableVariable.EMPTY_SET;

    private boolean contextInfoComputed=false;
    private boolean contextIsInPrefix=false;
     
    protected String tacletAsString;

    /** creates an empty Taclet */
    Taclet() {
	name = new Name("");
	ifseq = Sequent.EMPTY_SEQUENT;
	varsNew=SLListOfNewVarcond.EMPTY_LIST;
	varsNotFreeIn=SLListOfNotFreeIn.EMPTY_LIST;
	recursive=false;
	goalTemplates=SLListOfTacletGoalTemplate.EMPTY_LIST;
	heuristics=SLListOfHeuristic.EMPTY_LIST;
	noninteractive=false;	
	constraint=Constraint.BOTTOM;
	areaCache=new HashMapFromSchemaVariableToBoolean();
	prefixMap=MapAsListFromSchemaVariableToTacletPrefix.EMPTY_MAP;
    }

    /**
     * creates a Schematic Theory Specific Rule (Taclet) with the given
     * parameters.  
     * @param name the name of the Taclet 
     * @param applPart contains the application part of an Taclet that is
     * the if-sequence, the variable conditions
     * @param goalTemplates a list of goal descriptions.
     * @param heuristics a list of heurisics for the Taclet
     * @param constraint the Constraint under which the Taclet is valid
     * @param attrs attributes for the Taclet; these are boolean values
     * indicating a noninteractive or recursive use of the Taclet. 
     *
     */
    Taclet(Name name, TacletApplPart applPart,  
	 ListOfTacletGoalTemplate goalTemplates, 
	 ListOfHeuristic heuristics,
	 Constraint constraint, TacletAttributes attrs,
	 MapFromSchemaVariableToTacletPrefix prefixMap
	 ){

	this.name          = name;
	ifseq              = applPart.ifseq();
	varsNew            = applPart.varsNew();
	varsNotFreeIn      = applPart.varsNotFreeIn();
	this.goalTemplates = goalTemplates;
	this.heuristics    = heuristics;
	recursive          = attrs.recursive();
	noninteractive     = attrs.noninteractive();
	this.constraint    = constraint;
	areaCache          = 
	    new HashMapFromSchemaVariableToBoolean();
	this.prefixMap     = prefixMap;
        this.displayName   = attrs.displayName();
        if (this.displayName == null)  this.displayName = name.toString();
    }

    protected void cacheMatchInfo() {
	boundVars = SetAsListOfQuantifiableVariable.EMPTY_SET;
	varInScope = getVarInScope();
	IteratorOfInScopePair it=varInScope.iterator();
	while (it.hasNext()) {
	    boundVars = 
		boundVars.add((QuantifiableVariable)it.next().second());
	}

    }

    /** looks for a Schemavariable var in Term term
     * @param var the SchemaVariable to be looked for
     * @param term the Term to be looked through
     * @return true iff variable has been found in the term 
     */    
    protected boolean lookupInTerm(SchemaVariable var, Term term) {
	TacletVariableCollector vcv=new TacletVariableCollector();
	term.execPostOrder(vcv);
	return vcv.contains(var);
    }
    
    /** checks if a given variables is used in the rule this implies
     * in the ifseq, findexp, rwexp, addseq or addrule
     * @param var the Variable to be looked for
     * @return true iff Variable var occurs in the Taclet
     */
    protected abstract boolean lookup(SchemaVariable var);

    /** partial implementation of lookup. But lookup
     * must also check the findexp for FindTaclets. 
     * checks if a given variables is used in the rule this implies
     * the ifseq, addseq, rwexp and addrule. 
     * @param var the Variable to be looked for
     * @return true iff Variable var occurs in the Taclet
     */    
    protected boolean lookupImpl(SchemaVariable var) {
	if (!lookupInIfAddRwExp(var)) {
	    IteratorOfTacletGoalTemplate tplIt = goalTemplates().iterator();
	    while (tplIt.hasNext()) {
		IteratorOfTaclet ruleIt=tplIt.next().rules().iterator();
		while (ruleIt.hasNext()) {
		    if (ruleIt.next().lookup(var)) {
			return true;
		    }
		}    
	    }
	    return false;
	}
	return true;
    }	


    /** looks if a schema variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, addexp and rwexp, rwexp only if not RewriteTaclet
     * (overwrites this method)).
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInAreaOne(SchemaVariable var) {
	return lookupInIfAddRwExp(var);
    }


    /** looks if a variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, addexp and rwexp)
     * used by lookup)
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected abstract boolean lookupInIfAddRwExp(SchemaVariable var);
    

    /** checks if a variable occurs in
     * the addrules part of the Taclet. The method returns true if the
     * variable is found in two areas (not necessarily in two addrule
     * areas)
     * @param var the SchemaVariable to be looked for
     * @param foundBefore a boolean that has to be set true iff the
     * variable to be looked for has already been found in a former area
     * @return true iff. variable is found in more than two areas
     */
    protected boolean lookupInAddRulesArea(SchemaVariable var, 
					   boolean foundBefore) {
	boolean found = foundBefore;
	IteratorOfTacletGoalTemplate
	    goalTplIt=goalTemplates().iterator();
	while (goalTplIt.hasNext()) {
	    IteratorOfTaclet tacletIt=goalTplIt.next().rules().iterator();
		while (tacletIt.hasNext()) {
		    Taclet s=tacletIt.next();
		    if (s.lookup(var)) {
			if (found) { 
			    return true;
			}
			found = true;
		    }
		}
	}
	return false;
    }

    /** checks if a variable occurrs maximal in one instantiation area
     * @param var the SchemaVariable whose instantiation is checked
     * @return true iff instantiation is correct
     */
    protected abstract boolean
	correctInstantiationAreaOccurrence(SchemaVariable var); 

    /**
     * looks if a variable is declared as not free in
     * @param var the SchemaVariable to look for
     * @return true iff declared not free
     */
    private boolean varDeclaredNotFree(SchemaVariable var) {
	IteratorOfNotFreeIn it = varsNotFreeIn();
	while (it.hasNext()) {
	    if (it.next().first() == var) {
		return true;
	    }
	}
	return false;
    }

    /**
     * looks if a variable is declared as new and returns its sort to match
     * with or the schema variable it shares the match-sort with. Returns
     * null if the SV is not declared to as new.
     * @param var the SchemaVariable to look for
     * @return the sort of the SV to match or the SV it shares the same
     * match-sort with
     */
    public Object varDeclaredNew(SchemaVariable var) {
	IteratorOfNewVarcond it=varsNew.iterator();
	while (it.hasNext()) {
	    NewVarcond nv=it.next();
	    if (nv.getSchemaVariable()==var) {
		return nv.getSortDefiningObject();
	    }
	}
	return null;
    }

    /**
     * returns true iff the given variable is bound either in the
     * ifseq or in 
     * any part of the TacletGoalTemplates
     * @param v the bound variable to be searched 
     */
    protected boolean varIsBound(SchemaVariable v) {
	return (v instanceof QuantifiableVariable) 
	    ? boundVars.contains((QuantifiableVariable)v)
	    : false;
    }

    /**
     * returns pairs of variables (x,y) where y is bound in this Taclet and x
     * occurs in the scope of y; (y,y) is part of the set for all bound ys
     * with no x bound in the scope of y.
     * @return a set of variable pairs as specified above
     */
    protected SetOfInScopePair getVarInScope() {
	IteratorOfTacletGoalTemplate it=goalTemplates().iterator();
	SetOfInScopePair result
	    = SetAsListOfInScopePair.EMPTY_SET;
	while (it.hasNext()) {
	    result = result.union(it.next().getVarInScope());
	}
	return result.union(ifseq.getVarInScope());	
    }

    /** checks if the program element can potentially belong to the
     * prefix
     * @param pe the ProgramElement to check
     * @return true iff prefix compatible
     */
    public static boolean isPrefix(ProgramElement pe) {
	if (pe instanceof LabeledStatement || 
	    pe instanceof StatementBlock ||
	    pe instanceof Try || 
	    pe instanceof MethodCallStatement) {
	    return true;
	}
	return false;
    }

    /** returns a list of program elements whose first element is the
     * statement that is not in the prefix. Sideeffect: the position
     * of the given iterator will be the first element after the prefix as
     * result (or null if no such exists). The iterator will return 'null' in
     * cases like this: <{ {{{{}}}} }>  in cases like <{ { {} int i = 0; } }>
     * the iterator will return '{}'.
     * @param it the JavaIterator to use
     * @return a list of program elements who build the prefix      
     */
    private ListOfProgramElement getPrefix(JavaIterator it) {
	ListOfProgramElement prefixList = SLListOfProgramElement.EMPTY_LIST;

	ProgramElement lastElement = null;	
	ProgramElement el = it.element();	
	boolean labeled         = false;	
	boolean was_empty_block = false;
	boolean is_empty_block  = false;

	while (it.hasNext()) {
	    // the empty block thing catches the following to
	    // situations in case of {} stmnt; 
	    // the last thing not in prefix is the closing brace '}'
	    // in a situation like l1:{} we want to only l1:{ to be
	    // once inside the prefix 
	    if (isEmptyBlock(el)) {
		is_empty_block = true;		
	    } else {
		is_empty_block = false;
	    }

	    Debug.assert(el != null, "JavaIterator must have a bug.");

	    if (labeled && is_empty_block) {
		// we have a labeled empty block
		// next() method is called the empty block will be returned
		labeled = false;
		return prefixList;
	    } else if (is_empty_block) {
		return prefixList.prepend(el);
	    } else if (!isPrefix(el)) {		
		// Debug.out("taclet:    ProgramElement "+el+" is not prefix compatible.");
		labeled = false;
 		if (was_empty_block) {
 		    // the last thing not in prefix is the closing
 		    // brace '}' so set back twice		    
 		    it.back();
 		}
		return prefixList;
	    } else {		
		if (el instanceof StatementBlock) {
		    if (!labeled || 
			lastElement instanceof LabeledStatement) {
			prefixList = prefixList.prepend(el);
		    }
		} else {
		    prefixList = prefixList.prepend(el);
		}
		labeled = false;
	    }
	    if (el instanceof MethodCallStatement) {
		labeled = true;
		if (((MethodCallStatement)el).getProgramVariable() != null) {
		    // skip result var
		    it.next();
		}
	    } else if (el instanceof LabeledStatement) {
		labeled = true;
		// skip name
		it.next();	       
	    }
	    if (el instanceof Try) {
		labeled = true;
		// do not skip block because it may be the empty one
	    }
	    was_empty_block = is_empty_block;
	    lastElement = el;
	    el = it.next();
	}	
	return prefixList;
    }

    /** returns true iff the program element is StatementBlock with no
     * children 
     * @return true iff the program element is StatementBlock with no
     * children 
     */
    private boolean isEmptyBlock(ProgramElement pe) {
	if (pe != null && pe instanceof StatementBlock) {
	    if (((StatementBlock)pe).getChildCount() == 0) {
		return true;
	    }
	}
	return false;
    }

    /** returns the position in the program of the prefix if it can be
     * matched, null if no match has been found. Sideeffect: 
     * the iterator seqPE delivers after the next next() call the
     * first element that has to be matched 
     * @param seqPE the JavaIterator through the program element
     * part of a the term in the sequent for which we look if this
     * Taclet fits
     * @param tacletPE the JavaIterator through the program element
     * that is part of the find- or if-section of the Taclet
     * @return the position in the program of the prefix if it can be
     * matched, null if no match has been found
     */
    private PosInProgram matchPrefix( JavaIterator seqPE,
				      JavaIterator tacletPE ) {
	PosInProgram prefixPos = null;
      	ListOfProgramElement prefixList = getPrefix(seqPE); 
	// list of elements at the beginning of the program elements
	// of the Taclet (normally inside a context term) that are
	// potential prefix elements but that are not in the prefix
	// because the term in the find- or if-section of this sequent
	// that has to be matched starts with context compatible elements.
	// As every context term starts with a statement block as
	// wrapper, the wrapper appears in the prefix list and has to
	// be ignored. As we are only interested in the length of the
	// list, this is done by substracting 1 from the lists size.	
	int notInPrefixDepth = getPrefix(tacletPE).size() - 1;
	if (notInPrefixDepth > prefixList.size()) {
	    // the term we try to match with the taclet has no context
	    // compatible elements but some are required by the taclet
	    // term, so a matching is impossible and we return null
 	    return null;
 	} else if (notInPrefixDepth != 0) {
	    // we have now a maximal prefix. Now we have to go backwards
	    // so that 'notInPrefixDepth' elements are left that have to
	    // be checked if they match the taclet ones.
	    Debug.out("taclet: Prefixelements to match(nr):", notInPrefixDepth);
	    ProgramElement seqPEback = seqPE.element();
	    for (int i = 0; i < notInPrefixDepth; i++) { 		
		if (!isEmptyBlock(seqPEback)) {		   
		    seqPEback = seqPE.back();
		    if (seqPEback instanceof StatementBlock) {
			// take try { and methodcall..:{ as one
			seqPEback = seqPE.back();
			if (!(seqPEback instanceof Try ||
			    seqPEback instanceof ProgramVariable ||
			      seqPEback instanceof MethodCallStatement)) {
			    seqPEback = seqPE.next();
			} 
		    }
		} else {
		    // skip labeled empty block as we skip it when
		    // calculating the prefix 
		    ProgramElement lookAhead = seqPE.back();		    
		    if (lookAhead instanceof Label || 
			lookAhead instanceof Try || 
			lookAhead instanceof MethodCallStatement) {
			seqPEback = lookAhead;
		    } else if (lookAhead instanceof ProgramVariable) {
			// can happen if we have a methodcall construct
			ProgramElement lookAhead2 = seqPE.back();
			if (lookAhead2 instanceof MethodCallStatement) {
			    seqPEback = lookAhead2;
			} else {
			    seqPE.next();
			    seqPE.next();			    
			}		    
		    } else {
			seqPE.next();
		    }
		}
		if (seqPEback != null) {
		    if (seqPEback instanceof Label) {
			seqPEback = seqPE.back();		
			if (!(seqPEback instanceof LabeledStatement)) {
			    return null;
			} 
		    } else if (seqPEback instanceof ProgramVariable) {
			seqPEback = seqPE.back();
			if (!(seqPEback instanceof MethodCallStatement)) {
			    return null;
			} 			
		    }
		}
	    }
	}
	// get position of the prefix
	ProgramElement last = seqPE.element();
	if (last == null) {
 	    // at the end of the program ? can happen 
	    // e.g. if <{ {{{{{}}}}} }> has to be matched
	    // as all program elements disappear in the prefix
	    last = seqPE.back();
 	}
	// catch the case that we have a labeled empty block in these
	// case we want to have returned the labeled block and not
	// just the empty one
	if (isEmptyBlock(last)) {
	    last = seqPE.back();
	    if (last instanceof Label || 
		last instanceof ProgramVariable) {
		last = seqPE.back(); // back to labeled/method call statement
	    } else if (!(last instanceof Try || 
			 last instanceof MethodCallStatement)) {
		last = seqPE.next(); // no labeled or try block 
	    }
	}
	return seqPE.position();
    }



    /** this method is responsible for recognizing closing braces, as
     * we match a tree using a kind of iterator that works in a depth
     * first search mode, it is not this trivial to decide whether a
     * match has to fail because the pattern has a closing brace but
     * the program not, e.g. the template
     * <code> 
     * { {return #se;} #s  } 
     * </code> 
     * where <code>#se</code> is the schemavariable for simple
     * expression and <code>#s</code> a schemavariable for a
     * statemen. We have to avoid that this matches to a program like
     * <code> 
     *   { return 2; int i = 0; }
     * </code>
     *  that could easily happen because after matching
     * <code>#se </code> to <code> 2 </code> the iterator returns as
     * next element <code> int i = 0; </code> that matches with 
     * <code> #s </code> returned by the iterator of the template tree.
     */
    private boolean matchClosingBraces(PosInProgram prefixPos,
				       PosInProgram programPos,
				       PosInProgram lastProgramPos,		       
				       PosInProgram templatePos,
				       PosInProgram lastTemplatePos) {

	if (programPos == null) {
	    // nothing to check here
	    return true;
	}
	/* If the depth of the template is null, then the end of the
	   template that has to be matched is reached. As there is an
	   enclosing wrapper around the template the depth of it is
	   equal to one and not as one would expect to zero 
	 */
	int templateDepth = (templatePos == null ? 
			     1 : templatePos.depth()); 

	int programDepth  = programPos.depth();

	/* if the first element that does not belong to the prefix is
	   at a deeper position than the element of the program that
	   has to be matched, then the match fails because we have
	   left the part of the program to be matched and entered
	   parts of the enclosing context, e.g. 
	   <code> <{ {{int i = 0; }} int j=2;  }> </code>
	   is tried to be matched by
	   <code> <{ pi #t #v = #se; #s omega }> </code>
	   The schemavariable <code> #s </code> would be instantiated by
	   <code> int j=2; </code> but int j = 2; is in the context.
	*/	
	if (templatePos != null && prefixPos.depth() > programDepth) {
	    Debug.out("taclet: MATCH FAILED. Entered context but there "+
		      "are templates to be matched.");
	    return false; // FAILED
	}

	/* We calculate the difference of the depth of the position of
	   the last template element ans the current element that has
	   to be matched next. The difference is a measure for the
	   number of enclosing braces. (Same for the program element)
	*/
	int templateDiff = (templateDepth - lastTemplatePos.depth());	
	int matchDiff = (programDepth - lastProgramPos.depth());
	
	if (matchDiff > templateDiff) {
	    Debug.out("taclet: MATCH FAILED. One has left the "+
		      "container the other not.");
	    Debug.out("taclet: Program position  (current, last): ", 
		      programPos, lastProgramPos);
	    Debug.out("taclet: Template position (current, last):" , 
		      templatePos, lastTemplatePos);
	    return false;
	}
	return true;
    }


    /** returns true if the given SchemaVariable can stand for the
     * ProgramElement 
     * @param sv the SchemaVariable
     * @param match the ProgramElement to be matched
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return true if the SchemaVariable canstand for the given element
     */
    private boolean check(SchemaVariable sv, ProgramElement match,
			  JavaInfo javaInfo) {
	Debug.out("taclet: Check: SchemaVariable (name, sort): ",sv, sv.sort());
	Debug.out("taclet: With: ", match.getClass());
	if (match == null) {
	    Debug.out("taclet: MATCH FAILED. 'null' cannot be "+
		      "matched to a SchemaVariable.");
	    return false;
	}
	if (!((ProgramSVSort)sv.sort()).canStandFor(match, javaInfo)) {
	    Debug.out("taclet: MATCH FAILED. Sort mismatch.", match.getClass());	
	    return false; // FAILED
	}
	return true;
    }

    /** the following method is used to match schemavariables that can
     * stand for lists of programelements, e.g. list of statement or
     * catch-finally expressions
     * @param schemavar the SchemaVariable tried to be instantiated
     * @param programIterator the iterator through the program to be
     * matched
     * @param instantiations the SVInstantiations that has been found
     * until now
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return SVInstantiations with the new found matchings or null
     * if match failed
     */
    private SVInstantiations
	matchProgramElementList(SchemaVariable schemavar,
				JavaIterator programIterator,
				SVInstantiations instantiations,
				JavaInfo javaInfo) {
	SVInstantiations result = instantiations;
	// match all statements on this level
	// if the schemavariable has been matched we need to
	// check if the matching is equal
	ArrayOfProgramElement plist = (ArrayOfProgramElement)
	    result.getInstantiation(schemavar);
	ProgramElement program = programIterator.element();
	NonTerminalProgramElement parent =
	    (NonTerminalProgramElement) programIterator.parent();
	// length of statements to be matched is number of
	// child - current position
	int childCount = parent.getChildCount();
	int childPos   = parent.getIndexOfChild(programIterator.element());
	int size       = childCount-childPos;
	int count      = 0;
	boolean hasNextSibling = false;
	ProgramElement[] stmntList = new ProgramElement[(parent instanceof For ? size-1 : size)];
	do {
	    if (plist != null) {
		// schemavar has been matched before so we just
		// have to test if the instantiations are
		// equal
		if (count >= plist.size() ||
		    !(plist.getProgramElement(count).equals(program))) {
		    Debug.out("taclet: The SchemaVariable has been matched "+
			      "before with a different block");
		    return null; // FAILED
		}
	    } else if (!check(schemavar, program, javaInfo)) {
		return null; // FAILED
	    }
	    stmntList[count] = program;
	    hasNextSibling   = programIterator.hasNextSibling();
	    program          = programIterator.nextSibling();
            hasNextSibling   = hasNextSibling && (!(parent instanceof For) || programIterator.hasNextSibling());
            count ++;
	} while (hasNextSibling);
	if (!(parent instanceof For)) {
           Debug.assert(size == count);
        } else {
           Debug.assert(size-1 == count);
           programIterator.back();
        }
	if (plist == null) {
	    Debug.out("taclet: Match Success. Length of list ", stmntList.length);
	    result = result.add(schemavar, new ArrayOfProgramElement(stmntList));
	}
	return result;
    }

    /** this method checks the size of NonTerminalProgramElements and
     * tries to figure out if a match is possible. As for some
     * constructs (Try and Statements) a list matching is possible
     * the check cannot be restricted to size equality if one of these
     * constructs is to be matched. For other elements the stronger
     * check MUST be done.
     * @param template the NonTerminalProgramElement that is
     * required by this Taclet
     * @param program the NonTerminalProgramElement tried to be
     * matched
     * @return true iff the block size may be compatible
     */
    private boolean checkBlockSize(NonTerminalProgramElement template,
				   NonTerminalProgramElement program) {
	/*
	  if (a) template.size() > program.size() ---> FAIL
	     (b) template.size() == 0 && program.size() != 0
	                                            --> FAIL
	  otherwise even if they are unequal but not zero we
	  must look inside the blocks because of a possible
	  statmentlist schemavariable that can match 1..*
	  statements.
	  But if template is not a StatementBlock or Try the stronger
	  check(unequality) has to be performed
	*/
	int tplSize     = template.getChildCount();
	int programSize = program.getChildCount();
	if (!(template instanceof StatementBlock || template instanceof Try)) {
	    if (tplSize != programSize) {
		Debug.out("taclet: MATCH FAILED. Incompatible sizes "
			  +" of non terminal program elments");
		return false; //FAILED. Size incompatible
	    }
	} else if ( tplSize > programSize ||
		    (tplSize == 0 && programSize != 0)) {
	    // the second condition is because lsit matching must
	    // match at least one element
	    Debug.out("taclet: MATCH FAILED. Elements have incompatible size.");
	    return false; //FAILED. Size incompatible
	}
	return true;
    }


    /** matches java block when the taclet term is a context term
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     */
    protected SVInstantiations matchJavaProgram(JavaBlock block,
						ContextTerm ctxt,
						SVInstantiations inst,
						JavaInfo javaInfo) {
	Debug.out("taclet: Starting program match of rule: ", this);
	Debug.out("taclet: Match (program, template):", block.program(), ctxt.getBody());
	
	JavaIterator programIterator = new JavaIterator(block.program());
	JavaIterator ctxtIt = new JavaIterator(ctxt.getBody());
	PosInProgram prefixPos = matchPrefix(programIterator, ctxtIt);

	if (prefixPos == null) {
	    Debug.out("taclet: Failed to match prefix.");
	    return null;  //FAILED prefix matching
	}

	JavaIterator templateIt = new JavaIterator(ctxt.getBody());
	// skip wrapping block
	templateIt.next();

	return matchJavaProgram(block, programIterator,
				prefixPos, templateIt,
				inst, javaInfo);
    }

    /** matches the java block after the prefix has been matched or if 
     * no prefix is available this method is called directly
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     */
    protected SVInstantiations matchJavaProgram(JavaBlock programBlock, 
                                                JavaIterator programIterator, 
                                                PosInProgram prefixPos,
	                                        JavaIterator templateIt, 
                                                SVInstantiations inst, 
                                                JavaInfo javaInfo) {
	SVInstantiations result = inst;
	// the position before the last iteration
	ProgramElement program = programIterator.element();
	PosInProgram lastProgramPos = prefixPos;

	ProgramElement template =  templateIt.element();
	PosInProgram lastTemplatePos = templateIt.position();

	do {
	    if (program == null && template != null &&
		template instanceof StatementBlock) {
		// may be the empty block match
		break;
	    } else if (program == null || template == null) {
		return null;
	    }			
	    Debug.out("taclet: Match (program, class) ", program, program.getClass());
	    Debug.out("taclet: with (template, class) ", template, template.getClass());
	    if (template instanceof SchemaVariable) {
		SchemaVariable sv=(SchemaVariable)template;
		if (sv.sort() == ProgramSVSort.STATEMENTLIST ||
                    sv.sort() == ProgramSVSort.EXPRESSIONLIST ||
		    sv.sort() == ProgramSVSort.CATCHFINALLY) {
		    result = matchProgramElementList(sv, programIterator,
						     result, javaInfo);
		    if (result == null) { return null; } //FAILED
		} else { // schemavariables that are not matching on lists
		    if (!check(sv, program, javaInfo)) {
			Debug.out("taclet: MATCH FAILED. Sort of SchemaVariable cannot "+
				  "stand for the program");
			return null; //FAILED
		    }
		    Object instant=result.getInstantiation(sv);
		    if ( instant == null ||
			 instant.equals(program) ||
			 ((instant instanceof Term) 
			      && ((Term)instant).op().equals(program))) {
			Debug.out("taclet: Add instantiation (sv, program): ",sv, program);
			result = result.add(sv, program);
		    } else {
			Debug.out("taclet: MATCH FAILED 3. Former match of "+
				  " SchemaVariable incompatible with "+
				  " the current match.");
			return null; //FAILED mismatch
		    }
		}
	    } else if (template.getClass() != program.getClass()) {
		// if this match does not fail we can ssume in the
		// following cases that the type of the template and
		// the program to be matched are equal. Allows to
		// restrict instanceof checking to the template.
		Debug.out("taclet: MATCH FAILED. Different classes.");
		return null;  //FAILED
 	    } else if (template instanceof NonTerminalProgramElement) {
		if (!checkBlockSize
		    ((NonTerminalProgramElement)template,
		     (NonTerminalProgramElement)program)) {
		    return null; //FAILED
		}
 	    } else if (template instanceof Literal ||
		       template instanceof ProgramElementName) {
		if (!template.equals(program)) {
		    Debug.out("taclet: MATCH FAILED. Literal/Name mismatch.");
		    return null;  //FAILED
		}
	    }
	    lastProgramPos = programIterator.position();
	    // if the template is a schemavariable it has matched
	    // at least at the current subtree, so we have to skip
	    // it and not to descend into it.
	    program = (template instanceof SchemaVariable ?
		       programIterator.skip() : programIterator.next() );
	    lastTemplatePos = templateIt.position();
	    template = templateIt.next();
	    Debug.out("taclet: Next position is ", programIterator.position());	
	    if (!matchClosingBraces(prefixPos, 
				    programIterator.position(), 
				    lastProgramPos,
				    templateIt.position(),
				    lastTemplatePos)) {
		return null; //FAILED
	    }

	} while (programIterator.hasNext() && templateIt.hasNext());
	if (templateIt.hasNext()) {
 	    Debug.out("taclet: MATCH FAILED. Template has more unmatched elements.");
	    return null;
	} else {    
	    result = result.add(prefixPos, lastProgramPos, programBlock.program());
	    Debug.out("taclet: :-))) RULE MATCHED :-)))");
	    Debug.out("taclet: (EndPos, result)", lastProgramPos, result);
	    return result;
	}
    }

		
    /**
     * returns a SVInstantiations object iff the given Term
     * template can be instantiated to 
     * match the given Term term using the known instantiations stored in
     * svInst.  If a
     * matching cannot be found null is returned.
     * The not free in condition is checked in TacletApp. Collisions are resolved
     * there as well, if necessary.
     * @param term the Term that has to be matched
     * @param template the Term that is checked if it can match term
     * @param ignoreUpdates a boolean if set to true updates will be ignored as 
     * e.g. wanted if an if-sequent is matched
     * @param svInst the SVInstantiations that are required because of formerly
     * matchings 
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return the new SVInstantiations needed to match template with term , if
     * possible, null otherwise
     */
    protected SVInstantiations match(Term term, 
				     Term template, 
				     boolean ignoreUpdates,
				     SVInstantiations svInst,
				     JavaInfo javaInfo) {       
	return matchHelp(term, template, ignoreUpdates, svInst, javaInfo);
    }


    /**
     * same as the method above but with ignoreUpdates always false
     * @param term the Term that has to be matched
     * @param template the Term that is checked if it can match term
     * @param svInst the SVInstantiations that are required because of formerly
     * matchings 
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return the new SVInstantiations needed to match template with term , if
     * possible, null otherwise
     */
    protected SVInstantiations match(Term term, 
				     Term template, 
				     SVInstantiations svInst,
				     JavaInfo javaInfo) {
	return match(term, template, false, svInst, javaInfo);
    }



    /**
     * checks if the conditions for a correct instantiation are fulfilled
     * @param var the template Variable to be instantiated
     * @param term the Term to be mapped to var
     * @param svInst the SVInstantiations that are already known to be needed 
     * @param boundVars the ListOfQuantifiableVariable in whose
     * bounding area var is  
     * @return true iff conditions are fulfilled
     */
    private boolean checkVariableConditions(SchemaVariable var, Term term,
					    SVInstantiations svInst) {

// 	IteratorOfSchemaVariable varIt = varsNew();
// 	while (varIt.hasNext()) {
// 	    if (var==varIt.next()) {
// 		return false; // var new condition violated
// 	    }
// 	}
	if (!(term.op() instanceof QuantifiableVariable)) {
	    if (var.isVariableSV() || varIsBound(var) ||
		varDeclaredNotFree(var) ||
		(term.freeVars() != SetAsListOfQuantifiableVariable.EMPTY_SET 
		 && !correctInstantiationAreaOccurrence(var))
		) {
		// match(x) is not a variable, but the
		// corresponding template variable is bound
		// or declared non free (so it has to be
		// matched to a variable) or match(x) contains
		// free variables and the corresponding
		// template variable occurs in more than one
		// instantiation area			
		return false;
	    }
	} else {
	    if (varIsBound(var)) {
		IteratorOfInScopePair boundInIt = varInScope.iterator();
		while (boundInIt.hasNext()) {
		    InScopePair varpair = boundInIt.next();
		    if (varpair.second() == var) {
			if (term.equals(svInst.getInstantiation
					((SchemaVariable)varpair.first()))) {

			    return false;			    
			}
		    }
		}
	    }
	    if (!correctInstantiationAreaOccurrence(var)) {
		return false;

	    }
	}		

	return true;	
    }

    private SVInstantiations addInstantiation(Term term, SchemaVariable sv, 
					      SVInstantiations newSVInst) {   
	Term t = (Term) newSVInst.getInstantiation(sv);	   		
	if (t != null) {
	    if (!t.equalsModRenaming(term)) {
		return null; //FAILED;
	    } else {
		return newSVInst;
	    }
	}	
	// no former matching found
	if (!checkVariableConditions(sv,
				     term, 
				     newSVInst)) {
	    return null; //FAILED;
	}
	return newSVInst.add(sv, term);	   	    
    }
    
    /** returns a SVInstantiations object with the needed SchemaVariable to Term
     * mappings to match the given Term template to the Term term or
     * null if no matching is possible.
     * @param term the Term the Template should match
     * @param template the Term tried to be instantiated so that it matches term
     * @param ignoreUpdates a boolean if set to true updates will be ignored as 
     * e.g. wanted if an if-sequent is matched
     * @param svInst the SVInstantiations already known and required 
     * @return the SVInstantiations needed or null if no match is possible
     */
    private SVInstantiations matchHelp(Term term, Term template, 
				       boolean ignoreUpdates,
				       SVInstantiations svInst,
				       JavaInfo javaInfo) {
	SVInstantiations newSVInst = svInst;
	if (template.op() instanceof SchemaVariable) {
	    SchemaVariable sv=(SchemaVariable)template.op();
	    if (sv.isProgramSV()) {
		if (((ProgramSVSort)sv.sort()).canStandFor(term)) {
		    return addInstantiation(term, sv, newSVInst);
		} else {

		    return null; //FAILED
		}
	    } else if (!template.sort().extendsTrans(term.sort())) {
		return null; //FAILED
	    } else {
		return addInstantiation(term, sv, newSVInst);
	    }
	} else {
	    if (template.op() instanceof UpdateOp) {
		if (!(term.op() instanceof UpdateOp)) {
		    return null; //FAILED
		}
		// match updates
		UpdateOp templateOp = (UpdateOp) template.op();
		UpdateOp termOp = (UpdateOp) term.op();
		
		// variable matching
		if (templateOp.variable().op() instanceof SchemaVariable) {
		   
		    Term t = (Term) newSVInst.getInstantiation
			((SchemaVariable) templateOp.variable().op());
		    
		    if (t != null) {
			if (!(t.op().equals(termOp.variable().op()))) {
			    return null; //FAILED
			} 		
		    } else {
			newSVInst = newSVInst.add
			    ((SchemaVariable) templateOp.variable().op(),			     
			     termOp.variable());
		    }  
		} else {
		    if (!(templateOp.variable().op().equals(termOp.variable().op()))) {
			return null; //FAILED
		    }
		}
		// update matching
//  		newSVInst = matchHelp(termOp.update(), 
//  				      templateOp.update(),
//  				      newSVInst);
// 		if (newSVInst==null) {
// 		    return null; //FAILED
// 		}
// 		return matchHelp(term.sub(0), template.sub(0), 
// 				  newSVInst);
	    } else if (ignoreUpdates && term.op() instanceof UpdateOp) {
		// updates can be ignored 		    
		svInst = svInst.addUpdate((UpdateOp)term.op(), term.sub(0));
		return matchHelp(term.sub(1), template, true, svInst, javaInfo);
	    } else if (term.op() != template.op()) {	       
		if ((term.op() instanceof CondTermOp) 
		      && term.op().getClass()==template.op().getClass()) {
		    AttributeOp aop=((CondTermOp)term.op()).attribute();
		    ProgramVariable att=(ProgramVariable)aop.attribute();
		    CondTermOp templateOp=(CondTermOp)template.op();   
		    newSVInst=addInstantiation
			(TermFactory.DEFAULT.createVariableTerm(att), 
			 (SchemaVariable) templateOp.attribute().attribute(), 
			 newSVInst);
		    if (newSVInst==null) {
			return null; //FAILED
		    }
		} else {
		    return null; //FAILED;		
		}
	    }	    
    
	    //match java blocks:
	    if (term.javaBlock().isEmpty()) {
		if (!template.javaBlock().isEmpty()){
		    return null; //FAILED
		}
	    } else { //both javablocks!=null
		if (!(template.javaBlock().program() instanceof ContextTerm)) {	 
		    newSVInst = matchJavaProgram
			(term.javaBlock(), 
			 new JavaIterator(term.javaBlock().program()), 
			 PosInProgram.TOP,  
			 new JavaIterator(template.javaBlock().program()),
			 newSVInst, javaInfo);
		} else {
		    newSVInst = matchJavaProgram(term.javaBlock(),
						 (ContextTerm)
						 template.javaBlock().program(),  
						 newSVInst, javaInfo);
		}
		if (newSVInst == null) {
		    return null; //FAILED
		}
	    }

	    //match bound variables:
	    for (int j=0; j<term.arity(); j++) {		

		ArrayOfQuantifiableVariable bound = term.varsBoundHere(j);
		ArrayOfQuantifiableVariable tplBound=template.varsBoundHere(j); 
		
		if (bound.size() != tplBound.size()) {
		    return null; //FAILED
		}

		for (int i=0; i<bound.size(); i++) {		
		    if (tplBound.getQuantifiableVariable(i) 
			instanceof SchemaVariable) {			
			    if (!tplBound.getQuantifiableVariable(i)
				.sort().extendsTrans
				(bound.getQuantifiableVariable(i).sort())) {
			    return null;  //FAILED

			}
			Term foundMapping 
			    = (Term)newSVInst.getInstantiation
			    ((SchemaVariable)
			     tplBound.getQuantifiableVariable(i));
			if (foundMapping != null && 
			    ((QuantifiableVariable)foundMapping.op()) 
			    != bound.getQuantifiableVariable(i)) {
			    return null; //FAILED;			    

			} else if (foundMapping == null) {
			    Term varTerm
				= TermFactory.DEFAULT.createVariableTerm
				((LogicVariable)bound
				 .getQuantifiableVariable(i));
			    if (checkVariableConditions((SchemaVariable)
				       tplBound.getQuantifiableVariable(i),
							varTerm, newSVInst)) {
				newSVInst = newSVInst.add((SchemaVariable)tplBound
					      .getQuantifiableVariable(i), 
					      varTerm);
			    }
			} 
		    } else {
			if (tplBound.getQuantifiableVariable(i) 
			    != bound.getQuantifiableVariable(i)) {
			    return null; //FAILED;
			}
		    }		    
		}
	    }

	    for (int i = 0; i < term.arity(); i++) {
		newSVInst = match(term.sub(i), template.sub(i),
				  newSVInst, javaInfo);
		if (newSVInst == null) {
		    return null; //FAILED
		} 
	    }	    
	    return newSVInst;	
	}
    }


    /**
     * does the real work for matchIf from which it is called. First
     * with the antecedent, second with succedent of the ifseq
     * @param semiseq the Semisequent to be matched with the corressponding
     * if-semisequent 
     * @param inAntec a boolean , if true the antecdent is matched else the
     * succedent 
     * @param resList ListOfSVInstantiations with SVInstantiations of a former
     * matching (e.g. find-match or antecedent-match)
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of possible matches. Empty list if no match has
     * been found.
     */
    private ListOfSVInstantiations matchIfHelp(Semisequent semiseq,
					       boolean inAntec,
					       ListOfSVInstantiations
					       resList,
					       JavaInfo javaInfo) {

	ListOfSVInstantiations result = resList;	
	// get iterator of the semisequent that is considered
	IteratorOfConstrainedFormula ifSemiIt =
	    (inAntec ? ifseq().antecedent() :
	     ifseq().succedent()).iterator(); 
	while (ifSemiIt.hasNext()) {
	    // for each constraint formula of the if-sequent, 
	    // get iterator of the found mappings,
	    // delete old result list. This will not affect the iterator cause
	    // to immutable datastructures. So all mappings that fit will be put
	    // back to the list if they can match the if-sequent in a compatible
	    // way.
	    ConstrainedFormula cf = ifSemiIt.next();
	    IteratorOfSVInstantiations resultIt =
		result.iterator();
	    result = SLListOfSVInstantiations.EMPTY_LIST;
	    while (resultIt.hasNext()) {		
		IteratorOfConstrainedFormula semiIt = semiseq.iterator();
		SVInstantiations svInst = resultIt.next();
		while (semiIt.hasNext()) {		
		    SVInstantiations newSVInst = null;
		    if (svInst.getUpdateContext() == 
			SLListOfUpdatePair.EMPTY_LIST) {
			newSVInst = match(semiIt.next().formula(),
					  cf.formula(), false, svInst,
					  javaInfo);
		    } else {
			newSVInst = match
			    (semiIt.next().formula(),
			     TermFactory.DEFAULT.
			     createUpdateTerm(svInst.getUpdateContext(),
					      cf.formula()), false,
			     svInst, javaInfo);
		    }
		    if (newSVInst != null) {
			result = result.prepend(newSVInst);	    
		    }
		}
	    }
	}
	return result;	
    }

    /** same as method above but ignoreUpdates alsways false
     * @param term the Term the Template should match
     * @param template the Term tried to be instantiated so that it matches term
     * @param ignoreUpdates a boolean if set to true updates will be ignored as 
     * e.g. wanted if an if-sequent is matched
     * @param svInst the SVInstantiations already known and required 
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return the SVInstantiations needed or null if no match is possible
     */
    private SVInstantiations matchHelp(Term term, Term template, 
				       SVInstantiations svInst,
				       JavaInfo javaInfo) {
	return matchHelp(term, template, false, svInst, javaInfo);
    }  

    /**
     * matches if-sequent to the given sequent and returns the result 
     * of this trial 
     * @param toMatch the Sequent the if-sequent has to be matched to
     * @param svInst the SVInstantiations that have to be considered during the
     * match process 
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return a ListOfSVInstantiations with possible SVInstantiations, each of
     * them may be used to match the if sequent. The list is empty if no
     * matching has been found.
     */
    public ListOfSVInstantiations matchIf(Sequent toMatch,
					  SVInstantiations svInst,
					  JavaInfo javaInfo) {  
	ListOfSVInstantiations
	    resultList = SLListOfSVInstantiations.EMPTY_LIST.prepend(svInst);
	// match antecedent
	resultList = matchIfHelp(toMatch.antecedent(), true,
				 resultList, javaInfo);
	// match succedent
	return matchIfHelp(toMatch.succedent(), false, resultList, javaInfo);
    } 

    
    /** returns the name of the Taclet
     */
    public Name name() {
	return name;
    } 
    

    /** returns the display name of the taclet, or, if not specified -- 
     *  the canonical name
     */
    public String displayName() {
	return displayName;
    }
    
    
    public String helpText() {
       return helpText;
    }


   /** returns the if-sequence of the application part of the Taclet.
     */
    public Sequent ifseq() {
	return ifseq;
    } 
    
    /** returns an iterator over the variables that are new in the Taclet. 
     */
    public ListOfNewVarcond varsNew() {
	return varsNew;
    } 

    
    /** returns an iterator over the variable pairs that indicate that are 
     * new in the Taclet. 
     */
    public IteratorOfNotFreeIn varsNotFreeIn() { 
	return varsNotFreeIn.iterator();
    } 

    /** returns true iff the Taclet is to be applied recursive
     */
    public boolean recursive() {
	return recursive;
    } 
    
    /** returns the Constraint under which the Taclet is
     * valid */
    public Constraint constraint() {
	return constraint;
    }


    /** returns an iterator over the goal descriptions.
     */
    public ListOfTacletGoalTemplate goalTemplates() {
	return goalTemplates;
    } 

    /** returns an iterator over the heuristics. */
    public IteratorOfHeuristic heuristics() {
	return heuristics.iterator();
    } 

    /** returns true iff the Taclet is to be applied only noninteractive
     */
    public boolean noninteractive() {
	return noninteractive;
    }


    public MapFromSchemaVariableToTacletPrefix prefixMap() {
	return prefixMap;
    }

    public TacletPrefix getPrefix(SchemaVariable sv) {
	return prefixMap.get(sv);
    }

    /**
     * returns true iff a context flag is set in one of the entries in
     * the prefix map. Is cached after having been called
     * once. __OPTIMIZE__ is caching really necessary here?
     *
     * @return true iff a context flag is set in one of the entries in
     * the prefix map.
     */
    public boolean isContextInPrefix() {
	if (contextInfoComputed) {
	    return contextIsInPrefix;
	}
	contextInfoComputed=true;
	IteratorOfTacletPrefix it=prefixMap().valueIterator();
	while (it.hasNext()) {
	    if (it.next().context()) {
		contextIsInPrefix=true;
		return true;
	    }
	}
	contextIsInPrefix=false;
	return false;
    }

    public boolean equals(Object o) {
	if ( ! ( o instanceof Taclet ) )
	    return false;

	return name.equals(((Taclet) o).name);
    }


    /**
     * creates the initial instantiation of the SchemaVariables 
     * @return the initial SVInstantiations 
     */
    public SVInstantiations createInitialInstantiation() {
	return SVInstantiations.EMPTY_SVINSTANTIATIONS;
    }


    /** 
     * a new term is created by replacing variables of term whose replacement is
     * found in the given SVInstantiations 
     * @param term the Term the syntactical replacement is performed on
     * @param svInst the SVInstantiations containing the instantiations needed
     * for the replacement
     * @return the created Term 
     */
    protected Term syntacticalReplace(Term term,
				      JavaInfo javaInfo,
				      SVInstantiations svInst) {	

	//	Debug.out("taclet: SYNTACTICALREPLACE  "+ term+"  "+svInst);

	SyntacticalReplaceVisitor srVisitor=new	SyntacticalReplaceVisitor(javaInfo, svInst); 
	term.execPostOrder(srVisitor);
	return srVisitor.getTerm();
    }
    
    /**
     * adds ConstrainedFormula to antecedent or succedent depending on
     * position information or the boolean antec 
     * @param semi the Semisequent with the the ConstrainedFormulae to be added
     * @param goal the Goal that knows the node the formulae have to be added
     * @param pos the PosInOccurrence describing the place in the sequent
     * @param antec boolean true(false) if elements have to be added to the
     * antecedent(succedent) (only looked at if pos == null)
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations 
     */
    private void addToPos(Semisequent semi,
			  Goal goal,			  
			  PosInOccurrence pos,
			  boolean antec,
			  JavaInfo javaInfo, 
			  SVInstantiations svInst) 
    {
	IteratorOfConstrainedFormula it = semi.iterator();
	while (it.hasNext()) {
	    ConstrainedFormula cf = null;
	    if (svInst.getUpdateContext() == SLListOfUpdatePair.EMPTY_LIST) {
		cf = new ConstrainedFormula
		    (syntacticalReplace(it.next().formula(), 
					javaInfo,
					svInst),
		     constraint());
	    } else {
		Term formula=it.next().formula();
		Term replaced = syntacticalReplace(formula, javaInfo, svInst);
		UpdateSimplifier simplifier = new UpdateSimplifier();
		int simpDepth=UpdateSimplifier.depthWithoutUpdates(formula);
 		Term updated = TermFactory.DEFAULT
 		    .createUpdateTerm(svInst.getUpdateContext(), replaced);
		
 		cf = new ConstrainedFormula
 		    (simplifier.simplify(updated, simpDepth), constraint());

	    }
	    if (pos != null) {
		goal.addFormula(cf, pos);
	    } else {
		// cf : formula to be added , 1. true/false: antec/succ,
		// 2. true: at head 
		goal.addFormula(cf, antec, true);		
	    }	    
	}
    }

    /**
     * adds ConstrainedFormula to antecedent depending on
     * position information (if none is handed over it is added at the
     * head of the antecedent). Of course it has to be ensured that
     * the position information describes one occurrence in the
     * antecedent of the sequent.
     * @param semi the Semisequent with the the ConstrainedFormulae to be added
     * @param goal the Goal that knows the node the formulae have to be added
     * @param pos the PosInOccurrence describing the place in the
     * sequent or null for head of antecedent
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations 
     */
    protected void addToAntec(Semisequent semi,
			      Goal goal,
			      PosInOccurrence pos,
			      JavaInfo javaInfo,
			      SVInstantiations svInst) { 
	addToPos(semi, goal, pos, true, javaInfo, svInst);
    }

    /**
     * adds ConstrainedFormula to succedent depending on
     * position information (if none is handed over it is added at the
     * head of the succedent). Of course it has to be ensured that
     * the position information describes one occurrence in the
     * succedent of the sequent.
     * @param semi the Semisequent with the the ConstrainedFormulae to be added
     * @param goal the Goal that knows the node the formulae have to be added
     * @param pos the PosInOccurrence describing the place in the
     * sequent or null for head of antecedent
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations 
     */
    protected void addToSucc(Semisequent semi,
			     Goal goal,
			     PosInOccurrence pos,
			     JavaInfo javaInfo,
			     SVInstantiations svInst) {
	addToPos(semi, goal, pos, false, javaInfo, svInst);
    }

    protected abstract Taclet setName(String s);
    
    protected Taclet setName(String s, TacletBuilder b) {	
	b.setTacletGoalTemplates(goalTemplates());
	b.setName(new Name(s));
	b.setDisplayName(displayName());
	b.setIfseq(ifseq());
	b.setConstraint(constraint());
	b.addVarsNew(varsNew());
	b.addVarsNotFreeIn(varsNotFreeIn);
	return b.getTaclet();
    }

    /**
     * adds the given rules (i.e. the rules to add according to the Taclet goal
     * template to the node of the given goal
     * @param rules the rules to be added
     * @param goal the goal describing the node where the rules should be added
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations of the SchemaVariables
     */
    protected void applyAddrule(ListOfTaclet rules, Goal goal, 
				JavaInfo javaInfo,
				SVInstantiations svInst) {
                                
	IteratorOfTaclet it = rules.iterator();
	while (it.hasNext()) {
	    Taclet tacletToAdd = it.next(); 
	    String uniqueTail=""; // we need to name the new taclet uniquely

            TacletGoalTemplate replacewithCandidate = null;
	    IteratorOfTacletGoalTemplate actions = 
               tacletToAdd.goalTemplates().iterator();
            while (actions.hasNext()) {
               replacewithCandidate = actions.next();
               if (replacewithCandidate instanceof RewriteTacletGoalTemplate)
                  break;
            }
            if ((replacewithCandidate instanceof RewriteTacletGoalTemplate) & 
                (tacletToAdd instanceof FindTaclet)) {
                // we have _both_ FIND and REPLACEWITH
                Term find = ((FindTaclet)tacletToAdd).find();
                Term replwith = 
                   ((RewriteTacletGoalTemplate)
                      replacewithCandidate).replaceWith();
                      
                SyntacticalReplaceVisitor visitor = // now instantiate them!
                   new SyntacticalReplaceVisitor(javaInfo, svInst);
                visitor.visit(find);
                uniqueTail = "_" + visitor.getTerm();
                visitor = new SyntacticalReplaceVisitor(javaInfo, svInst);
                visitor.visit(replwith);
                uniqueTail += "_" + visitor.getTerm();
	    }

            if ("".equals(uniqueTail)) { // otherwise just number it
               uniqueTail = "_"+AUTONAME+(tacletNumber++);
            }

            tacletToAdd=tacletToAdd.setName(tacletToAdd.name()+uniqueTail);


	    // the new Taclet may contain variables with a known
	    // instantiation. These must be used by the new Taclet and all
	    // further rules it contains in the addrules-sections. Therefore all
	    // appearing (including the addrules) SchemaVariables have to be
	    // collected, then it is looked if an instantiation is known and if
	    // positive the instantiation is memorized. At last the Taclet with
	    // its required instantiations is handed over to the goal, where a
	    // new TacletApp should be built including the necessary instantiation
	    // information

	    SVInstantiations neededInstances = SVInstantiations.
		EMPTY_SVINSTANTIATIONS.addUpdateList
		(svInst.getUpdateContext());
	    TacletVariableCollector collector = new
		TacletVariableCollector(); 
	    collector.visit(tacletToAdd, true);// true, because
	                                     // descend into
					     // addrules
	    IteratorOfSchemaVariable svIt = collector.varIterator();
	    while (svIt.hasNext()) {
		SchemaVariable sv = svIt.next();
		if (svInst.isInstantiated(sv)) {
		    neededInstances = neededInstances.add
			(sv, svInst.getInstantiationEntry(sv));
		} 
	    }
	    goal.addTaclet(tacletToAdd, neededInstances);
	}
    }

    protected void applyAddProgVars(SetOfSchemaVariable pvs, Goal goal,
				    SVInstantiations svInst) {
	IteratorOfSchemaVariable it = pvs.iterator();
	while (it.hasNext()) {
	    ProgramVariable inst
		= (ProgramVariable)svInst.getInstantiation(it.next());
	    goal.addProgramVariable(inst);
	}
    }
    
    /**
     * adds the goal to prove that the if sequent can be derived and
     * therefore the Taclet can be applied
     * @param goal the Goal to be updated
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the SVInstantiations with the neccessary 
     * instantiations of the SchemaVariables
     */
    private void applyAddIfGoal(Goal goal,
				JavaInfo javaInfo,
				SVInstantiations svInst) {	
	
	addToSucc(ifseq().antecedent(), goal, null, javaInfo, svInst);
	addToAntec(ifseq().succedent(), goal, null, javaInfo, svInst);
    }


    /** 
     * adds the if conditions to the proof sequent
     * @param goal the Goal to be updated
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the SVInstantiations with the neccessary 
     * instantiations of the SchemaVariables
     */
    private void addIfCondition(Goal goal,
				JavaInfo javaInfo,
				SVInstantiations svInst) {
	addToAntec(ifseq().antecedent(), goal, null, javaInfo, svInst); 
	addToSucc(ifseq().succedent(), goal, null, javaInfo, svInst); 
    }

    /** creates two new goals. The first one is the modified
     * original Taclet, so that the if condition of the Taclet is fulfilled
     * The second one is the goal needed to prove that the if-sequent holds.
     * @return array of Goals the first entry is the sequent for the
     * if sequent, the second the modified original goal on which the
     * Taclet is applied 
     */
    protected Goal[] createIfGoals(Goal goal,
				   JavaInfo javaInfo,
				   SVInstantiations svInst) 
    {
	Goal[] result = new Goal[2];
	ListOfGoal goals = goal.split(2);	    	
	result[0] = goals.head();
	result[1] = goals.tail().head();
	// adds if condition to original goal so that the taclet can
	// be applied
	addIfCondition(result[0], javaInfo, svInst);
	// creates goal to prove the if sequent 
	applyAddIfGoal(result[1], javaInfo, svInst);
	return result;
    }


    /**
     * the rule is applied on the given goal using the
     * information of rule application. A goal for the if sequent is
     * not added. The method acts the same way as an invocation of 
     * apply(goal, ruleApp, false) would do
     * @param goal the goal that the rule application should refer to.
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param startApp the rule application that is executed.
     */
    public ListOfGoal apply(Goal goal, JavaInfo javaInfo,
			    RuleApp tacletApp) {
	ListOfGoal r= apply(goal, javaInfo, tacletApp, false);
	return r;
    }


    /** the rule is applied to the given goal using the
     * information of rule application. 
     * @param goal the goal that the rule application should refer to.
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param tacletApp the rule application that is executed.
     * @param proveIfSequent a boolean that is true if a goal has to
     * be created in order to prove that the sequent in the if part of
     * the Taclet can be derived and therefore the application of the
     * rule is correct (normally true when the if-sequent does not match)
     * @return list of the goals created by the rule which have to
     * be proved
     */
    public abstract ListOfGoal apply(Goal goal, JavaInfo javaInfo, 
				     RuleApp tacletApp, boolean proveIfSequent);


    /**
     * computes the formula capturing the meaning of the if-Part of this rule
     */
    protected abstract Term calculate_b_IF();
    
    /**
     * computes the formula capturing the meaning of the replacewith-Part of the i-th
     * goal template of this rule where 'rwexp' is the rwexp - term or sequent of the i-th goal template.
     */
    protected abstract Term calculate_b_RW(Object rwexp);
    
    /**
     * computes the formula capturing the meaning of the add-Part of the i-th
     * goal template of this rule where 'addseq' is the addseq - Sequent and 'rwexp' is the
     * rwexp - term or sequent (if one exists) of the i-th goal template.
     */
   
    protected abstract Term calculate_b_A(Sequent addseq, Object rwexp);
    /**
     * computes the formula capturing the meaning of the addrules-Part of the i-th
     * goal template of this rule where 'addedRules' are the added concrete rule of
     * the i-th goal template and the 'addedRuleVariables' are the added rule variables in the
     * corresponding addrules-part of the i-th goal template.
     */
    /*    protected Term calculate_b_AR(ListOfTaclet addedRules, ListOfRulevariable addedRulevariables){
	// The implementation was done in this class because the algorithm is the same for
	// all types of Taclet (subclasses of Taclet);

	//     returns the conjunctive formula :
	//          b_po(ar1) /\ b_po(ar2) /\ .... (conrete rules) ... 
	//          ... /\ proofobl(arv1) /\ proofobl(arv2) /\ ... (rule vars) ...

	
	
	TermFactory tf = TermFactory.DEFAULT;
	
	if (addedRules.size() + addedRulevariables.size() == 0){
	    return (tf.createJunctorTerm(Op.TRUE));
	} else {
	    IteratorOfTaclet rulesIterator = addedRules.iterator();
	    Taclet currentRule;
	    Term firstArg = null;
	    Term secondArg = null;
	    if (rulesIterator.hasNext()) {
		firstArg = rulesIterator.next().calculate_b_PO(); // calculate meaning formula b_po ...
		while(rulesIterator.hasNext()){
		    secondArg = rulesIterator.next().calculate_b_PO();
		    // append conjunctive ...
		    firstArg = tf.createJunctorTerm(Op.AND,firstArg,secondArg);
		}
	    }
	    Rulevariable rv;
	    IteratorOfRulevariable ruleVarsIterator = addedRulevariables.iterator();
	    if (firstArg == null) { // there has to be at least one rule variable ...
	        rv = ruleVarsIterator.next();
		// create a free formula variable instead of proofobl-terms ...
		firstArg = tf.createProofObligationTerm(rv);
	    }
	    while(ruleVarsIterator.hasNext()){
		rv = ruleVarsIterator.next();
		// create a free formula variable instead of proofobl-terms ...
		secondArg = tf.createProofObligationTerm(rv);
		// append conjunctive ...
		firstArg = tf.createJunctorTerm(Op.AND,firstArg,secondArg);
	    }

	    return firstArg;
	} 
	
	
	}*/



    StringBuffer toStringIf(StringBuffer sb) {
	if (!ifseq.empty()) {
	    sb = sb.append("if (").append(ifseq()).append(") ");
	    sb = sb.append("\n");
	}
	return sb;
    }

    StringBuffer toStringVarCond(StringBuffer sb) {
	IteratorOfNewVarcond itVarsNew=varsNew().iterator();
	IteratorOfNotFreeIn itVarsNotFreeIn=varsNotFreeIn();
	if (itVarsNew.hasNext() || itVarsNotFreeIn.hasNext()) {
	    sb=sb.append("varcond(");
	    while (itVarsNew.hasNext()) {
		sb=sb.append(itVarsNew.next());
		if (itVarsNew.hasNext() || itVarsNotFreeIn.hasNext())
		    sb=sb.append(", "); 
	    }
	    while (itVarsNotFreeIn.hasNext()) {
		NotFreeIn pair=itVarsNotFreeIn.next();
		sb=sb.append(pair.first());
		sb=sb.append(" not free in ").append(pair.second());	 
		if (itVarsNotFreeIn.hasNext()) sb=sb.append(", ");
	    }	    
	    sb=sb.append(")\n");	    
	}
	return sb;
    }

    StringBuffer toStringGoalTemplates(StringBuffer sb) {
	IteratorOfTacletGoalTemplate itGoalTempl=goalTemplates().iterator();
	while (itGoalTempl.hasNext()) {
	    sb=sb.append(itGoalTempl.next());
	    if (itGoalTempl.hasNext()) sb = sb.append(";");
	    sb = sb.append("\n");
	}
	return sb;
    }

    StringBuffer toStringHeuristics(StringBuffer sb) {
	IteratorOfHeuristic itHeu=heuristics();
	if (itHeu.hasNext()) {
	    sb=sb.append("heuristics(");
	    while (itHeu.hasNext()) {
		sb=sb.append(itHeu.next());
		if (itHeu.hasNext()) sb=sb.append(", ");
	    }
	    sb=sb.append(")");
	}
	return sb;
    }

    StringBuffer toStringAttribs(StringBuffer sb) {
	if (noninteractive()) sb = sb.append(" noninteractive");
	if (recursive()) sb = sb.append(" recursive");
	return sb;
    }
    
    /**
     * returns a representation of the Taclet as String
     * @return string representation
     */
    public String toString() {
	if (tacletAsString == null) {
	    StringBuffer sb=new StringBuffer();
	    sb = sb.append(name()).append(" {\n");
	    sb = toStringIf(sb);
	    sb = toStringVarCond(sb);
	    sb = toStringGoalTemplates(sb);
	    sb = toStringHeuristics(sb);
	    sb = toStringAttribs(sb); 
	    tacletAsString = sb.append("}").toString();
	}
	return tacletAsString;
    }



}
