// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/** this class wraps a Taclet and collects additional information needed
 * to be able to apply it to sequent. These informations are: 
 * <ul>
 * <li> instantiations, </li>
 * <li> if necessary position of the find term </li>
 * </ul>
 */
package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.TacletInstantiationsTableModel;
import java.util.HashSet;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.*;

public abstract class TacletApp implements RuleApp {

    /** the taclet for which the application information is collected */
    private Taclet taclet;
    /** contains the instantiations of the schemavarioables of the
     * Taclet 
     */
    protected SVInstantiations instantiations = 
	SVInstantiations.EMPTY_SVINSTANTIATIONS;


    /** set of schemavariables that appear in the Taclet and need to be
     * instantiated but are not instantiated yet. This means
     * SchemaVariables in addrule-sections have to be ignored
     */
    private SetOfSchemaVariable missingVars = null;

    /** constructs a TacletApp for the given taclet, with an empty instantiation 
     *	map 
     */
    TacletApp(Taclet taclet) {
	this.taclet = taclet;
    }

    /** constructs a TacletApp for the given taclet. With some information about
     * required instantiations.
     */
    TacletApp(Taclet taclet, SVInstantiations instantiations) {
	this.taclet = taclet;
	this.instantiations = instantiations;
    }

    /** collects all bound variables above the occurrence of the schemavariable
     * whose prefix is given
     * @param prefix the TacletPrefix of the schemavariable
     * @param instantiations the SVInstantiations so that the
     * find(if)-expression matches 
     * @return set of the bound variables 
     */
    protected static SetOfQuantifiableVariable 
	boundAtOccurrenceSet(TacletPrefix prefix,
			     SVInstantiations instantiations) {	
	return collectPrefixInstantiations(prefix, 
					   instantiations);
    }


    /** collects all those logic variable that are instances of the variableSV
     * of the given prefix
     * @param pre the TacletPrefix of a SchemaVariable that is bound
     * @param instantiations the SVInstantiations to look at
     * @return the set of the logic variables whose elements are the
     * instantiations of a bound SchemaVariable appearing in the TacletPrefix
     */ 
    private static SetOfQuantifiableVariable collectPrefixInstantiations
	(TacletPrefix pre, SVInstantiations instantiations) {

	SetOfQuantifiableVariable instanceSet 
	    = SetAsListOfQuantifiableVariable.EMPTY_SET;
	IteratorOfSchemaVariable it = pre.prefix().iterator();
	while (it.hasNext()) {
	    SchemaVariable var = it.next();
	    instanceSet = 
		instanceSet.add((LogicVariable)
				((Term)instantiations.getInstantiation(var)).op());
	}
	return instanceSet;
    }


    /** returns the taclet the application information is collected for
     * @return the Taclet the application information is collected for
     */
    public Taclet taclet() {
	return taclet;
    }

    /** returns the rule the application information is collected for
     * @return the Rule the application information is collected for
     */
    public Rule rule() {
	return taclet;
    }

    /**
     * returns the instantiations for the application of the Taclet at the
     * specified position.
     * @return the SVInstantiations needed to instantiate the Taclet
     */
    public SVInstantiations instantiations() {
	return instantiations;
    }


    /**
     * resolves collisions between bound SchemaVariables in an SVInstantiation
     *
     * @param insts the original SVInstantiations
     * @return the resolved SVInstantiations
     */
    protected static SVInstantiations resolveCollisionVarSV
	(Taclet taclet, SVInstantiations insts) {

	HashMapFromLogicVariableToSchemaVariable collMap = 
	    new HashMapFromLogicVariableToSchemaVariable();
	IteratorOfEntryOfSchemaVariableAndInstantiationEntry it = 
	    insts.pairIterator();
	while (it.hasNext()) {
	    EntryOfSchemaVariableAndInstantiationEntry pair = it.next();
	    if (pair.key().isVariableSV()) {
		SchemaVariable varSV = pair.key();		
		Term value = ((TermInstantiation)pair.value()).getTerm();
		if (!collMap.containsKey((LogicVariable)
					 value.op())) {
		    collMap.put((LogicVariable)value.op(), 
				varSV);
		} else {
		    insts = replaceInstantiation(taclet, insts, varSV);	
		}
	    }
	}
	return insts;
    }

    /**
     * delivers the term below the (unique) quantifier of a bound
     * SchemaVariable in the given term.
     *
     * @param varSV the searched Bound Schemavariable
     * @param term the term to be searched in
     * @return the term below the given quantifier in the given term
     */
    private static Term getTermBelowQuantifier(SchemaVariable varSV, Term term) {
	for (int i=0; i<term.arity(); i++) {
	    for (int j=0; j<term.varsBoundHere(i).size(); j++) {
		if (term.varsBoundHere(i).getQuantifiableVariable(j)==varSV) {
		    return term.sub(i);
		} 
	    }
	    Term rec=getTermBelowQuantifier(varSV, term.sub(i));
	    if (rec!=null) {
		return rec;
	    }
	}
	return null;
    }

    /**
     * delivers the term below the (unique) quantifier of 
     * a bound SchemaVariable varSV in the find and if-parts of the Taclet 
     *
     * @param varSV the searched bound SchemaVariable 
     * @return the term below the given quantifier in the find and if-parts of 
     * the Taclet
     */
    private static Term getTermBelowQuantifier(Taclet taclet, SchemaVariable varSV) {
	IteratorOfConstrainedFormula it=taclet.ifseq().iterator();
	while (it.hasNext()) {
	    Term result=getTermBelowQuantifier(varSV, it.next().formula());
	    if (result!=null) {
		return result;
	    }
	}

	if (taclet instanceof FindTaclet) {
	    return getTermBelowQuantifier(varSV, ((FindTaclet)taclet).find());
	} else	
	    return null;
    }


    /** returns true iff the instantiation of a bound SchemaVariable contains the given
    * Logicvariable 
    * @param boundVars ArrayOfQuantifiableVariable with the
    * bound SchemaVariables
    * @param x the LogicVariable that is looked for
    * @param insts the SVInstantiations where to get the necessary
    * instantiations of the bound SchemaVariables
    * @return true iff the instantiation of a Bound Schemavariable contains the given
    * Logicvariable 
    */
    static private boolean contains(ArrayOfQuantifiableVariable boundVars,
			     LogicVariable x,
			     SVInstantiations insts) {
	for (int i=0; i<boundVars.size(); i++) {
	    Term instance = (Term) insts.getInstantiation
		((SchemaVariable)boundVars.getQuantifiableVariable(i));
	    if (instance.op() == x) {
		return true;
	    } 
	}
	
	return false;
    }

    /**
     * returns a new SVInstantiations that modifies the given
     * SVInstantiations insts at the bound SchemaVariable varSV to a new LogicVariable.
     */
    protected static SVInstantiations replaceInstantiation
	(Taclet taclet, SVInstantiations insts, SchemaVariable varSV){
	Term term=getTermBelowQuantifier(taclet, varSV);
	LogicVariable newVariable = new LogicVariable 
	    (new Name(((Term)insts.getInstantiation(varSV)).op()
		      .name().toString()+"0"),
	     varSV.sort());
	// __CHANGE__ How to name the new variable?	
	Term newVariableTerm 
	    = TermFactory.DEFAULT.createVariableTerm(newVariable);
	return replaceInstantiation(insts, term, varSV, newVariableTerm);
    }

    /**
     * returns a new SVInstantiations that modifies the given
     * SVInstantiations insts at the bound SchemaVariable u to the Term (that is
     * a LogicVariable) y.
     */
    static private SVInstantiations replaceInstantiation
	(SVInstantiations insts, Term t,  SchemaVariable u, Term y){

	SVInstantiations result = insts;
	LogicVariable x = (LogicVariable)
	    ((Term)insts.getInstantiation(u)).op();
	if (t.op() instanceof SchemaVariable) {
	    if (!((SchemaVariable)t.op()).isVariableSV()) {
		SchemaVariable sv=(SchemaVariable)t.op();
		ClashFreeSubst cfSubst
		    = new ClashFreeSubst(x,y); 
		result = result.replace
		    (sv, 
		     cfSubst.apply((Term)insts.getInstantiation(sv)));
	    }
	    else {
		for (int i=0; i<t.arity(); i++) {
		    if (!contains(t.varsBoundHere(i), x, insts)) {	       
			result = replaceInstantiation(result, t.sub(i),
						      u, y);
		    }
		}
		
	    }
	}
	result=result.replace(u, y);
	return result;
    }

    /** applies the specified rule at the specified position 
     * if all schema variables have been instantiated
     * @param goal the Goal at which the Taclet is applied
     * @param javaInfo the JavaInfo encapsulating all java information
     * @return list of new created goals 
     */
    public ListOfGoal execute(Goal goal, JavaInfo javaInfo) {
	if (!complete()) {	  
	    throw new IllegalStateException("Tried to apply rule "
					    +"that is not complete.");
	}
	goal.addAppliedRuleApp(this);	
	return taclet().apply(goal, javaInfo, this);
    }    

    /** applies the specified rule at the specified position 
     * if all schema variables have been instantiated
     * and creates a new goal if the if-sequent does not match.
     * The new goal proves that the if sequent can be derived.
     * @param goal the Goal where to apply the rule
     * @param javaInfo the JavaInfo encapsulating all java information
     * @return list of new created goals 
     */
    public ListOfGoal executeForceIf(Goal goal, JavaInfo javaInfo) {
	if (!complete()) {
	    throw new IllegalStateException("Applied rule not complete.");
	}
	goal.addAppliedRuleApp(this);	
	return taclet().apply(goal, javaInfo, this, true);
    }    

    /** calculate needed SchemaVariables that have not been
     * instantiated yet. This means to ignore SchemaVariables that
     * appear only in addrule-sections of Taclets
     * @return SetOfSchemaVariable that need to be instantiated but
     * are not
     */
    protected SetOfSchemaVariable calculateNonInstantiatedSV() {
	if (missingVars == null) {
	    missingVars = SetAsListOfSchemaVariable.EMPTY_SET;
	    TacletVariableCollector coll = new TacletVariableCollector(instantiations);
	    coll.visitWithoutAddrule(taclet());	    
	    IteratorOfSchemaVariable it = coll.varIterator();
	    while (it.hasNext()) {
		SchemaVariable var=it.next();
		if (!instantiations().isInstantiated(var)) {
		    missingVars = missingVars.add(var);	
		}
	    }	    
	}
	
	return missingVars;
    }

    /** creates a new Tacletapp where the SchemaVariable sv is
     * instantiated with the the given Term term. Sort equality is
     * checked. If the check fails an IllegalArgumentException is
     * thrown
     * @param sv the SchemaVariable to be instantiated
     * @param term the Term the SchemaVariable is instantiated with
     * @return the new TacletApp
     */
    public TacletApp addCheckedInstantiation(SchemaVariable sv, 
					   Term term) 
	throws IllegalArgumentException {

	if (!term.sort().extendsTrans(sv.sort())) {
	    throw new IllegalArgumentException
		("Could not add the instantiation of "+sv
		 +" (sort: "+sv.sort()+") because "
		 +"the sort of "+
		 term+" (sort: "+term.sort()+") is different.");
	} else if (sv.isVariableSV() && 
		   !(term.op() instanceof LogicVariable)) {
	    throw new IllegalArgumentException("Could not add "+
	       "the instantiation of "+sv+" because "+
		term+" is no variable.");
	} 
	return addInstantiation(sv, term);    
    }

    /**
     * returns the variables that have not yet been instantiated and
     * need to be instantiated to apply the Taclet. (These are not all
     * SchemaVariables like the one that appear only in the addrule
     * sections)
     * @return SetOfSchemaVariable with SchemaVariables that have not
     * been instantiated yet 
     */
    public SetOfSchemaVariable uninstantiatedVars() {
	return calculateNonInstantiatedSV();
    }
    
    
    public TacletApp tryToInstantiate(JavaInfo javaInfo) {
        TacletApp app = this;
	IteratorOfSchemaVariable it = uninstantiatedVars().iterator();
	while (it.hasNext()) {
	    SchemaVariable sv=it.next();
	    if (sv.sort()==ProgramSVSort.VARIABLE ||
		sv.sort()==ProgramSVSort.BOOLEANVARIABLE ||
		sv.sort()==ProgramSVSort.LABEL) {
		
		String proposal = TacletInstantiationsTableModel.
		   getNameProposalForSchemaVariable(sv);
		ProgramElement pe = TacletInstantiationsTableModel.
		   getProgramElement(this, proposal, sv, javaInfo);
		app = this.addInstantiation(sv, pe);
	    } else {
		return null;
	    }
	}
	return app;
    }
    
    
    /** adds a new instantiation to this TacletApp 
     * @param sv the SchemaVariable to be instantiated
     * @param term the Term the SchemaVariable is instantiated with
     * @return the new TacletApp
     */
    public abstract TacletApp addInstantiation(SchemaVariable sv, Term term);


    /** adds a new instantiation to this TacletApp 
     * @param sv the SchemaVariable to be instantiated
     * @param term the ProgramElement the SV is instantiated with
     * @return the new TacletApp
     */
    public abstract TacletApp addInstantiation(SchemaVariable sv, 
					     ProgramElement pe);


    /**
     * creates a new Taclet application containing all the
     * instantiations given 
     * by the SVInstantiations
     * @param svi the SVInstantiations whose entries are the needed
     * instantiations 
     * @return the new Taclet application
     */
    public abstract TacletApp addInstantiation(SVInstantiations svi);


    /**
     * creates a new Taclet application containing all the
     * instantiations given 
     * by the SVInstantiations and forget the old ones
     * @param svi the SVInstantiations whose entries are the needed
     * instantiations 
     * @return the new Taclet application
     */
    public abstract TacletApp setInstantiation(SVInstantiations svi);



    /**
     * returns a new PosTacletApp that is equal to this TacletApp except
     * that the position is set to the given PosInOccurrence
     * @param pos the PosInOccurrence of the newl created PosTacletApp
     * @return the new TacletApp
     */
    public PosTacletApp setPosInOccurrence(PosInOccurrence pos) {
	if (taclet() instanceof NoFindTaclet) {
	    throw new IllegalStateException("Cannot add position to an taclet"
					    +" without find");
	}
	return PosTacletApp.createPosTacletApp((FindTaclet)taclet(), 
					   instantiations(), pos);
    }

    /** returns true iff all necessary informations are collected, so
     * that the Taclet can be applied.
     * @return true iff all necessary informations are collected, so
     * that the Taclet can be applied.
     */
    public abstract boolean complete();

    /**
     * returns the PositionInOccurrence (representing a ConstrainedFormula and
     * a position in the corresponding formula)
     * @return the PosInOccurrence
     */
    public abstract PosInOccurrence posInOccurrence();


    /** compares the given Object with this one and returns true iff both are
     * from type TacletApp with equal taclets, instantiations and positions.
     */
    public boolean equals(Object o) {
	if (o==this) return true;
	if (!(o instanceof TacletApp)) return false;
	TacletApp s=(TacletApp)o;
	return (s.taclet().equals(taclet()) 
		&& s.instantiations().equals(instantiations()));
    }

    public String toString() {
	return "Application of Taclet " +taclet()+ " with "
	    +instantiations();
    }


    /**
     * checks if there are name conflicts (i.e. there are
     * two matched bound SchemaVariable that are matched to variables with an
     * equal name); if yes a new TacletApp is returned that equals this
     * TacletApp except that the name conflict is resolved by replacing the
     * instantiation of one of the conflict-causing SchemaVariables by a 
     * bound SchemaVariable with a new name; 
     * if the check is negative, the same TacletApp is returned.
     *
     * @return a conflict resolved TacletApp, remainder equal to this TacletApp
     */
    public TacletApp prepareUserInstantiation() {
	TacletApp result = this;
	SchemaVariable varSV = varSVNameConflict();
	while (varSV!=null) {
	    result = setInstantiation(replaceInstantiation
				    (taclet, result.instantiations(), varSV));
	    varSV = result.varSVNameConflict();
	}
	return result;
    }

    protected abstract SetOfQuantifiableVariable contextVars(SchemaVariable sv);

    /**
     * creates a new variable namespace by adding names of the instantiations
     * of the schema variables in the context of the given schema
     * variable and (if the TacletApp's prefix has the context flag set)
     * by adding names of the logic variables of the context.
     *
     * @param sv the schema variable to be considered
     * @param var_ns the old variable namespace
     * @return the new created variable namespace
     */
    public Namespace extendVarNamespaceForSV(Namespace var_ns, 
					     SchemaVariable sv) {
	Namespace ns=new Namespace(var_ns);
	IteratorOfSchemaVariable it=taclet().getPrefix(sv).prefix().iterator();
	while (it.hasNext()) {
	    LogicVariable var = (LogicVariable) 
		((Term)instantiations().getInstantiation(it.next())).op();
	    ns.add(var);
	}
	if (taclet().getPrefix(sv).context()) {	    
	    IteratorOfQuantifiableVariable lit
		= contextVars(sv).iterator();
	    while (lit.hasNext()) {
		ns.add(lit.next());
	    }
	}
	return ns;
    }


    /**
     * returns the bound SchemaVariable that causes a name conflict (i.e. there are
     * two matched bound SchemaVariables that are matched to variables with an
     * equal name, the returned SchemaVariable is (an arbitrary) one of these
     * SchemaVariables) or null if there is no such conflict
     *
     * @return null iff there is no conflict and one of the
     * conflict-causing bound SchemaVariables
     */
    public SchemaVariable varSVNameConflict() {
	IteratorOfSchemaVariable svIt=uninstantiatedVars().iterator();
	while (svIt.hasNext()) { 
	    SchemaVariable sv=svIt.next();
	    if (!(sv.isVariableSV())) {
		TacletPrefix prefix=taclet().getPrefix(sv);
		HashSet names=new HashSet();	    
		if (prefix.context()) {
		    IteratorOfQuantifiableVariable contextIt
			= contextVars(sv).iterator();
		    while (contextIt.hasNext()) {
			names.add(contextIt.next().name());
		    }
		}
		IteratorOfSchemaVariable varSVIt=prefix.iterator();
		while(varSVIt.hasNext()) {
		    SchemaVariable varSV=varSVIt.next();
		    Term inst = (Term)
			instantiations().getInstantiation(varSV);
		    if (inst!=null) {
			Name name = inst.op().name();
			if (!names.contains(name)) {
			    names.add(name);
			} else {
			    return varSV;
			}
		    }
		}
	    }
	}
	return null;
    }




}
