// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.metaconstruct;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.ArrayOfStatement;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.declaration.ParameterDeclaration;
import de.uka.ilkd.key.java.expression.operator.Instanceof;
import de.uka.ilkd.key.java.expression.Assignment;
import de.uka.ilkd.key.java.expression.operator.CopyAssignment;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.logic.op.ProgramVariable;

/** This class is used to perform program transformations needed 
 * for the symbolic execution of a try-throw as 
 * try {throw x; alpha} beta; where beta is the catch-finally part
 */
public class TryThrowToIf extends ProgramMetaConstruct {

     /** creates a try-throw-to-if ProgramMetaConstruct 
     * @param tryThrow the Statement contained by 
     * the meta construct 
     */
    public TryThrowToIf(Statement tryThrow) {
	super("try-throw-to-if", tryThrow); 
    }

    /** inserts a statment in the given block
     * @param stmnt the Statement to insert
     * @param b theStatementblock where to insert
     * @param pos int giving the position where to insert
     */
    private StatementBlock insertStatementInBlock(Statement stmnt, 
						  StatementBlock b, 
						  int pos) {
	
	Statement[] block = new Statement[b.getStatementCount()+1];
	int offset = 0;
	for (int i = 0; i < block.length-1; i++) {
	    if (i == pos) {
		block[i] = stmnt;
		offset = 1;
	    } 
	    block[i+offset] = b.getStatementAt(i);
	}
	// may be it should have been insert at last one
	if (offset == 0) {
	    block[pos] = stmnt;
	}
	return new StatementBlock(new ArrayOfStatement(block));	
    }

    /** performs the program transformation needed for symbolic
     * program transformation 
     * @param javaInfo the JavaInfo with all necessary information 
     * about the java programs
     * @return the transformated program
     */
    public ProgramElement symbolicExecution(ProgramElement pe, 
					    JavaInfo javaInfo,
					    SVInstantiations insts) {
	Try try_stmnt = (Try) pe;
	Throw throw_stmnt = (Throw) (try_stmnt.getBody().getChildAt(0));

	int try_branches = try_stmnt.getBranchCount()-1;
	// build else block
	StatementBlock else_block;
	Finally finally_stmnt = null;
	if (try_stmnt.getBranchAt(try_branches) instanceof Finally) {
	    finally_stmnt = (Finally) try_stmnt.getBranchAt(try_branches);
	    StatementBlock finally_block = (StatementBlock)finally_stmnt.getBody();	    
	    else_block = insertStatementInBlock
		(throw_stmnt, 
		 finally_block, finally_block.getStatementCount());
	    try_branches--;
	} else {
	    else_block = new StatementBlock(throw_stmnt);
	}
	
	// we start backwards	
	Statement if_cascade = null;
	for (int i = try_branches; i >=0; i--) {
	    ParameterDeclaration exc_decl = ((Catch)try_stmnt.getBranchAt(i)).
		getParameterDeclaration();
	    // a catch block declares exactly one parameter 
	    // catch (Exc1 e) b1
	    // the if-condition has the form (x instanceof Exc1) so we
	    // build if-condition
	    Instanceof instof = new Instanceof(throw_stmnt.getExpression(),
					       exc_decl.getTypeReferenceAt(0));
	    
	    // create then-branch
	    // finally present: { try {e = x; b1} finally {r} }
	    // finally not present: { e = x; b1} 

	    // first create assignment e = x;
	    ProgramVariable e = (ProgramVariable) exc_decl.
		getVariableSpecification().getProgramVariable();

	    Assignment assign_x_to_e = 
		new CopyAssignment(e, throw_stmnt.getExpression());
	    // now create then block
	    StatementBlock catch_block = (StatementBlock)
		((Catch)try_stmnt.getBranchAt(i)).getBody();
	    // finally not present: { e = x; b1} 
	    StatementBlock then_block = insertStatementInBlock(assign_x_to_e, catch_block, 0);
	    if (finally_stmnt != null) {
		// finally present: { try {e = x; b1} finally {r} }
		then_block = new StatementBlock
		    (new ArrayOfStatement(new Statement[]{
			new Try(then_block, new Branch[]{finally_stmnt})}));
	    } 
	    Then then_branch = new Then(then_block);
	    // create else
	    Else else_branch; 
	    if (if_cascade == null) {
		else_branch = new Else(else_block);
	    } else {
		else_branch = new Else(if_cascade);
	    }	    
	    // create if 
	    if_cascade = new If(instof, then_branch, else_branch);
	}
	if (if_cascade == null) {
	    return else_block;
	} else {
	    return if_cascade;
	}
    }
}
